
Scheduling Real-Time I/O in the HyperScore ToolKit
		Stephen Travis Pope

HUG Tech Note #3 - April, 1989

Introduction

This technical note describes the three current real-time input/output schemes available for music processing in the HyperScore ToolKit (HSTK) version 5.0 and later.
An understanding of the HyperScore ToolKit basics as described in the Description and Examples document is assumed.

Three Scheduling Policies

Because of the wide range of hardware platforms on which HSTK is being used, and the range of input/output media it can be connected to, there are several different sets of classes related to general-purpose event scheduling, MIDI event I/O, and primitive-level schedulers.
Each of these three systems is described below and illustrated in code examples.
They are called the ApointmentBook scheduler, the Score/MidiInterface system, and the MidiPrimitive scheduler )or alternatively "Worlds" 1, 2, and 3).
Their use is usually mutually exclusive; the platform and virtual machine in use will determine which scheme is appropriate.

World 1 - The AppointmentBook Scheduler

Mark Lentczner wrote the original Smalltalk-80 AppointmentBook Scheduler for Apple's first Smalltalk-80+MIDI experiments.  It uses a schedule object (an instance of class EventScheduler) which holds onto a sorted list of clients, which can be events, event lists or event-generating processes.  It can be used for time-merging and sorting for non-real-time performance such as generating cmusic notelist files.  The Boolean doWait tells it whether to run in real-time, or as fast as possible.
The scheduler's innermost loop sends the message scheduleAt: aTime to its clients components.  Note Event and EventList classes implement this for real performance.  NoteEvents pass themselves on to their voices for performance (e.g., self voice playEvent: self), which is what gives the system the pluggability of output media via voices.

The class comment for EventScheduler is included below.

The class EventScheduler is a simple driver for real-time message-passing among classes that can return timed event association values.  One uses Schedules by setting up clients which a master scheduler process continuously evaluate for their next timed activations.
Clients can return appointments which are sorted into the scheduler's list of (time -> event) associations that is evaluated continuously when running.  Changed messages are sent from the accessor messages to simplify EventScheduler browsers and inspectors.
This implementation is HSTK V5.2 based on Mark Lentczner's original 1986 design.

Instance variables:
	clients 			<Collection> client objects
	appointments 	<SortedCollection of (time -> event) Associations> the schedule
	running 		<Boolean> running schedule?
	doWait			<Boolean> whether or not the scheduler should do real-time scheduling.
					Set to false if there is a lower level of scheduling (e.g., primitive-level)

Class variable:
	Schedule 		<EventScheduler> shared instance accessed by the class message 
					masterSchedule

The class protocol of EventScheduler is:
	('class initialization' initialize release)
	('instance creation' clients:)
	('accessing ' addAppointment: runAppointments schedule)
	('interrupting' interrupt)

The instance protocol of EventScheduler is:
	('initializing' initialize)
	('accessing' addAppointment: appointments: clients: dontWait running: wait)
	('running' callNextAppointment runAppointments setUpFirstAppointments)
	('updating' release update:)
	('interrupting' interrupt)

The real work is done in the methods runAppointments and callNextAppointment, which are shown below:

runAppointments
	"Set up the first meetings and then run them all till no one wants a meeting anymore."

	running ifTrue: [^nil].
	self setUpFirstAppointments.		"set up the starting conditions"
	self running: true.
	[(running == false) or: [appointments isEmpty]] 
		whileFalse: [self callNextAppointment].
	self running: false.

callNextAppointment
	"Calls upon the next appointment to be made and then reschedules the next one"

	| entry entryClient nextTime now time |
	entry _ appointments removeFirst.			"this is an association"
	time _ entry key.							"whose key is its time"
	entryClient _ entry value.					"and whose value is the event"
	doWait ifTrue:								"wait"
		[(Delay untilMilliseconds: time) wait].
"	Transcript show: entryClient printString; cr."
	nextTime _ entry value scheduleAt: time.	"really perform the client's value now"
	nextTime == nil ifFalse: 						"it returns the next start time or nil"
			[appointments add: (nextTime => entryClient)]

Examples of implementors of scheduleAt: aTime are:

NoteEvent>>scheduleAt: aTime
	"play me then on my voice"
	
	voice == nil 
			ifTrue: [Voice defaultVoice playEvent: self at: aTime]
			ifFalse: [(Voice named: self voice) playEvent: self at: aTime].
	^nil		"return nil because i'm a leaf of the event list"

EventList>>scheduleAt: aTime 
	"really expand an eventList for the appointment scheduler"

	| newEventAssociation |
	index == nil ifTrue: [index _ 1].
	self events isEmpty ifTrue: [^nil].
	newEventAssociation _ self map: (self events at: index) at: aTime.
	EventScheduler addAppointment: newEventAssociation.
	index _ index + 1.
	index > self events size 
		ifTrue: [index _ nil.
			^nil].
	^(self events at: index) key + aTime		"return the time of my next appointment"

PlainPeal>>scheduleAt: time
	"Play the next note of the receiver"

	| order nextTime |
	. . . "compute the next note to play"
	(noteIndex = order) & secondMeasure
		ifFalse: [nextTime _ self playNote: (self at: noteIndex) at: time]
		ifTrue: [nextTime _ self playDoubleNote: (self at: noteIndex) at: time].
	. . . 

Examples of the use of the EventScheduler are:

(EventLists schedule themselves)
	EventScheduler addAppointment: (aTime => self).
	EventScheduler runAppointments

(FastMidiVoice puts the noteOff command in the schedule)
	EventScheduler addAppointment: 
		((Time millisecondClockValue + anEvent duration asMS) => noteOff).

Voices and Scheduling with the AppointmentBook Scheduler

Scores in HSTK can be kept portable across back-end performance mechanisms by using voices and voice maps to schedule them.
An event list can have the name or number of the voices it uses and this can be expanded at perform-time because the event list can perform its events by forwarding them to their voices (i.e., the event play method says something like "voice play: self")

The class hierarchy of Voices is:
	Voice ('name' 'instruments' 'stream' )
		CmusicVoice ()
		FastMidiVoice ()
		MidiVoice ()
		OpcodeVoice1 ('tempo' 'time' 'command' 'byte2' 'byte3' 'eventList' )
		OpcodeVoice2 ('fileType' 'tracks' 'ppq' 'tempo' )
		PrimMidiVoice ('clockOffset' )

MidiVoice is a simple object that can hold onto a collection of device->channelNumber associations
MidiVoice
	('initialize-release' reInitialize reset)
	('accessing' addDevice:channel: number)
	('events' playEvent: playEvent:at:)

A MidiDevice has a port and can play byteArrays out it
MidiDevice
	('initialization' initialize: reInitialize reset)
	('note events' play:on:)
	('parameter mapping' mapAmplitude: mapDuration: mapPitch:)
	('backward compatibility' play:amp: play:at:for: play:at:for:amp:)
	('private' playNoteOff:on: playNoteOn:on:)

The standard MidiVoice uses the EventScheduler and simple MidiDevice.
The MidiDevice has a serial port output stream and builds strings to dump to it. The relevant methods are included below.

MidiVoice methodsFor: 'events'
playEvent: anEvent at: aTime
	"send the given Event to all device/channel pairs"

	instruments do: [:anIn |
		anIn value play: anEvent on: anIn key]

MidiDevice methodsFor: 'note events'
play: anEvent on: channelNumber
	"play the given event on my port as a default MIDI noteOn/noteOff pair"

	| dur |
	dur _ self mapDuration: anEvent dur.
	self playNoteOn: anEvent on: channelNumber.
	dur > 0 
		ifTrue: [[(Delay forMilliseconds: dur) wait.
			self playNoteOff: anEvent on: channelNumber] fork]

MidiDevice methodsFor: 'private'
playNoteOn: anEvent on: channelNumber
	"play the note-on part of the given event"

	| bytes |
	bytes _ (ByteArray new: 3).
	bytes at: 1 put: (16r90  bitOr: channelNumber);
		at: 2 put: (self mapPitch: anEvent pitch);
		at: 3 put: (self mapAmplitude: anEvent ampl).
	port == nil 
		ifTrue: [MStream nextPutAll: Time millisecondClockValue printString, 
				'  NOn: ', bytes printString; cr]
		ifFalse: [port send: bytes length: 3]

Several other types of voices use simple model of output streams, such as the CmusicVoice.
In this case, the EventScheduler is used but the no-wait mode is set because the output file is time-stamped.

CmusicVoice methodsFor: 'events'
playEvent: anEvent at: aTime 
	"send the given Event to all device/channel pairs in cmusic format (note start instrument dur p-fields)"

	| pFields aString |
							"print out the standard fields"
	stream nextPutAll: 'note	' , (aTime / 1000) asFloat printString , '	' ,
			name , '	', anEvent duration printString.
							"select the relevant p-fields (i.e., properties with names like p5 or p21)"
	anEvent properties == nil ifFalse:
		[pFields _ anEvent properties keys select: [:aSymbol | 
			aString _ aSymbol asString.
			(aString at: 1 = $p)
				and: [(Number readFrom: (aString copyFrom: 2 to: aString size)) !!= 0]].
							"and now the extra p-fields"
		pFields asSortedCollection do: [:aProperty | 
			stream nextPutAll: '	', (anEvent perform: aProperty printString)]].
	stream nextPut: $;; cr

Setting Up and Using the Basic EventScheduler 

initialize
	"Set up the global Scheduler"
	"EventScheduler initialize."

	Schedule == nil 
		ifFalse: [self release].
	Schedule _ self new initialize.
	Smalltalk addDependent: Schedule

Evaluation of the Basic EventScheduler 

Because it is very simple and general-purpose, the AppointmentBook Scheduler can be used to schedule any event types that understand the message scheduleAt: aTime.  Because it runs in real-time in Smalltalk, it is effected by other Smalltalk-80 processes (such as the user interface), and is subject to the time that it takes to actually schedule each event.
The complexity of processing and user interface that can be run with this scheduler is related to the available compute horsepower; on very fast machines, it runs quite acceptably.  It is too slow to be used for accurate performance on Macs in combination with serial-port-based Midi output.

World 2 - Performance in John's Score/MidiInterface World

John Maloney at the University of Washington (jmaloney@june.cs.washington.edu) has contributed a faster Smalltalk-only scheduler based on his Midi-specific version of EventLists called Scores.  Scores schedule themselves using the MidiInterface object by sending their NoteElement objects out at their start times and maintaining a list of active notes for generating noteOff events at the right times.  The instance variables of NoteElements are assumed to be Midi smallIntegers instead of abstract Durations, Pitches and Amplitudes.

The class comment for MidiInterface is included below.

MidiInterface - John Maloney's Musical Instrument Digital Interface

This class allows Smalltalk programs to control synthesizers and other musical devices attached via the RS232 port and a midi interface box. Because the Midi port(s) are a shared resource, there are no instances of this class. Instead, Midi control messages are directed to the class itself. You can use the "class refs" facility to find all the users of Midi operations in the system.

The class comment for Score is included below.

I am a subclass of OrderedCollection used to store collections of notes. I know how to perform the score I contain.

My instance variables are:
	nextIndex			index of the next note to play during a performance
	activeNotes		list of active notes (notes that have been struck but not released)
	scoreTime			a cache for the score duration (scoreTime = nil or the duration)

The class protocol of MidiInterface is:
	('initialize-debug' debugOff debugOn initialize primOff primOn)
	('port control' closePort closePortPrim openPort openPort: openPortPrim: openPortStandard update:)
	('midi output' allNotesOff control:value:chan: noteOff:chan: noteOn:vel:chan: pitchBend:chan: program:chan: synthInit)
	('midi monitor' monitor monitorAllBut: monitorRaw)
	('low level - output' send:with:on: send:with:with:on:)
	('low level - input' flushInput getBuffer getByte readStream)
	('primitive tests' noopTest openCloseTest receiveTest sendTest)
	('primitives' primClose: primGet:bytes:size: primNoop primOpen: primPut:bytes:count:)

The work of this scheme takes place in the method Score>>playFrom:rate:

Score methodsFor: performance
playFrom: time rate: aRate
	"Plays the entire score at the given rate. aRate is 1 for normal speed, 2 for twice speed, 3/4 for three-quarter speed, etc."

	| mSecsAtStart currTime |
	"Scan for the first note at or after the given time, processing all control changes 
		and program changes during this scanning to establish the proper synthesizer state."
	self reset; sendEventsThrough: time - 1 silent: true.

	mSecsAtStart _ (Delay millisecondClockValue) - (time * 10 // aRate).
	[self done | Sensor anyButtonPressed] whileFalse:
		[currTime _ 
			((Delay millisecondClockValue - mSecsAtStart) * aRate) // 10.
		self sendEventsThrough: currTime silent: false.
		Processor yield].
	MidiInterface allNotesOff.	"Turn off notes in case I was interrupted."

Score methodsFor: 'private'
sendEventsThrough: time silent: silentFlag
	"Processes the score up through the given time. If silentFlag is true, then notes are skipped but all control and program changes are sent to the synthesizer."

	| event |
	self turnOffNotesAt: time.
	[(nextIndex <= self size) and: [(event _ self at: nextIndex) time <= time]]
		whileTrue:
			[((silentFlag not) or: [event isTone not])
				ifTrue: [self sendEvent: event].
			nextIndex _ nextIndex + 1].

Score methodsFor: 'private'
sendEvent: aScoreElement
	"Send the appropriate Midi messages for the given score element."

	| chan |
	aScoreElement isTone
		ifTrue: [chan _ aScoreElement chan.
			"MidiInterface program: (aScoreElement prog) chan: chan."
			MidiInterface noteOn: (aScoreElement midiKey) vel: (aScoreElement vel) chan: chan.
			activeNotes add: aScoreElement].

Examples of the use of the MidiInterface are:

	"assume a Midi-specific event object as receiver"
	MidiInterface
		noteOn: (self pitch)
		vel: (self velocity)
		chan: (self voice).

	"assume anEvent with music magnitudes as parameters"
	MidiInterface 
		noteOn: anEvent pitch asMidi 
		vel: anEvent ampl asMidi 
		chan: anEvent voice.

Setting Up and Using the MidiInterface

"set up the SerialPort Midi world"
[	MidiInterface initialize; debugOff; primOff.
	MidiInterface openPort]

MidiInterface SerialPort Primitives

John has also written a fast version of serial port output for ParcPlace Smalltalk-80 VI 2.3 running on Macintosh computers.  His notes on this are as follows.

1. Be sure to have a Midi interface attached the port you are using. The interface should have a clock speed of 1 Mhz (most do, including the one from Apple). It is possible to hang the system if you open a port that does not have a Midi interface attached (and turned on) because the port depends on a clock signal supplied by the Midi interface -- no clock, no action!!

2. The Midi driver is interrupt driven and the interrupt code is part of the Smalltalk interpreter. Thus, if you quit from Smalltalk with one of the Midi ports open and an interrupt occurs (due to a character arriving, say) then your Mac will crash because the drivers aren''t there anymore. Included in this filein is a modification of Smalltalk''s quit method that closes all Midi ports. (Closing a non-open port is harmless.) It is recommended that you include a similar mechanism in any applications you develop yourself. You will have to re-open the Midi port(s) you were using after restarting the image.

3. There is a limit of 1000 bytes per call to the Midi put primitive; you can pass it a bigger buffer but the extra bytes will not be sent.

4. The Midi put primitive puts its bytes in an internal buffer until all previous bytes have been sent. If there is not enough room in the buffer, the put primitive waits until there is room. If there IS room in in the buffer, the bytes of the current call are buffered and the call returns to Smalltalk immediately even though not all the bytes have actually been sent.

5. If the transmit buffer is completely full, it takes nearly three seconds to empty itself. (Midi data is sent at 3125 bytes/second, so you have to send a lot of data fast to actually fill up the 8192 byte transmit buffer.) If you close the Midi port before the buffer is empty, some bytes will never be sent. My practice is to keep the Midi port open continuously during a Smalltalk session but if your application opens and closes the port you may wish to put a three second delay before each close operation to absolutle guarantee that all the bytes have been sent.

Voices and MidiInterface

The FastMidiVoice uses the MidiInterface class's facilities for output.
It is usually driven by the EventScheduler, but it does not have the abstraction level of device drivers and only uses standard MIDI commands.
It can schedule standard EventLists.

FastMidiVoice methodsFor: 'note events'
play: anEvent on: channelNumber
	"play the given event on my port as a default MIDI noteOn/noteOff pair"

	| noteOff |
	MidiInterface 
		noteOn: anEvent pitch asMidi 
		vel: anEvent ampl asMidi 
		chan: anEvent voice.
	noteOff _ anEvent copy.
	noteOff ampl: 0.
	EventScheduler addAppointment: 
		((Time millisecondClockValue + anEvent duration asMS) => noteOff).

Evaluation of the Score/MidiInterface System

Because it assumes Midi-style events and maintains active note lists, this scheme is much more acceptable on smaller machines.  Scores and EventLists can be freely transformed into each other (with the messages asEventList and asScore respectively), and this scheme is useful for systems where higher-levels of interactivity are desired.  When using the serial port primitives, very good interactive performance is achieved on a Mac-II with the scheduler running in real-time in Smalltalk.

World 3 - User Primitive Scheduling in Lee and Guy's PrimitiveScheduler World

The third scheduling mechanism was written by Lee Boynton (Lee_Boynton@Next.com) and Guy Garnett (GEG%CCRMA-F4@Sail.Stanford.edu).  It involves a special Macintosh (or Sun or Atari) driver for scheduling MIDI I/O at the lowest level.
The level of abstraction seen by the Smalltalk primitives is sending time-stamped strings (really ByteArrays) to a special primitive scheduler.  There are primitives that query or set the low-level scheduler's clock and tick size.  Output can be scheduled arbitrarily long in advance.

The class protocol of Midi is:
	('mousing around' fromMouse)
	('playing' allOff play:at:dur:amp: play:at:dur:amp:voice:)
	('constants' midiMode rawMode)
	('primitive operations' close getClockTime getMode getReceiveCount getTickSize getTransmitCount 
		isOpen killIO next: nextEvent: open openAndInitialize put:count: set:data: setClockTime: 
		setReadMode: setReceiveOff setReceiveOn setTickSize: status: test)
	('accessing' getLong:at: putLong:into:at:)

The primary output method for playing MIDI events is MIDI>>play:at:dur:amp:voice:. It is shown below.

play: aPitch at: aTime dur: aDur amp: anAmp voice: aVoice

	| bytes endTime iValue |
	bytes _ ByteArray new: 8.
									"stuff start time into the first 4 bytes of the array"
	Midi putLong: aTime into: bytes at: 1.
	iValue _ 16r90.
	aVoice isNumber 
		ifTrue: [iValue _ iValue bitOr: (aVoice bitAnd: 16r0F)].
									"put the event into the array"
	bytes at: 5 put: 1; 
			at: 6 put: iValue; 
			at: 7 put: aPitch; 
			at: 8 put: anAmp.
	Midi put: bytes count: 8.		"primitive"
	endTime _ aTime + aDur.
									"stuff end time into the array"
	Midi putLong: endTime into: bytes at: 1.
	bytes at: 8 put: 0.				"amplitude 0"
	Midi put: bytes count: 8		"primitive"

An examples of the use of the MidiPrimitive scheduler is in its class example fromMouse.

fromMouse
	"Midi fromMouse"

	| pScale vScale aPoint pitch vel |
	pScale _ 127 / Display width.
	vScale _ 127 / Display height.
	[Sensor yellowButtonPressed]
		whileFalse: 
			[Sensor leftShiftDown
				ifTrue: [aPoint _ Sensor waitButton]
				ifFalse: [aPoint _ Sensor waitClickButton].
			pitch _ aPoint x * pScale.
			vel _ aPoint y * vScale.
			Midi play: pitch rounded
				at: 0
				dur: 1000
				amp: vel rounded]

The class PrimMidiVoice uses the primitive scheduler and can play events ahead of real time and have them scheduled by the driver.  No Smalltalk-level scheduler is used here, rather the play method in class EventList is changed to dump events out the driver at full speed.

PrimMidiVoice methodsFor: 'events'
playEvent: anEvent at: aTime
	"send the given Event to the global output - midi assumes"

	| realTime | 
	realTime _ ((aTime - Time millisecondClockValue) max: 0) + clockOffset.
	^Midi play: anEvent pitch asMidi
		at: realTime
		dur: anEvent duration asMS
		amp: anEvent ampl asMidi
		voice: anEvent voice

Evaluation of the Midi Primitive System

The actual output scheduling or Midi bytes is performed in Lee's driver (originally written for MidiLISP), it is independent of the Smalltalk process world and, if one schedules events ahead of real-time, can be completely insensitive to Smalltalk interaction or other processing.  The same primitive interface has now been implemented on Macintosh's, Atari's and Sun/4's.
This is the most desirable of the three schemes.

Switching between Schedulers

The main place that the schedulers are selected is in the method EventList>>play.  Several methods can be used here; all but 1 are commented out.

EventList methodsFor: 'scheduling'
play
	"play NoteEvent data from my instance variables starting now"
	"(EventLists named: #temp.1) play."

"	CHOOSE ONE OF:
				--use the AppointmentBook and my voice
	^self playOn: self voice at: (Time millisecondClockValue)
				--turn me into a Score and use the MidiInterface
	^self asScore play
				--use the primitive scheduler
	^self playPrim
"
	^self playPrim

Implementors of the messages play:on:at:, play:at:, playPrim: etc. may also depend on the type of scheduler being used.

EventList methodsFor: 'scheduling'
playOn: aVoice at: aTime
	"play event data from my instance variables starting then"
	"(EventLists named: #p1) play."

	self voice: aVoice.
	EventScheduler addAppointment: (aTime => self).
	EventScheduler runAppointments


EventList methodsFor: 'scheduling'
playPrimAt: start
	"play NoteEvent data from my instance variables starting at the given time--use the scheduling primitives"
	"(EventList named: #le1) playPrim."

	| time event relOffset |
"	Midi isOpen
			ifFalse: [Midi openAndInitialize].
"	relOffset _ Midi getClockTime.
	events do: [:anAssociation |
		time _ anAssociation key.
		event _ anAssociation value.
		(event isKindOf: EventList)
			ifTrue: [event playPrimAt: time]
			ifFalse: [Midi play: event pitch asMidi
						at: time + start + relOffset
						dur: event duration asMS
						amp: event ampl asMidi
						voice: event voice]]

Choosing the Proper Scheduler

As mentioned above, the choice of scheduler is determined by the hardware platform and Smalltalk-80 virtual machine you are using.
If you are running on a machine for which a primitive scheduler exists (Mac, Sun/4, Atari) and you are only concerned with Midi output for performance, you'll probably want to use the scheduling primitives in class Midi (i.e., world 3).
If you're on some other platform, but are using Midi output exclusively, world 2 is probably best.
If you are interested in scheduling other types of events that only Midi output (e.g., animation events) or want to use other music output media that Midi (e.g., a sound compiler or a Platypus), you can use the AppointmentBok scheduler (world 1).
