'From Smalltalk-80, Version 2.4 of 28 January 1989 on 14 March 1989 at 3:30:27 pm'!

Object subclass: #MidiInterface
	instanceVariableNames: ''
	classVariableNames: 'Debug LastCmd PitchBends Port ProgramNumbers SendBuf '
	poolDictionaries: ''
	category: 'Music-ScoreRepresentation'!
MidiInterface comment:
'See my class comment.'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

MidiInterface class
	instanceVariableNames: ''!
MidiInterface class comment:
'MidiInterface - John Maloney''s Musical Instrument Digital Interface

This class allows Smalltalk programs to control synthesizers and other musical devices attached via the RS232 port and a midi interface box. Because there is only one device, there are no instances of this class. (One could make instances but they wouldn''t do much...) Instead, messages are directed to the the class object, MidiInterface, itself.

All MidiInterface control messages are class messages!!!!
'!


!MidiInterface class methodsFor: 'initialize-debug'!

debugOff
	"MidiInterface debugOff."

	Debug _ false.!

debugOn
	"MidiInterface debugOn."

	Debug _ true.!

initialize
	"MidiInterface initialize."

	Debug isNil ifTrue: [Debug _ false].

	ProgramNumbers _
		"Fill program number cache with impossible values"
		(OrderedCollection new: 16) addAll: ((1 to: 16) collect: [:x | -1]).
	PitchBends _
		"Fill pitch bend cache with impossible values"
		(OrderedCollection new: 16) addAll: ((1 to: 16) collect: [:x | -1]).! !

!MidiInterface class methodsFor: 'port control'!

closePort
	"Close and release the serial port."
	"MidiInterface closePort."

	(Port notNil and: [Port isOpen])
		ifTrue:
			[Port close.
			 Port release].
	Port _ nil.!

isOpen
	"Answer true if the serial port appears to be open."

	^Port notNil!

openPort
	"Open the RS-232C port unless debugging is turned on."
	"MidiInterface openPort."

	(Debug not & Port isNil)
		ifTrue: [Port _ (MidiPort portNamed: '/dev/ttyb').
			 "Port setRawMode; setMidiMode"].

	SendBuf _ BufferStream on: (String new: 100).! !

!MidiInterface class methodsFor: 'midi output'!

allNotesOff
	"Turns off all midi notes on all midi channels."

	1 to: 16 do: [: ch | self control: 123 value: 0 chan: ch].!

control: control value: value chan: channel
	"Sets the given control to the given new value."

	self send: 176 with: control with: value on: channel.!

noteOff: key chan: channel
	"Turns off the given note."
	"Detail: Send a 'noteOn' command with zero velocity since not all synthesizers recognize the 'noteOff' command."

	self send: 144 with: key with: 0 on: channel.!

noteOn: key vel: velocity chan: channel
	"Turns on the given note."

	self send: 144 with: key with: velocity on: channel.!

pitchBend: bend chan: channel
	"Updates the pitch bend on this channel to the given new value."
	"Detail: The current  pitch bend for each channel is cached, to avoid sending unnecessary pitch bend commands."

	((PitchBends at: channel) == bend)
		ifFalse:
			[self
				send: 224
				with: (bend bitAnd: 127)
				with: (bend bitShift: -7)
				on: channel.
			 PitchBends at: channel put: bend].!

program: program chan: channel
	"Changes the synthesizer's 'program number' for this channel to the given new value."
	"Detail: The current  program number for each channel is cached, to avoid sending unnecessary program change commands."

	((ProgramNumbers at: channel) == program)
		ifFalse:
			[self send: 192 with: (program - 1) on: channel.
			 ProgramNumbers at: channel put: program].!

synthInit
	"Resets the synthesizer. This may need to be changed since devices differ; the idea is to put the synthesizer into a known state for all parameters which might be changed through Midi."

	1 to: 16 do:
		[: channel |
			self control: 123 value: 0 chan: channel.	"All notes off on this channel."
			PitchBends at: channel put: -1.				"Force a cache miss on the next pitch bend command."
			self pitchBend: 8192 chan: channel.
			ProgramNumbers at: channel put: -1.		"Force a cache miss on the next program command."
			"self program: 1 chan: channel."
				"commented out to avoid annoying changes of synth presets"
			"You may wish to reset various other control values here."].! !

!MidiInterface class methodsFor: 'midi monitor'!

midiMonitor
	"This is a simple midi monitor program which deciphers incoming
	midi commands and prints them to the transcript until a mouse
	button is pressed."

	self monitorAllBut: #().!

monitorAllBut: cmdList
	"Print midi messages to the transcript until any mouse button is pressed."
	"MidiInterface midiMonitor"

	| recorder |
	recorder _ MidiRecorder new.
	"ignore the given commands"
	cmdList do: [: cmd | recorder ignoreCmd: cmd].
	recorder monitor.!

monitorRaw
	"Shows the raw midi data as it arrives."
	"MidiInterface monitorRaw."

	| c |

	self flushInput.
	[Sensor anyButtonPressed]
		whileFalse: [
			c _ self getByte.
			((c ~= #NoInput) & (c ~= 254))
				ifTrue: [Transcript show: c printString; cr].
		].! !

!MidiInterface class methodsFor: 'low level - output'!

flushOutbuf
	"Flushes buffered output characters."

	(SendBuf isEmpty not & Debug not & Port notNil)
		ifTrue:
			[Port readAppendStream nextPutAll: SendBuf contents.
			 SendBuf reset
			"Transcript show: SendBuf contents printString; cr"].!

putByte: aByte
	"Transmits aByte to the Midi device."

	Debug
		ifTrue: [Transcript show: (aByte printString, '\') withCRs]
		ifFalse: [SendBuf nextPut: aByte].!

send: midiCode with: arg on: channel
	"Sends a one-argument midi channel message. Checks argument and channel number."

	| ch cmd |
	ch _ channel - 1.
	cmd _ midiCode bitOr: ch.
	((ch bitAnd: 15) == ch)
		ifFalse: [self error: 'bad channel number']
		ifTrue: [((arg bitAnd: 127) == arg)
				ifFalse: [self error: 'bad Midi argument']
				ifTrue: [Debug
						ifTrue: [Transcript show:
									(cmd printString, ' ',
									 arg printString, '\') withCRs]
						ifFalse: [SendBuf nextPut: cmd asCharacter.
								  SendBuf nextPut: arg asCharacter]]]!

send: midiCode with: arg1 with: arg2 on: channel
	"Sends a two-argument midi channel message. Checks arguments and channel number."

	| ch cmd |
	ch _ channel - 1.
	cmd _ midiCode bitOr: ch.
	((ch bitAnd: 15) == ch)
		ifFalse: [self error: 'bad channel number']
		ifTrue: [(((arg1 bitAnd: 127) == arg1) and: [(arg2  bitAnd: 127) == arg2])
				ifFalse: [self error: 'bad Midi argument']
				ifTrue: [Debug
						ifTrue: [Transcript show:
										"(Time millisecondClockValue / 1000) asInteger printString,"
										((cmd printStringRadix: 16), ' ',
										 (arg1 printStringRadix: 16), ' ',
										 (arg2 printStringRadix: 16), '\') withCRs]
						ifFalse: [SendBuf nextPut: cmd asCharacter.
								 SendBuf nextPut: arg1 asCharacter.
								 SendBuf nextPut: arg2 asCharacter]]]! !

!MidiInterface class methodsFor: 'low level - input'!

flushInput
	"Empties the serial port's input buffer."

	Port setMidiMode.
	Port stream == nil 
		ifTrue: [ Port receiveBuffer ].
"	[Port atEnd]
		whileFalse: [Port receiveBuffer]."!

getByte
	"Gets input without waiting. Answers #NoInput if there isn't any."

	(Debug | Port isNil)
		ifTrue: [^#NoInput]
		ifFalse:
			[Port atEnd ifTrue: [Port receiveBuffer].	"try reading into buffer"
			 Port atEnd	"still at end?"
				ifTrue:  [^#NoInput]
				ifFalse: [Port next]].!

readStream

	Port receiveBuffer.
	^Port stream! !

MidiInterface initialize!
