Object subclass: #KeySignature
	instanceVariableNames: 'notes accidentals '
	classVariableNames: 'PitchClasses '
	poolDictionaries: ''
	category: 'Music-Basics'!
KeySignature comment:
'KeySignatures are objects that can influence Pitch transposition or printing via the pitch map functions.

Instance variables:
	notes			<Array of Numbers> the notes of the KS
	accidentals		<Array of Numbers> which notes have accidentals

(N.B. these remain to be implemented)'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

KeySignature class
	instanceVariableNames: ''!


!KeySignature class methodsFor: 'examples'!

ksExample
	"Use a KeySig to constrina a note"
	"KeySignature ksExample"

	| aP aKS |
	aP _ Pitch value: 46. "a#3"
	aKS _ (KeySignature on: #f) major
	Transcript show: aP asSymbol.
	
	Transcript show: (aP mapToKey: aKS) asSymbol.! !

Object subclass: #PitchClass
	instanceVariableNames: 'name accidentals ordering '
	classVariableNames: 'Modulus NameSpace ThePitches '
	poolDictionaries: ''
	category: 'Music-Basics'!


!PitchClass methodsFor: 'accessing'!

baseName
	"Answer the name that this pitch class is based on"

	^ name!

order
	"Answer the order within the chromatic scale that this pitch class falls into."

	^ ordering! !

!PitchClass methodsFor: 'scales'!

major
	"Answer a collection of pitch classes taht form a major scale starting of the receiver."

	| scale lastClass |
	scale _ OrderedCollection new.
	scale addFirst: self.
	lastClass _ self.
	#(2 4 5 7 9 11) do: [ :offset |
		lastClass _ (PitchClass after: lastClass
								order: ordering + offset).
		scale addLast: lastClass. ].
	^ scale!

minor
	"Answer a collection of pitch classes taht form a minor scale starting of the receiver."

	| scale lastClass |
	scale _ OrderedCollection new.
	scale addFirst: self.
	lastClass _ self.
	#(2 3 5 7 9 10) do: [ :offset |
		lastClass _ (PitchClass after: lastClass
								order: ordering + offset).
		scale addLast: lastClass. ].
	^ scale! !

!PitchClass methodsFor: 'printing'!

printOn: aStream
	"Write out the name of the receiver pitch class"

	name printOn: aStream.
	1 to: accidentals do: [aStream nextPut: $#].
	accidentals to: -1 do: [aStream nextPut: $b].! !

!PitchClass methodsFor: 'private'!

name: baseName accidentals: offset order: value
	"Initialize the receiver from the given values"

	name _ baseName.
	accidentals _ offset.
	ordering _ value.!

nextBaseName
	"Answer the base name next in the series after the receiver's base name."

	| nameIndex |
	nameIndex _ NameSpace indexOf: name.
	nameIndex _ nameIndex \\ NameSpace size + 1.
	^ (NameSpace at: nameIndex) key! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

PitchClass class
	instanceVariableNames: ''!


!PitchClass class methodsFor: 'initialization'!

initialize
	"Default PitchClass set si the Twelve Tone system."

	self initializeTwelveTonal.
	self initializeFirstPitches.

	"PitchClass initialize"!

initializeFirstPitches
	"Given that NameSpace and Modulus are set up, add the base pitch classes to ThePitches"

	ThePitches _ Array new: Modulus.
	1 to: Modulus do: [ :i | ThePitches at: i put: Set new].
	NameSpace do: [ :nameDefinition |
		(ThePitches at: nameDefinition value + 1) add:
			(self name: nameDefinition key
				accidentals: 0
				order: nameDefinition value) ].!

initializeTwelveTonal
	"Set up the class to work within the twelve tone system"

	Modulus _ 12.
	NameSpace _ OrderedCollection new: 7.
	'CDEFGAB' with: #(0 2 4 5 7 9 11) do: [ :className :order |
		NameSpace addLast: (String with: className) -> (order printString) ].! !

!PitchClass class methodsFor: 'accessing'!

after: anotherClass order: value
	"Answer the pitch class after anotherClass (in name) that has the order indicated.  Do this by looking it up in our private list of
pitch classes, and if it's not there then create it."

	| newBase newPc |
	newBase _ anotherClass nextBaseName.
	(ThePitches at: value + 1) do: [ :pc |
		pc baseName = newBase ifTrue: [ ^ pc ] ].
	^ self name: newBase order: value!

named: baseName
	"Answer the pitch class whoes named is as given and has no accidentals.  Do this by looking it up in our private list of pitch
classes."

	| order |
	order _ (NameSpace at: (NameSpace indexOf: baseName)) value.
	(ThePitches at: order + 1) do: [ :pc |
		pc baseName = baseName ifTrue: [ ^ pc ] ].
	^ self error: 'UnInitialized base name ' , baseName! !

!PitchClass class methodsFor: 'private'!

name: baseName accidentals: offset
	"Actually create an instance of the receiver who's base name and accidentals are as given.  To do this, the order of the resultant
class must be determined."

	| baseOrder |
	baseOrder _ (NameSpace at: (NameSpace indexOf: baseName)) value.
	^ self name: baseName
		accidentals: offset
		order: baseOrder + offset \\ Modulus!

name: baseName accidentals: offset order: value
	"Actually create an instance of the receiver and answer it"

	| newSelf |
	newSelf _ super new.
	newSelf name: baseName accidentals: offset order: value.
	(ThePitches at: value + 1) add: newSelf.
	^ newSelf!

name: baseName order: value
	"Actually create an instance of the receiver who's base name and order are as given.  To do this, the number of accidentals must be
determined."

	| baseOrder offset |
	baseOrder _ (NameSpace at: (NameSpace indexOf: baseName)) value.
	offset _ value - baseOrder \\ Modulus.
	offset > (Modulus / 2) ifTrue: [ offset _ offset - Modulus ].
	^ self name: baseName
		accidentals: offset
		order: value! !

PitchClass initialize!
