Association subclass: #EventAssociation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Events'!
EventAssociation comment:
'EventAssociations are used for the elements of EventLists.
They are just like regular Associations, but require both key and value to be equal for the Association to be equal.
EventAssociations are created with the message ''=>'' sent to any object, e.g., 
	21 => anEvent'!


!EventAssociation methodsFor: 'accessing'!

event
	"answer my event"

	^self value!

start
	"answer my start"

	^self key! !

!EventAssociation methodsFor: 'comparing'!

= anAssociation
	"Answer whether the receiver is equal to the argument."

"	^key = anAssociation key
"	key = anAssociation key
		ifTrue: [^value == anAssociation value]
		ifFalse: [^false]! !

!EventAssociation methodsFor: 'printing'!

printOn: aStream 
	"Append to the argument, aStream, the two elements of the 
	Association separated by a right arrow."

	key printOn: aStream.
	aStream nextPutAll: '=>'.
	value printOn: aStream! !

Object variableSubclass: #MEvent
	instanceVariableNames: 'duration props '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Events'!
MEvent comment:
'MEvent is the most abstract event class used in the HSTK V5.0 event list processing library.
MEvents are simply objects with names and possible durations that can also be used as dictionaries or property lists themselves.
Their properties can be accessed either with at: and at:put: or by using the new property names themselves as selectors (e.g., (aMEventSubClassInstance at: #color put: #green) or (aMEventSubClassInstance color: #green)).
	
Instance variables:
	duration 	<number, symbol, special duration class> my duration
	props		<Dictionary or nil> my extended property list
					standard items are #name, etc.

The global dictionary called Events can be used for sharing event instances'!


!MEvent methodsFor: 'accessing'!

copy
	"to avoid stepping on other events' parameters--do a deep copy"

	^self deepCopy!

duration
	"return my duration."

	^duration!

duration: newValue
	"set my duration."

	duration _ newValue!

inspect
	"use a special inspector for MEvent types"

	EventInspectorView openWithButtons: (EventInspector inspect: self)!

name
	"return my name."

	^self at: #name!

name: newValue
	"set my name."

	self at: #name put: newValue!

species
	"all subclasses look like me"

	^MEvent! !

!MEvent methodsFor: 'properties'!

at: aProp
	"return a value from the property list dictionary"

	props == nil 
		ifTrue: [^nil].
	^props at: aProp ifAbsent: [nil]!

at: aProp ifAbsent: otherCase
	"return a value from the property list dictionary"

	props == nil 
		ifTrue: [^otherCase].
	^props at: aProp ifAbsent: otherCase!

at: aProp put: aVal
	"set a value in the property list dictionary"

	props == nil ifTrue: [
		props _ Dictionary new  ].
	props at: aProp put: aVal!

doesNotUnderstand: aMessage 
	"override DNU to try to access the property dictionary.
	If this is unsuccessful, announce that the receiver does not understand the argument as a message."
	"one may want to turn this off for debugging"

	| sel val symbol |
	sel _ aMessage selector.
	(sel  includes: $:) ifFalse: [		"access property"
		(props == nil or: [(props includesKey: sel) == false]) ifTrue: [
			NotifierView openContext: thisContext
				label: 'Message not understood: ', aMessage selector
				contents: thisContext shortStack
		  ].
		^props at: sel
	]  ifTrue: [				"set property"
		^self at: (sel copyUpTo: $:) asSymbol put: (aMessage arguments at: 1)
	].!

properties
	"return my property dictionary"

	^props! !

!MEvent methodsFor: 'testing'!

isEvent
	"we're all events somehowe"

	^true! !

!MEvent methodsFor: 'scheduling'!

scheduleAt: aTime
	"set a schedule for me to play myself at my special time."

	^self subclassResponsibility! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

MEvent class
	instanceVariableNames: ''!


!MEvent class methodsFor: 'class initialization'!

initialize
	"Initialize the global dictionary of Events."
	"MEvent initialize"

	Smalltalk at: #Events put: Dictionary new! !

MEvent initialize!


Object subclass: #EventScheduler
	instanceVariableNames: 'clients appointments running doWait '
	classVariableNames: 'Schedule '
	poolDictionaries: ''
	category: 'Music-Events'!
EventScheduler comment:
'The class EventScheduler is a simple driver for real-time message-passing among classes that can return timed event association values.  
One uses Schedules by setting up clients which a master scheduler process continuously evaluate for their next timed activations.
Clients can return appointments which are sorted into the scheduler''s list of (time -> event) associations that is evaluated continuously when running.
Changed messages are sent from the accessor messages to simplify EventScheduler browsers and inspectors.
This implementation is based on Mark Lentczner''s original 1986 design.

Instance variables:
	clients 			<Collection> client objects
	appointments 	<SortedCollection of (time -> event) Associations> the schedule
	running 		<Boolean> running schedule?
	doWait			<Boolean> whether or not the scheduler should do real-time scheduling
						set to false if there is a lower level of scheduling 
						(e.g., primitive-level) or if output need not be real-time

Class variable:
	Schedule 		<EventScheduler> shared instance accessed by the class message masterSchedule
'!


!EventScheduler methodsFor: 'initializing'!

initialize
	"set up the default Schedule"
	
	self clients: nil.
	self running: false.
	doWait _ false! !

!EventScheduler methodsFor: 'accessing'!

addAppointment: theApp
	"add the argument, 'theApp', to the instance variable 'appointments'."

	appointments == nil ifTrue: [
			self setUpFirstAppointments ].
	appointments add: theApp.
	self changed: #appointments.!

appointments: theAppointments
	"Accept the argument, 'theAppointments', as the new instance variable 'appointments'."

	appointments _ theAppointments.
	self changed: appointments!

clients: theClients
	"Accept the argument, 'theClients', as the new instance variable 'clients'."

	clients _ theClients.
	self changed: #clients!

dontWait
	"the scheduler should NOT do the waiting--i.e., there is a lower level of scheduling going on"
	"EventScheduler schedule dontWait"

	doWait _ false!

running: theBool
	"Accept the argument, 'theBool', as the instance variable 'running'."

	running _ theBool.
"	Transcript show: 'Schedule: ', theBool printString; cr.
"	self changed: #running!

wait
	"the scheduler should do the waiting"
	"EventScheduler schedule wait"

	doWait _ true! !

!EventScheduler methodsFor: 'running'!

callNextAppointment
	"Calls upon the next appointment to be made and then reschedules the next one"

	| entry entryClient nextTime now time |
	entry _ appointments removeFirst.			"this is an association"
	time _ entry key.
	entryClient _ entry value.
	(doWait == true) ifTrue:
		[(Delay untilMilliseconds: time) wait ].	"wait"
"	Transcript show: entryClient printString; cr."
	nextTime _ entry value scheduleAt: time.	"it returns the next start time or nil"
	nextTime == nil ifFalse: 
			[appointments add: (nextTime => entryClient)]!

runAppointments
	"Set up the first meetings and then run them all till no one wants a meeting anymore."

	running ifTrue: [ ^nil ].
	self setUpFirstAppointments.
	self running: true.
	[(running == false) or: [appointments isEmpty]] whileFalse:
		[self callNextAppointment].
	self running: false.!

setUpFirstAppointments
	"Create the appointment list and initialize it with the initial appointments."

	| firstTime |
	appointments == nil ifTrue: [
		appointments _ SortedCollection new ].
	clients == nil ifFalse: [
		clients do: [:user |
			firstTime _ user firstSchedule.
			appointments add: (firstTime => user) ] ]! !

!EventScheduler methodsFor: 'updating'!

release
	"clean up the schedule"

	appointments _ nil.
	clients _ nil.!

update: anAspect
	"check if clients are waiting"

	appointments == nil ifFalse: [
		self running: false.
		appointments _ nil
	]! !

!EventScheduler methodsFor: 'interrupting'!

interrupt
	"stop all running schedules by brute force"

	self running: false! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

EventScheduler class
	instanceVariableNames: ''!


!EventScheduler class methodsFor: 'class initialization'!

initialize
	"Set up the global Scheduler"
	"EventScheduler initialize."

	Schedule == nil ifFalse: [
		self release
	].
	Schedule _ self new initialize.
	Smalltalk addDependent: Schedule!

release
	"clear away the global EventScheduler"
	"EventScheduler release"

	Smalltalk removeDependent: Schedule.
	Schedule release.
	Schedule _ nil! !

!EventScheduler class methodsFor: 'instance creation'!

clients: clientList
	"add a client list to the newly-created Scheduler"

	| newSelf |
	newSelf _ self new.
	newSelf clients: clientList.
	^ newSelf! !

!EventScheduler class methodsFor: 'accessing '!

addAppointment: anApp
	"add the argument, 'theApp', to the instance variable 'appointments'."

	Schedule addAppointment: anApp!

runAppointments
	"Set up the first meetings and then run them all till no one wants a meeting anymore."

	Schedule runAppointments!

schedule
	"return the global Scheduler"

	^Schedule! !

!EventScheduler class methodsFor: 'interrupting'!

interrupt
	"stop all running schedules by brute force"

	EventScheduler allInstances do: [ :thisOne |
		thisOne interrupt ]! !

EventScheduler initialize!


MEvent variableSubclass: #NoteEvent
	instanceVariableNames: 'voice pitch loudness '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Events'!
NoteEvent comment:
'Instances of class NoteEvent (subclass of MEvent), are concrete musical note event objects used for eventLists and eventGenerators.
Voices are now used as drivers for numerical or symbolic event parameters and there is generally no need to subclass NoteEvent.

Their instance variables are:
	pitch 		<Pitch> my pitch or note number
	voice 		<Symbol, Voice or Integer> my timbre, instrument or voice
	loudness 	<Amplitude> my amplitude, loudness or volume'!


!NoteEvent methodsFor: 'accessing'!

ampl
	"return my loudness"

	^loudness!

ampl: aValue
	"set my loudness"

	^self loudness: aValue!

dur
	"return my duration"

	^duration!

dur: aValue
	"set my duration"

	^self duration: aValue!

duration
	"return my duration"

	^duration!

duration: aValue
	"set my duration"

	(aValue isKindOf: Duration)
		ifFalse: [ duration _ Duration value: aValue ]
		ifTrue: [ duration _ aValue ]!

loudness
	"return my loudness"

	^loudness!

loudness: aValue
	"set my loudness"

	(aValue isKindOf: Amplitude)
		ifFalse: [ loudness _ Amplitude value: aValue ]
		ifTrue: [ loudness _ aValue ]!

pitch
	"return my pitch"

	^pitch!

pitch: aValue
	"set my pitch"

	pitch _ aValue!

voice
	"return my voice"

	^(voice == nil) ifTrue: [0] ifFalse: [voice]!

voice: aValue
	"set my voice"

	voice _ aValue! !

!NoteEvent methodsFor: 'scheduling'!

play
	"play me now"
	
	voice playEvent: self!

schedule
	"play me now"
	
	voice playEvent: self!

scheduleAt: aTime
	"play me then on my voice"
	
	voice == nil ifTrue: [
		Voice defaultVoice playEvent: self at: aTime.
	] ifFalse: [
		(Voice named: self voice) playEvent: self at: aTime ].
	^nil! !

!NoteEvent methodsFor: 'printing'!

printOn: aStream 
	"format and print the receiver note"

	aStream nextPutAll: ' Evt'.
	duration == nil ifFalse:
		[ aStream nextPutAll: ' ', duration printString ]. 
	pitch == nil ifFalse: 
		[ aStream nextPutAll: ' ', pitch printString ].
	voice == nil ifFalse: 
		[ aStream nextPutAll: ' V: ', voice printString ].
	loudness == nil ifFalse: 
		[ aStream nextPutAll: ' ', loudness printString ].
	aStream cr!

storeOn: aStream 
	"format and store the source the receiver note"

	aStream nextPutAll: '(NoteEvent'.
	duration == nil ifFalse:
		[ aStream nextPutAll: ' dur: ', duration value storeString ]. 
	pitch == nil ifFalse: 
		[ aStream nextPutAll: ' pitch: ', pitch value storeString ].
	loudness == nil ifFalse: 
		[ aStream nextPutAll: ' ampl: ', loudness value storeString ].
	voice == nil ifFalse: 
		[ aStream nextPutAll: ' voice: ', voice storeString ].
	aStream nextPutAll: ')'! !

!NoteEvent methodsFor: 'processing'!

transposeBy: aStep
	"add the given step to the receiver note--assumes we're woring with key numbers, not pitch degrees"

	self pitch: (Pitch value: (self pitch + aStep))! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

NoteEvent class
	instanceVariableNames: ''!


!NoteEvent class methodsFor: 'instance creation'!

dur: aD pitch: aP
	"return an initialized NoteEvent"
	"NoteEvent dur: 1000 pitch: 30"
	"NoteEvent dur: 1/4 pitch: #c3"

	^self new duration: (Duration value: aD); pitch: (Pitch value: aP)!

dur: aD pitch: aP ampl: anA
	"return an initialized NoteEvent"
	"NoteEvent dur: 1000 pitch: 30 ampl: 127 "
	"NoteEvent dur: 1/4 pitch: #c3 ampl: #mf "

	^self new duration: (Duration value: aD); pitch: (Pitch value: aP); 
		loudness: (Amplitude value: anA)!

dur: aD pitch: aP ampl: anA voice: aVoice
	"return an initialized NoteEvent"
	"NoteEvent dur: 1000 pitch: 30 ampl: 127 voice: 2"
	"NoteEvent dur: 1/4 pitch: #c3 ampl: #mf voice: #flute"

	^self new duration: (Duration value: aD); pitch: (Pitch value: aP); 
		loudness: (Amplitude value: anA); voice: aVoice!

dur: aD pitch: aP voice: aVoice
	"return an initialized NoteEvent"

	^self new duration: (Duration value: aD); pitch: (Pitch value: aP); voice: aVoice!

pitch: aP
	"return an initialized NoteEvent"
	"NoteEvent pitch: 440.0"
	"NoteEvent pitch: #c3"

	^self new pitch: (Pitch value: aP)!

pitch: aP ampl: anA voice: aVoice
	"return an initialized NoteEvent"

	^self new pitch: (Pitch value: aP); 
		loudness: (Amplitude value: anA); voice: aVoice! !

NoteEvent variableSubclass: #EventList
	instanceVariableNames: 'events index '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Events'!
EventList comment:
'Instances of class EventList (subclass of NoteEvent in this release, could be generalized to MEvent with little cost), are used for holding onto multiple timed events.
EventLists are events themselves and can have arbitrary properties as well as a collection of (time -> event) associations.

The Instance variables are:
	events 		<SortedCollection of (time -> event) Associations> my events
	index 		<Integer> the current position (if you care)

Standard properties are #tempoScale for the tempo scaling factor (used by the CMN event list editors as the MS/whole note scale).
The pool dictionary EventLists is a shared source of instances that can be referenced with their own symbols and often have gensym-like names such as #temp.23.'!


!EventList methodsFor: 'initialize-release'!

initializeAnonymous
	"make an un-named instance of me with default state."

	duration _ 0.
	self index: 1.
	self name: 'anonymous'.
	self code: #durations.
	self events: SortedCollection new.!

initializeNamed: aName
	"make a named instance of me with default state."

	self initializeAnonymous.
	self name: aName asString.
	EventLists at: aName asSymbol put: self! !

!EventList methodsFor: 'accessing'!

add: anEorA
	"add a new event or association to myself."

	(anEorA isKindOf: Association) 
		ifTrue:
			[self events add: anEorA.
			anEorA value duration == nil ifFalse:
				[self duration: (duration max: 
					(anEorA key + anEorA value duration asInteger))]]
		ifFalse:
			[self events add: (duration => anEorA).
			self duration: duration + anEorA duration].
	self changed: #added!

add: anEvent at: aTime
	"add a new event to myself."

	self events add: (aTime => anEvent).
	anEvent duration == nil 
		ifTrue: [duration _ duration max: aTime ]
		ifFalse: [duration _ duration max: (aTime + anEvent duration) ].
	self changed: #added!

addAll: anEventList
	"add the given event list's events to myself."

	anEventList events do:
		[ :assoc |
			self events add: (assoc key copy => assoc value deepCopy).
			assoc value duration == nil 
				ifTrue: [duration _ duration max: assoc key ]
				ifFalse: [duration _ duration max: (assoc key + assoc value duration) ]].
	self changed: #added!

addModifier: aModifier
	"add the given event modifier to my collection"

	(props at: #modifiers) == nil 
		ifTrue: [props at: #modifiers put: OrderedCollection new].
	(props at: #modifiers) add: aModifier!

code
	"set the code about my events, usually #durations or #noteOnOff"

	^props at: #code ifAbsent: [nil]!

code: aSymbol
	"set the code about my events, usually #durations or #noteOnOff"

	self at: #code put: aSymbol!

deepCopy
	"be careful about copying these"

	| newEL |
	newEL _ EventList newAnonymous.
	newEL name: self name deepCopy.	
	props == nil ifFalse:
		[ props associationsDo: [ :anAssociation |
			newEL at: anAssociation key put: anAssociation value ]].
	self events do: [ :anAssociation |
		newEL add: ((anAssociation key copy) -> (anAssociation value copy)) ].
	^newEL!

defaultEventClass
	"What kind of Events do i normally use?"

	^NoteEvent!

duration: aValue
	"set my duration"

	duration _ aValue!

empty
	"throw away my events"
	self duration: 0.
	self index: 1.
	self events: SortedCollection new.!

events
	"return my eventDictionary."

	^events!

events: anEL
	"set my eventDictionary."

	events _ anEL.
	self changed: #added!

index: aNumber
	index _ aNumber!

isEventList
	"WE ARE"

	^true!

links
	"return the links i have to others."

	^props at: #links ifAbsent: [nil]!

linkTo: anEL
	"set a link to another eventList."

	(props at: #links) == nil 
		ifTrue: [props at: #links put: Set new ].
	(props at: #links) add: anEL!

next
	index == nil ifTrue: [ index _ 1 ].
	^self events at: (index _ index + 1)!

play: aNote at: aTime dur: aDur amp: anAmp
	"add the given event to the list--useful for using lists to play generators onto"

	self add: ((self defaultEventClass new) 
				pitch: (Pitch value: aNote); 
				dur: (Duration value: aDur); 
				ampl: (Amplitude value: anAmp))
		at: aTime!

removeModifier: aModifier
	"remove the given event modifier from my collection"

	(props at: #modifiers ifAbsent: [nil]) == nil 
		ifTrue: [^nil].
	(props at: #modifiers) remove: aModifier ifAbsent: []!

reset
	index _ 1!

shallowCopy
	"be careful about copying these"

	| newEL |
	newEL _ self shallowCopy.
	newEL events: SortedCollection new.
	self events do: [ :anAssociation |
		newEL add: ((anAssociation key copy) => (anAssociation value copy)) ].
	^newEL!

species
	"for my subclasses, we're all eventLists"

	^EventList!

tempo: someValue
	"set the tempo I scale my events by"

	self at: #tempo put: someValue! !

!EventList methodsFor: 'collecting'!

collect: aBlock
	"iterate the given block over my events, collecting the results"

	| newList |
	newList _ EventList new.
	self events do: [ :anE |
		newList add: anE key => (aBlock value: anE) ].
	^newList!

do: aBlock
	"iterate the given block over my events"

	self events do: aBlock!

expanded
	"expand all sub-event lists and return one large (non-hierarchical) list"

	| newList newEvent |
	newList _ EventList newAnonymous.
	events do: [ :anEvent |					"iterate over all my events"
		newEvent _ anEvent value "copy".		"copy them"
		(newEvent species == EventList) 
			ifTrue:							"unfold sub-event lists recursively"
				[ newEvent expanded do: [ :aSubEvent |
					newList add: (newEvent map: 
						((anEvent key + aSubEvent key) 
							-> aSubEvent value)) ]]
			ifFalse: 							"map and add normal events"
				[ newList add: ( anEvent key -> newEvent) ]].
	^newList!

expandedFrom: start to: end
	"expand all sub-event lists and return one large (non-hierarchical) list"

	| newList newEvent |
	newList _ EventList newAnonymous.
	events do: 
		[ :anEvent |				"iterate over all my events"
		(anEvent key between: start and: end) 
			ifTrue: [(newEvent isEventList)		"map and add normal events"
				ifFalse: [newList add: 
							(anEvent key => anEvent value copy)]
										"unfold sub-event lists recursively"
				ifTrue: [(newEvent expandedFrom: 0 
								to: (end - anEvent key)) 
						do: [ :aSubEvent |
							newList add: (newEvent map: 
									((anEvent key + aSubEvent key) 
										=> aSubEvent value))]]]].
	^newList!

includes: anAss
	"answer whether the given association is in the event list"

		^(self events includes: anAss)!

notesDo: aBlock
	"iterate the given block over my events' events"

	^self events do: [:assoc | aBlock value: (assoc event) ]!

remove1: anA
	"remove the given event from my collection"

	| victim |
	victim _ self events detect: [ :item |
		(item key = anA key) and: [ item value = anA value ]].
	Transcript show: victim printString; cr.
	self events removeIndex: (self events indexOf: victim).
	self changed: #deleted!

remove: anA
	"remove the given event from my collection"

	self events remove: anA.
	self changed: #deleted!

select: aBlock
	"iterate the given block over my events"

	| newList |
	newList _ EventList new.
	self events do: [ :anE |
		 (aBlock value: anE) ifTrue: [
			newList add: anE key => (aBlock value: anE) ] ].
	^newList!

timesDo: aBlock
	"iterate the given block over my events' start times"

	^self events do: [:assoc | aBlock value: (assoc key) ]! !

!EventList methodsFor: 'processing'!

addVoice: theVoice
	"set the voice of all of my events to be the given argument"

	| anEvent |
	self events do: [ :anAssociation |
		anEvent _ anAssociation value.
		(anEvent isKindOf: EventList)
			ifTrue: [ anEvent addVoice: theVoice ]
			ifFalse: [ anEvent voice: (anEvent voice + theVoice) ]]!

apply: aFunction to: aSelector
	"apply the given function to the range of the given selector"

	^self apply: aFunction to: aSelector startingAt: 0!

apply: aFunction to: aSelector startingAt: sTime
	"apply the given function to the range of the given selector"

	| aStartTime anEvent val |
	self events do:
		[ :anAssoc |
		aStartTime _ anAssoc key.
		anEvent _ anAssoc value.
		anEvent isEventList 
			ifTrue: [anEvent apply: aFunction to: aSelector startingAt: aStartTime]
			ifFalse: [Sensor leftShiftDown ifTrue: 
						[Transcript show: (((anEvent perform: aSelector) * 
							(aFunction at: (aStartTime / self duration))) 
											rounded) printString; tab].
					anEvent perform: ((aSelector, ':') asSymbol) 
						with: ((anEvent perform: aSelector) * 
							(aFunction at: (aStartTime / self duration))) rounded]]!

chooseDuration: theDurationRange
	"set the duration to an element from the given interval or array"

	| anEvent rand size |
	rand _ Random new.
	size _ theDurationRange size.
	self events do: [ :anAssociation |
		anEvent _ anAssociation value.
		(anEvent isKindOf: EventList)
			ifTrue: [ anEvent setDuration: theDurationRange ]
			ifFalse: [ anEvent duration: 
						(theDurationRange at: ((rand next * size) truncated + 1))]]!

chooseLoudness: theLoudnessRange
	"set the loudness to an element from the given interval or array"

	| anEvent rand size |
	rand _ Random new.
	size _ theLoudnessRange size.
	self events do: [ :anAssociation |
		anEvent _ anAssociation value.
		(anEvent isKindOf: EventList)
			ifTrue: [ anEvent setLoudness: theLoudnessRange ]
			ifFalse: [ anEvent loudness: 
						(theLoudnessRange at: ((rand next * size) truncated + 1))]]!

chooseVoice: theVoiceRange
	"set the voice to an element from the given interval or array"

	| anEvent rand size |
	rand _ Random new.
	size _ theVoiceRange size.
	self events do: [ :anAssociation |
		anEvent _ anAssociation value.
		(anEvent isKindOf: EventList)
			ifTrue: [ anEvent setVoice: theVoiceRange ]
			ifFalse: [ anEvent voice: 
						(theVoiceRange at: ((rand next * size) truncated + 1))]]!

scaleDuration: theScale
	"scale the duration of all of my events by the given argument"

	| anEvent |
	self events do: [ :anAssociation |
		anEvent _ anAssociation value.
		(anEvent isKindOf: EventList)
			ifTrue: [ anEvent scaleDuration: theScale ]
			ifFalse: [ anEvent duration: (anEvent duration asMS * theScale) rounded ]]!

scaleLoudness: theScale
	"scale the loudness of all of my events by the given argument"

	| anEvent |
	self events do: [ :anAssociation |
		anEvent _ anAssociation value.
		(anEvent isKindOf: EventList)
			ifTrue: [ anEvent scaleLoudness: theScale ]
			ifFalse: [ anEvent loudness: (anEvent loudness * theScale) ]]!

scaleTempo: theScale
	"scale the duration of all of my events by the given argument"

	| anEvent |
	self events do: [ :anAssociation |
		anEvent _ anAssociation value.
		(anEvent isKindOf: EventList)
			ifTrue: [ anEvent scaleTempo: theScale ]
			ifFalse: [ anAssociation key: (anAssociation key * theScale) rounded ]]!

setDuration: theDuration
	"set the duration of all of my events to be the given argument"

	| anEvent |
	self events do: [ :anAssociation |
		anEvent _ anAssociation value.
		(anEvent isKindOf: EventList)
			ifTrue: [ anEvent setDuration: theDuration ]
			ifFalse: [ anEvent duration: theDuration ]]!

setLoudness: theLoudness
	"set the loudness of all of my events to be the given argument"

	| anEvent |
	self events do: [ :anAssociation |
		anEvent _ anAssociation value.
		(anEvent isKindOf: EventList)
			ifTrue: [ anEvent setLoudness: theLoudness ]
			ifFalse: [ anEvent loudness: theLoudness ]]!

setVoice: theVoice
	"set the voice of all of my events to be the given argument"

	| anEvent |
	self events do: [ :anAssociation |
		anEvent _ anAssociation value.
		(anEvent isKindOf: EventList)
			ifTrue: [ anEvent setVoice: theVoice ]
			ifFalse: [ anEvent voice: theVoice ]]!

spreadDuration: theRandomPercent
	"spread the duration of all of my events by the given random range (%+-)"

	| anEvent rand |
	rand _ Random new.
	self events do: [ :anAssociation |
		anEvent _ anAssociation value.
		(anEvent isKindOf: EventList)
			ifTrue: [ anEvent spreadDuration: theRandomPercent ]
			ifFalse: [ anEvent duration: 
				(anEvent duration asMS * 
					(1 + ((rand next - 0.5) * theRandomPercent / 50))) rounded ]]!

spreadLoudness: theRandomPercent
	"spread the loudness of all of my events by the given random range (%+-)"

	| anEvent rand |
	rand _ Random new.
	self events do: [ :anAssociation |
		anEvent _ anAssociation value.
		(anEvent isKindOf: EventList)
			ifTrue: [ anEvent spreadLoudness: theRandomPercent ]
			ifFalse: [ anEvent loudness: 
				(anEvent loudness asMidi * 
					(1 + ((rand next - 0.5) * theRandomPercent / 50))) rounded ]]! !

!EventList methodsFor: 'scheduling'!

play
	"play NoteEvent data from my instance variables starting now"
	"(EventLists named: #temp.1) play."

"	CHOOSE ONE OF:
	^self playOn: self voice at: (Time millisecondClockValue)
	^self asScore play
	^self playPrim
"
	^self playPrim!

playAt: aTime
	"play NoteEvent data from my instance variables starting then"
	"(EventLists named: #p1) playAt: (Time millisecondClockValue + 100)."

"	^self playOn: self voice at: aTime"
	self playPrimAt: aTime!

playOn: aVoice
	"play NoteEvent data from my instance variables starting then"
	"(EventLists named: #p1) playOn: Voice defaultMap."

	^self playOn: aVoice at: (Time millisecondClockValue + 100)!

playOn: aVoice at: aTime
	"play event data from my instance variables starting then"
	"(EventLists named: #p1) play."

	self voice: aVoice.
	EventScheduler addAppointment: (aTime => self); 
					runAppointments!

playPrim
	"play NoteEvent data from my instance variables starting now--use the scheduling primitives"
	"(EventLists named: #temp.1) playPrim."

	^self playPrimAt: 0!

playPrimAt: start
	"play NoteEvent data from my instance variables starting at the given time--use the scheduling primitives"
	"(EventList named: #le1) playPrim."

	| time event relOffset |
"	Midi isOpen
			ifFalse: [ Midi openAndInitialize ].
"	relOffset _ Midi getClockTime.
	events do: [ :anAssociation |
		time _ anAssociation key.
		event _ anAssociation value.
		(event isKindOf: EventList)
			ifTrue: [event playPrimAt: time]
			ifFalse: [Midi play: event pitch asMidi
						at: time + start + relOffset
						dur: event duration asMS
						amp: event ampl asMidi
						voice: event voice ]]!

scheduleAt: aTime 
	"really expand an eventList for the appointment scheduler"

	| newEventAssociation |
	index == nil ifTrue: [ index _ 1 ].
	self events isEmpty ifTrue: [ ^nil ].
	newEventAssociation _ self map: (self events at: index) at: aTime.
	EventScheduler addAppointment: newEventAssociation.
	index _ index + 1.
	index > self events size ifTrue: [
		index _ nil.
		^ nil ].
	^(self events at: index) key + aTime		"return the time of my next appointment"! !

!EventList methodsFor: 'user interface'!

edit
	"open your favorite editor on me"

	Sensor leftShiftDown
		ifFalse: [HSEventListView openWith8ButtonsAndHAndVScrollingOn: self]
		ifTrue: [PianoView open2StaffOn: self]!

inspect
	"use a special inspector for MEvent types"

	EventInspectorView openWithButtons: (EventInspector inspect: self)!

setRanges
	"set my header variables with my instruments, ranges, etc."
	"(EventList named: #b1) setRanges."

	| p1 p2 a1 a2 instrs du duE |
	p1 _ 255.	"min pitch and amplitudes"
	a1 _ 255.
	p2 _ 0.	"max pitch and amplitudes"
	a2 _ 0.
	du _ 0.
	instrs _ Set new.	"instruments"
	events values do: [ :anE | 
		anE pitch > p2 ifTrue: [ p2 _ anE pitch ].
		anE pitch < p1 ifTrue: [ p1 _ anE pitch ].
		anE ampl > a2 ifTrue: [ a2 _ anE ampl ].
		anE ampl < a1 ifTrue: [ a1 _ anE ampl ].
		(instrs includes: anE voice)
			ifFalse: [ instrs add: anE voice ] 
	].
	self at: #prange put: (Array with: p1 with: p2).
	self at: #arange put: (Array with: a1 with: a2).
	instrs do: [ :anI | anI _ anI bitAnd: 15 ].	"mask instrument numbers"
	self at: #instruments put: instrs.
	self duration: du! !

!EventList methodsFor: 'code conversion'!

asScore
	"create and return a MIDI Score for the receiver's events"

	| list score time event |
	list _ self expanded.
	score _ Score new.
	list events do: [ :association |
		time _ association key.
		event _ association value.
		score add: (NoteElement new: (event pitch asMidi) 
				at: time // 10
				dur: (event duration asMS // 10) 
				vel: (event loudness asMidi)
				voice: (event voice == nil ifTrue: [1] ifFalse: [event voice max: 1]))].
	^score!

findNoteOff0: anAssociation
	"find the noteOff event that corresponds to anAssociation - return the sum of the delta times"

	| instrument assoc1 assoc2 time2 event2 event1 |
	instrument _ anAssociation value voice.
	event1 _ anAssociation value.
	assoc1 _ anAssociation.
	[ true ] whileTrue:
		[ assoc2 _ events after: assoc1.  		"check next"
		assoc2 = assoc1 ifTrue: [ self halt ].
		assoc2 isNil ifTrue: 
				[ ^self duration - anAssociation key ].	"if missing."
		event2 _ assoc2 value.
"		Transcript show: '.'."
		event2 voice = instrument ifTrue:		"if same instrument"
			[ event2 ampl asMidi = 0 ifTrue:	"if ampl = 0"
				[ event2 pitch = event1 pitch ifTrue: 		"if same note"
					[ ^assoc2 key - anAssociation key ]]]. 	"^time"
		event1 _ event2.
		assoc1 _ assoc2 ].					"next event"!

findNoteOff: anAssociation
	"find the noteOff event that corresponds to anAssociation - return the sum of the delta times"

	| instrument time assoc1 assoc2 event2 event1 |
	instrument _ anAssociation value voice.
	time _ 0.
	event1 _ anAssociation value.
	assoc1 _ anAssociation.
	[ true ] whileTrue: [
		assoc2 _ events after: assoc1.  			"check next"
		assoc2 isNil ifTrue: [ ^time ].				"if missing."
		time _ time + (assoc2 key).					"sum delta times"
		event2 _ assoc2 value.
		event2 voice = instrument ifTrue: [		"if same instrument"
			event2 ampl asMidi = 0 ifTrue: [		"if ampl = 0"
				event2 pitch = event1 pitch ifTrue: [	"if same note"
					^ assoc2 key - anAssociation key 		"^time"
				 ]]].
				event1 _ event2.
				assoc1 _ assoc2 ].		"next event"!

setDurations
	"take a delta-time encoded event list and determine the durations"

	| time sTime event newEventList anAss ind |
	(self code == #noteOnOff) ifFalse:				"if i'm delta not coded"
				[^nil ].
	newEventList _ SortedCollection new.
	ind _ 1.
	[ ind <= events size ] whileTrue: [			"for all note on events"
		anAss _ events at: ind.
		sTime _ anAss key.
		event _ anAss value deepCopy.
		(event ampl value) > 0 ifTrue:			"until note off found"
			[ ((event voice) bitAnd: 16rF0) = 16r90 ifTrue:
				[ time _ self findNoteOff: ind.
				time isNil ifTrue:
					[ self error: 'no note off found...' ].
				event duration: (Duration value: time).
				event voice: (event voice bitAnd: 16r0F)."strip the Midi command"
				newEventList add: (sTime => event).
				Transcript show: ind printString, ' ' ]
			ifFalse:							"meta or control event"
				[Transcript show: 'meta event'; cr.
				newEventList add: (sTime => event) ]].
		ind _ ind + 1 ].
	self code: #durations.
						"now remove noteOffs"
	self events: newEventList.
	self start: newEventList first key.
	self at: #events put: (events size).
	self changed! !

!EventList methodsFor: 'printing'!

printOn: aStream
	"print out all of my events"

	| anEvent count |
	aStream nextPutAll: '(EventList ', (self name printString), '  dur:	', duration printString, 
			'  events:	', events size printString.
	pitch == nil ifFalse: [ aStream nextPutAll: '	pitch:	', pitch printString ].
	(self at: #tempo) == nil ifFalse: [ aStream nextPutAll: '	tempo:	', (self at: #tempo) printString ].
	aStream cr.
	Sensor leftShiftDown 
			ifFalse: [ count _ events size min: 20 ]
			ifTrue: [ count _ events size ].
	1 to: count do: [ :ind |
		anEvent _ events at: ind.
		aStream nextPutAll: anEvent key printString, ' -> '.
		anEvent value printOn: aStream].
	count = events size 
			ifTrue: [aStream nextPut: $); cr ]
			ifFalse: [aStream nextPutAll: ' ... )'; cr ].!

storeOn: aStream
	"print out all of my events"

	| anEvent count |
	aStream nextPutAll: '((EventList  newNamed: ', (self name printString), ')'.
	pitch == nil ifFalse: [aStream nextPutAll: '	pitch: ', pitch printString].
	(self at: #tempo) == nil ifFalse: [pitch == nil ifFalse: [aStream nextPut: $;].
				aStream nextPutAll: ' tempo:	', (self at: #tempo) printString].
	aStream cr.
	count _ events size.
	1 to: count do: [ :ind |
		anEvent _ events at: ind.
		aStream nextPutAll: 'add: (', anEvent key printString, ' => '.
		anEvent value storeOn: aStream.
		aStream nextPut: $).
		ind < count ifTrue: [aStream nextPut: $;; cr]].
	aStream nextPutAll: ')'.!

storeOnFile: aName
	"store out all of my events on the named file"

	| fs |
	fs _ FileStream fileNamed: aName.
	self storeOn: fs.
	fs close.! !

!EventList methodsFor: 'private'!

map: anAssociation
	"map my special properties and/or eventModifiers onto the given event"

	| aTime anE |
	aTime _ anAssociation key.
	anE _ anAssociation value.
	pitch == nil 			"if i'm transposed"
		ifFalse: [ anE pitch: (Pitch value: (anE pitch + pitch)) ].
	loudness == nil 		"or scaled"
		ifFalse: [ anE loudness: (anE loudness + loudness) ].
		(props == nil) 
			ifFalse: [(props at: #tempo ifAbsent: [nil]) == nil		"or have a tempo map"
				ifFalse: [aTime _ aTime * ((props at: #tempo) at: aTime).
					anE duration: anE duration * ((props at: #tempo) at: anE duration)]].
	^anAssociation!

map: anAssociation at: startTime
	"map my special properties and/or eventModifiers onto the given event"

	| aTime anE |
	aTime _ anAssociation key.
	anE _ anAssociation value copy.
	(self at: #pitch) == nil 			"if i'm transposed"
		ifFalse: [ anE pitch: (anE pitch + (self at: #pitch)) ].
	(self at: #loudness) == nil 		"or scaled"
		ifFalse: [ anE loudness: (anE loudness + (self at: #loudness)) ].
	(self at: #tempo) == nil 			"or have a tempo map"
		ifFalse: [  aTime _ aTime * ((self at: #tempo) at: aTime).
				anE duration: anE duration * ((self at: #tempo) at: anE duration) ].
	^(aTime + startTime) -> anE! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

EventList class
	instanceVariableNames: ''!


!EventList class methodsFor: 'instance creation'!

named: myName fromPitches: pitchCollection
	"answer an event list made with the given collection of pitches"
	"EventList named: (EventList nextName) fromPitches: #(36 38 40 43 45)"

	| me |
	me _ self newNamed: (myName asSymbol).
	pitchCollection do:
		[ :item |
		me add: (me defaultEventClass pitch: item)
			at: 0 ].
	^me!

new
	"set up and return a new default-named instance of me"

	^super new initializeNamed: self nextName!

newAnonymous
	"set up and return a new default un-named instance of me"

	^super new initializeAnonymous!

newNamed: aName
	"set up and return a new named instance of me"

	^super new initializeNamed: aName! !

!EventList class methodsFor: 'class initialization'!

initialize
	"set up the shared EL dictionary"
	"EventList initialize."

	Smalltalk at: #EventLists put: Dictionary new! !

!EventList class methodsFor: 'named constants'!

loadAllFrom: pattern
	"EventList loadAllFrom: 'Scores:'"

	| size |
	size _ pattern size + 1.
	(FileDirectory filesMatching: pattern, '*.gio') do:
		[ :fName |
		Transcript show: (fName copyFrom: size to: fName size - 4); cr.
		EventLists at: (fName copyFrom: size to: fName size - 4) asSymbol
				put: (AdagioParser parse: 
						(FileStream oldFileNamed: fName)) asEventList].
	(FileDirectory filesMatching: pattern, '*.el') do:
		[ :fName |
		(FileStream oldFileNamed: fName) fileIn].!

named: aName
	"return the named instance or nil"

	^self named: aName ifAbsent: [ nil ]!

named: aName ifAbsent: theBlock
	"return the named instance or nil"

	^EventLists at: aName ifAbsent: theBlock value!

nextName
	"return a gensym-type of event list name"

	| temp |
	temp _ 1.
	[EventLists includesKey: ('temp.', temp printString)] whileTrue: [
		temp _ temp + 1].
	^'temp.', temp printString! !

EventList initialize!
