Path subclass: #Function
	instanceVariableNames: 'size min max array '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Functions'!
Function comment:
'Function is an abstraction of a 1-dimensional function of time.
Functions are created from an array of points over the unity interval; x varies from 0.0 to 1.0 and y is free over that range.  
Examples are a ramp from 0 to 1 such as:
	LinSeg from: #((0 @ 0) (1 @ 1))

or a spline that traces a sine-like path centered around 1:
	SplineSeg from: #((0@1) (0.33@2) (0.67@0) (1@1))

Instance varibles:
	size		<Integer> number of elements
	min		<Number> min value
	max	<Number> max value
	array 	<Array of Numbers> my value array'!


!Function methodsFor: 'accessing'!

at: anIndex
	"return the value at the given index"

	^self subclassResponsibility! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Function class
	instanceVariableNames: ''!


!Function class methodsFor: 'instance creation'!

from: anArrayOfPoints
	"answer a function with the given points"

	| me |
	me _ self new: anArrayOfPoints size.
	anArrayOfPoints do: [ :anArray | me add: (anArray first @ anArray last) ].
	^me! !

!Function class methodsFor: 'examples'!

functionExample
	"Function usage example; make a roll-type eventList and apply a crescendo/decrescendo to it"
	"Function functionExample"

	| temp fcn |
	temp _ EventList newNamed: #test3.
	(0 to: 4000 by: 50) do: 	"4 seconds, 20 notes per second"
			[ :index |			"add the same note"
			temp add: (NoteEvent dur: 100 pitch: 36 ampl: 100) 
					at: index].
	fcn _ LinSeg from: #((0 @ 0) (0.5 @ 1) (1 @ 0)).
	temp apply: fcn to: #loudness.
	temp play! !

Function subclass: #SplineSeg
	instanceVariableNames: 'myForm '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Functions'!
SplineSeg comment:
'Instances of SplineSeg are cubic splines betwen their points.'!


!SplineSeg methodsFor: 'accessing'!

at: anIndex
	"return the value at the given index in my range--take it out of my computed form"

	| count realIndex box |
	box _ self computeBoundingBox.
	count _ 1.
	realIndex _ anIndex * (box width) / (myForm width).
	[myForm valueAt: (realIndex @ count) = 0] 
		whileTrue:
				[count _ count + 1].
	^count * (box width) / (myForm width)! !

!SplineSeg methodsFor: 'computing'!

computeCurve
	"compute the block by cubic interpolation--draw into a 512@100 form."

	| box newSpline xScale yScale xOffset yOffset |
	box _ self computeBoundingBox.
	box extent: (box extent - self form extent).
	Transcript show: box printString; cr.
	myForm _ Form extent: 512 @ 100.		"cached Form extent"
	xScale _ myForm width / box width.
	yScale _ myForm height / box height.
	xOffset _ box top * box width.
	yOffset _ box left * box height.
	newSpline _ Spline new: self size.
	collectionOfPoints do:
		[ :aPoint |
		newSpline add: (((aPoint x * xScale + xOffset) rounded) 
						@ ((aPoint y * yScale + yOffset)) rounded)].
	newSpline form: (Form dotOfSize: 1).
	newSpline computeCurve; displayOn: myForm.
"	myForm displayAt: 20@20."
	^self! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SplineSeg class
	instanceVariableNames: ''!


!SplineSeg class methodsFor: 'instance creation'!

from: anArrayOfPoints
	"answer a function with the given points"

	| me |
	me _ super from: anArrayOfPoints.
	^me computeCurve! !

!SplineSeg class methodsFor: 'examples'!

SSExample
	"make a SplineSeg"

"	| fcn |
	fcn _ SplineSeg from: #((0@1) (0.35@1.3) (0.65@0.7) (1@1)).
	fcn inspect.
	^fcn at: 0.14 
"! !

Function subclass: #Fourier
	instanceVariableNames: 'myForm myArray '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Functions'!
Fourier comment:
'Instances of Fourier are functions that interpret their points as 3-element arrays (harmonic, amplitude, phase), and sum sine waves into their array given the fourier summation of these components.

Instance variables:
	myForm		<Form> my function plot--default size = 512@100
	myArray	<Array of Floats> my value array--default length = 512'!


!Fourier methodsFor: 'accessing'!

at: theIndex
	"return the value from my form--assume an index in the range 0 to 1"

	^myArray at: (theIndex * myArray size)!

form
	^myForm!

function
	"answer the form I displayed on"

	^myArray! !

!Fourier methodsFor: 'computing'!

computeCurve
	"compute the block by sine summation--draw into a 512@100 form and set the values in the cached array."

	| width height x y twoPi scale |
	width _ 512.
	height _ 50.0.
	myForm _ Form extent: width @ (height * 2) truncated.		"cached Form extent"
	myArray _ Array new: width.
	twoPi _ 2.0 * Float pi.
	scale _ twoPi / width.
	form _ Form dotOfSize: 1.
	1 to: width do:
		[ :index |
		x _ index asFloat* scale.
		y _ 0.0.
		collectionOfPoints do:
			[ :aZPoint |				"each point is (number @ amplitude @ phase)"
			y _ y + (((x * aZPoint x asFloat) + (aZPoint z asFloat / 360.0 * twoPi)) sin 
							* aZPoint y asFloat)].
		myArray at: index put: y.
		y _ (y negated * height + height) rounded.
		form displayOn: myForm at: index@y].! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Fourier class
	instanceVariableNames: ''!


!Fourier class methodsFor: 'instance creation'!

from: anArrayOfZPoints
	"answer a sum-of-sines function with the given points"

	| me |
	me _ self new: anArrayOfZPoints size.
	anArrayOfZPoints do: 
			[ :anArray | 
			me add: (anArray first @ (anArray at: 3) @ anArray last) ].
	me computeCurve.
	^me! !

!Fourier class methodsFor: 'examples'!

FExample
	"make a Sine summation that approaches a square wave"

"	| fcn |
	fcn _ Fourier from: #((1@1@0) (3@0.3@0) (5@0.2@0) 
							(7@0.15@0) (9@0.11@0) (11@0.09@0)).
	fcn form displayOn: Display.
	Transcript show: (fcn at: 0.14) printString; cr.
	fcn inspect.
"! !

Function subclass: #LinSeg
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Functions'!
LinSeg comment:
'Instances of LinSeg are line-segment functions of one free variable.'!


!LinSeg methodsFor: 'accessing'!

at: anIndex
	"return the value at the given index in my range--do the linear interpolation by hand"

	| xVal count x1 x2 y1 y2|
	xVal _ collectionOfPoints first x.
	count _ 1.
	[xVal <= anIndex] whileTrue:
		[count _ count + 1.
		xVal _ (collectionOfPoints at: count) x].
	xVal = anIndex 
		ifTrue: [^(collectionOfPoints at: count) y]
		ifFalse: [x1 _ (collectionOfPoints at: (count - 1)) x.
			x2 _ (collectionOfPoints at: count) x.
			y1 _ (collectionOfPoints at: (count - 1)) y.
			y2 _ (collectionOfPoints at: count) y.
			^y2 - ((x2 - anIndex) * (y2 - y1)) ]! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

LinSeg class
	instanceVariableNames: ''!


!LinSeg class methodsFor: 'examples'!

LSExample
	"make a LinSeg"

"	| fcn |
	fcn _ LinSeg new: 2.
	fcn add: (0@0.5); add: (1@1).
	^fcn at: 0.14
"! !