Object subclass: #Midi
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-MidiPrimitives'!
Midi comment:
'Midi driver written by Guy E. Garnett using the Macintosh driver written by Lee Boynton.
See the class messages for output and examples.
Standard Midi Command format
		Note On = 9n kk vv
		Note Off = 8n kk vv
		Key Pressure = An kk vv
		Pitch Wheel = En ll ll
		After Touch Channel Pressure = Dn vv
		Control Change = Bn cc vv
		Program change = Cn pp
		System Exclusive = F0
		End of Sys Ex =	F7
		Active Sensing = FE

Relevant Constants for description of MIDI commands:
		n = 4-bit channel number
		kk = 7-bit key number
		vv = 7-bit key velocity
		ll = 7-bit low-order value
		hh = 7-bit high-order value
		cc = 7-bit control number
		pp = 7-bit program number
'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Midi class
	instanceVariableNames: ''!


!Midi class methodsFor: 'mousing around'!

fromMouse
	"Midi fromMouse"

	| pScale vScale aPoint pitch vel |
	pScale _ 127 / Display width.
	vScale _ 127 / Display height.
	[Sensor yellowButtonPressed]
		whileFalse: 
			[Sensor leftShiftDown
				ifTrue: [aPoint _ Sensor waitButton]
				ifFalse: [aPoint _ Sensor waitClickButton].
			pitch _ aPoint x * pScale.
			vel _ aPoint y * vScale.
			Midi
				play: pitch rounded
				at: 0
				dur: 1000
				amp: vel rounded]! !

!Midi class methodsFor: 'playing'!

allOff
	"Midi allOff"

	| bytes |
	Midi killIO.
	bytes _ ByteArray new: 8 withAll: 0.
	0 to: 127 do: [:aKey | 0 to: 15 do: 
			[:aChan | 
			bytes at: 5 put: 1; at: 6 put: 144 + aChan; at: 7 put: aKey.
			Midi put: bytes count: 8]]!

play: aPitch at: aTime dur: aDur amp: anAmp 
	| bytes endTime |

"	Transcript show: aPitch printString, '	', aTime printString, '	', 
		aDur printString, '	', anAmp printString; cr.
"
	bytes _ ByteArray new: 8.
	Midi putLong: aTime into: bytes at: 1.
	bytes at: 5 put: 1; 
			at: 6 put: 144; 
			at: 7 put: aPitch; 
			at: 8 put: anAmp.
	Midi put: bytes count: 8.
	endTime _ aTime + aDur.
	Midi putLong: endTime into: bytes at: 1.
	bytes at: 8 put: 0.
	Midi put: bytes count: 8!

play: aPitch at: aTime dur: aDur amp: anAmp voice: aVoice

	| bytes endTime iValue |
	bytes _ ByteArray new: 8.
	Midi putLong: aTime into: bytes at: 1.
	iValue _ 16r90.
	aVoice isNumber 
		ifTrue: [iValue _ iValue bitOr: (aVoice bitAnd: 16r0F)].
	bytes at: 5 put: 1; 
			at: 6 put: iValue; 
			at: 7 put: aPitch; 
			at: 8 put: anAmp.
	Midi put: bytes count: 8.
	endTime _ aTime + aDur.
	Midi putLong: endTime into: bytes at: 1.
	bytes at: 8 put: 0.
	Midi put: bytes count: 8! !

!Midi class methodsFor: 'constants'!

midiMode

	^(1 bitShift: 24)!

rawMode

	^0! !

!Midi class methodsFor: 'primitive operations'!

close

	<primitive: -102>
	^Transcript show: 'Could not close Midi. probably not open anyway!!';cr!

getClockTime

	^self status: 0!

getMode
	^(self status: 4)
		bitShift: -16!

getReceiveCount
	| results data | 

	results _ Array new: 2.
	data _ self status: 5.
	results at: 1 put: (data bitShift: -16); at: 2 put: (data bitAnd: 16rFFFF).
	^results!

getTickSize

	^self status: 7!

getTransmitCount
	| results data | 

	results _ Array new: 2.
	data _ self status: 6.
	results at: 1 put: (data bitShift: -16); at: 2 put: (data bitAnd: 16rFFFF).
	^results!

isOpen
	"answer whether or not the Midi output primitives are initialized"
	"Midi isOpen"

	^(self getTickSize) ~= -21!

killIO

	^self set: 1 data: 0!

next: aByteArray

	<primitive: -103>
	^self primitiveFailed!

nextEvent: aByteArray
	| time | 

	self next: aByteArray.
	[(aByteArray at: 5) = 2]
		whileTrue:[self next: aByteArray].
	time _ Midi getLong: aByteArray at: 1.
	^time!

open
	"Midi open"

	<primitive: -101>
	^self primitiveFailed!

openAndInitialize
	"open and set-up Midi output primitives"
	"Midi openAndInitialize"

	self isOpen 
			ifTrue: [ self close ].
	self open.
	self isOpen
			ifFalse: [ self error: 'Midi is down...' ].
	self setClockTime: 0.
	self setTickSize: 1!

put: aByteArray count: anInteger

	<primitive: -104>
	^self primitiveFailed!

set: anIntCode data: anInteger
	"anIntCode = 0, setClockTime; 	anInteger = milliSeconds
				= 1, kill IO; 		anInteger ignored
				= 2, receive On;	anInteger ignored
				= 3, receive Off;	anInteger ignored
				= 4, readMode;	anInteger = 0, raw mode else standardMode
				= 7, set tickSize; 	anInteger = milliSeconds"

	<primitive: -106>
	^self primitiveFailed!

setClockTime: msTime

	^self set: 0 data: msTime!

setReadMode: flag
	"flag is 0 for raw mode, else standard midi data format"

	^self set: 4 data: flag!

setReceiveOff

	^self set: 3 data: 0!

setReceiveOn

	^self set: 2 data: 0!

setTickSize: inMilliseconds

	^self set: 7 data: inMilliseconds!

status: anIntCode
	"anIntCode = 0, getClockTime
				= 4, readMode
				= 5, receiveCount
				= 6, transmitCount
				= 7, return tickSize"

	<primitive: -105>
	^self primitiveFailed!

test
	<primitive: -201>
	^self primitiveFailed! !

!Midi class methodsFor: 'accessing'!

getLong: aByteArray at: pos 
	"| aBA | 
	aBA _ ByteArray new: 8. 
	Midi putLong: 1234567 into: aBA at: 1. 
	Midi getLong: aBA at: 1."

	| a32BitW |
	a32BitW _ (aByteArray at: pos) bitShift: 24.
	a32BitW _ a32BitW + ((aByteArray at: pos + 1) bitShift: 16).
	a32BitW _ a32BitW + ((aByteArray at: pos + 2) bitShift: 8).
	a32BitW _ a32BitW + (aByteArray at: pos + 3).
	^a32BitW!

putLong: a32BitW into: aByteArray at: pos

	aByteArray at: pos put: ((a32BitW bitShift: -24) bitAnd: 16rFF).
	aByteArray at: pos+1 put: ((a32BitW bitAnd: 16rFF0000) bitShift: -16).
	aByteArray at: pos+2 put: ((a32BitW bitAnd: 16rFF00) bitShift: -8).
	aByteArray at: pos+3 put: (a32BitW bitAnd: 16rFF).
	^aByteArray! !