ScrollController subclass: #PianoRollController
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-MidiScores'!
PianoRollController comment:
'I control interactions which modify the wirelist''s ordering.  I use a PopUpMenu to get my instructions.'!


!PianoRollController methodsFor: 'initialize'!

initialize
	"Set up my yellow button menu."

	| myMenu myMenuMessages |

	myMenu _
		PopUpMenu
			labels:
'Play All
Play Window
Play Selected Notes
Play Time Range
Clear Selection
Voice -> Selection
Selection -> Voice
Delete Selection
Record
Set Rate
Inspect Notes
Show Voice Counts'
			lines: #(4 5 7).
	myMenuMessages _
		#(playAll
		   playWindow
		   playSelectedNotes
		   playTimeRange
		   clearSelection
		   addVoiceToSelection
		   moveSelectionToVoice
		   deleteSelection
		   record
		   setRate
		   inspectNotes
		   showVoiceCounts).

	super initialize.
	self
		yellowButtonMenu: myMenu
		yellowButtonMessages: myMenuMessages.! !

!PianoRollController methodsFor: 'controller handling'!

controlActivity
	"Handle mouse activities."

	| nearestNote cursorPoint newPoint mode |

	"we only deal with red button activities"
	(Sensor redButtonPressed)
		ifFalse: [^super controlActivity].

	"drag a time selection marker"
	(view selectionBar containsPoint: Sensor cursorPoint)
		ifTrue:
			[view dragTimeMarker.
			 ^self].

	"find the nearest note to the initial mouse position"
	nearestNote _ view nearestNote: Sensor cursorPoint.

	"return if no nearby notes"
	(nearestNote isNil)
		ifTrue: [
			[Sensor redButtonPressed] whileTrue: ["wait for button to be released"].
			^self].

	"if the indicated note is not already in the selection, use add mode; otherwise, use remove mode"
	(view selected includes: nearestNote)
		ifFalse: [
			mode _ #adding.
			view addToSelection: nearestNote]
		ifTrue: [
			mode _ #removing.
			view removeFromSelection: nearestNote].

	[Sensor redButtonPressed]
		whileTrue:
			[nearestNote _ view nearestNote: Sensor cursorPoint.
			 nearestNote notNil
				ifTrue: [
					(mode == #adding)
						ifTrue: [view addToSelection: nearestNote]
						ifFalse: [view removeFromSelection: nearestNote]]].!

isControlActive
	"Should we handle this mouse event?"

	^super isControlActive & sensor blueButtonPressed not! !

!PianoRollController methodsFor: 'menu messages'!

addVoiceToSelection
	"Prompts for a voice number and adds the notes of that voice to the selection."

	| voice empty |
	voice _ (FillInTheBlank request: 'Voice?') asNumber.
	empty _ true.
	model do:
		[: note |
			(note voice = voice)
				ifTrue: [empty _ false. view addToSelection: note]].
	empty
		ifTrue: [self notify: 'No notes in that voice'].!

clearSelection
	"Clear the current selection."

	view clearSelection.!

deleteSelection
	"Delete selected notes after asking for confirmation."

	BinaryChoice
		message: 'Delete all selected notes?' 
		displayAt: Sensor cursorPoint 
		ifTrue:
			[view selected do: [: note | model remove: note].
			 view cacheUpdate.
			 view clearSelection].!

inspectNotes
	"Bring up an inspector on the selected note or notes."

	| selectedNotes |

	selectedNotes _ view selected asSortedCollection.
	selectedNotes isEmpty ifTrue: [^self].				"nothing selected"
	selectedNotes size > 1
		ifTrue: [selectedNotes inspect]					"multiple notes selected"
		ifFalse: [(selectedNotes at: 1) inspect].			"single note selected"!

moveSelectionToVoice
	"Prompts for a voice number and adds all selected notes to that voice."

	| voice |

	voice _
		(FillInTheBlank
			request: 'Move selected notes to which voice?') asNumber.

	voice == 0 ifTrue: [^self].

	view selected do:
		[: note | (note voice: voice)].!

playAll
	"Perform the entire score."

	Cursor execute
		showWhile: [model playFrom: 0 rate: view rate / 100.0].!

playSelectedNotes

	| score |

	score _ Score new.
	score addAll: (view selected asSortedCollection).
	score isEmpty ifTrue: [^self].

	Cursor execute
		showWhile: [score playFrom: (score first time) rate: view rate / 100.0].!

playTimeRange

	| thisNote startIndex stopIndex tempScore |

	1 to: model size do:
		[: index |
			thisNote _ model at: index.
			((startIndex isNil) and: [thisNote time >= view timeRangeStart])
				ifTrue: [startIndex _ index].
			((stopIndex isNil) and: [thisNote time > view timeRangeEnd])
				ifTrue: [stopIndex _ index - 1]].

	(startIndex isNil) ifTrue: [^self].
	(stopIndex isNil) ifTrue: [stopIndex _ model size].

	tempScore _ model copyFrom: startIndex to: stopIndex.
	(tempScore isEmpty)
		 ifTrue: [^self].

	Cursor execute
		showWhile: [tempScore playFrom: (tempScore first time) rate: view rate / 100.0].!

playWindow

	| tempScore endTime endIndex |

	tempScore _ Score new.
	tempScore addAll: (view selected asSortedCollection).
	tempScore isEmpty ifTrue: [tempScore _ model].

	endTime _ view windowEndTime.
	endIndex _ tempScore size.
	1 to: tempScore size do: [: i |
		(((tempScore at: i) time > endTime) and:
		 [endIndex isNil])
			ifTrue: [endIndex _ i]].
		
	Cursor execute showWhile: [
		tempScore _ tempScore copyFrom: 1 to: endIndex.
		tempScore playFrom: (view timeOffset) rate: view rate / 100.0].!

record

	| selectedNotes scoreToPlay startTime newNotes |

	self notify: 'Proceed to record into voice 17'.

	selectedNotes _ view selected.
	(selectedNotes isEmpty)
		ifTrue:
			[scoreToPlay _ model.
			 startTime _ 0]
		ifFalse:
			[scoreToPlay _ (Score new: selectedNotes size) addAll: selectedNotes asSortedCollection; yourself.
			 startTime _ scoreToPlay first time].

	newNotes _ (MidiRecorder new) overDub: scoreToPlay playFrom: startTime rate: view rate / 100.0.
	newNotes do: [: note | note voice: 17. note time: (note time + startTime)].
	model _ model mergedWith: newNotes.

	view model: model.
	view cacheUpdate.
	view displayView.!

setRate

	| newRate |

	newRate _
		(FillInTheBlank
			request: 'Type the rate followed by a CR'
			initialAnswer: view rate printString) asNumber.

	newRate > 0 ifTrue: [view rate: newRate].!

showVoiceCounts

	| tallys |
	tallys _ Bag new.
	model do:
		[: scoreElement |
			(scoreElement isTone)
				ifTrue: [tallys add: scoreElement voice]].
	Transcript cr.
	(tallys sortedElements) do:
		[: assoc |
			Transcript show: 'Voice ', assoc key printString, ': ', assoc value printString; cr].	
	Transcript cr.! !

!PianoRollController methodsFor: 'scrolling'!

canScrollDown

	^view timeOffset > 0!

canScrollUp

	^view timeOffset < model scoreTime!

computeMarkerRegion

	| length |

	length _ (self percentVisible * scrollBar inside height) rounded.
	^0@0 extent: 10@length!

percentVisible

	^(1000.0 / (model scoreTime max: 10) asFloat) min: 100.0!

scrollAmount

	^(super scrollAmount asFloat / (self percentVisible asFloat * 9.5)) truncated!

scrollView
	"Absolute scrolling: the user is thumbing to an absolute location."

	| scorePosition |

	scorePosition _ ((self viewDelta negated asFloat / 1013.0) * model scoreTime asFloat) truncated.
	(model scoreTime = scorePosition) ifFalse: [view scrollTo: scorePosition].
	self moveMarker: self viewDelta.!

scrollView: delta
	"Relative scrolling by the given amount."

	| newTimeOffset | 

	newTimeOffset _ view timeOffset - delta.
	view scrollTo: newTimeOffset.
	self moveMarker: self viewDelta.! !

Object subclass: #MergeSorter
	instanceVariableNames: 'lists currList lastAdded '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-MidiScores'!
MergeSorter comment:
'I collect a list of ordered sub-sequences which can be merged into a single sequenceable collection on demand. Objects may be added to the current sub-sequence using add: or addCheck:. AddCheck: will check the assumption that the sub-sequence is really in ascending order. A new sub-sequence is started with the message startNewSublist.
'!


!MergeSorter methodsFor: 'initialize-release'!

reset
	"Reset to pristine, empty state."

	lists _ OrderedCollection new: 40.
	currList _ nil.
	lastAdded _ nil.
	self startNewSublist.! !

!MergeSorter methodsFor: 'accessing'!

add: anItem
	"Adds anItem to the current list."

	currList addLast: anItem.
	lastAdded _ anItem.!

addCheck: anItem
	"Adds anItem to my current list. Checks the assumption that anItem is >= to the last item added."

	((lastAdded notNil) and: [anItem < lastAdded])
		ifTrue: [self error: 'Attempt to add an item out of sequence'].
	self add: anItem.!

size
	"Answer the total count of all items added to me since the last time I was reset."

	^(lists inject: 0 into: [:subTotal :next | subTotal + next size]) + currList size!

startNewSublist
	"Start a new ordered sublist. Subsequent additions will be appended to this list until the next startNewSublist message is received."

	((currList notNil) and: [currList isEmpty not])
		ifTrue: [lists addLast: currList].
	currList _ Score new: 1000.
	lastAdded _ nil.! !

!MergeSorter methodsFor: 'converting'!

asClass: aCollectionClass
	"Return the merge of all my lists as a sorted sequenceable collection of the given class."

	| out list1 list2 |
	self startNewSublist.		"Adds currList to list (my list of sorted sublists)."
	(lists size == 0) ifTrue: [^aCollectionClass new].	"we're empty so return an empty collection"
	lists _ lists asSortedCollection: [: list1 : list2 | list1 size < list2 size].
	[lists size > 1]
		whileTrue:
			[list1 _ lists removeFirst.
			 list2 _ lists removeFirst.
			 lists add: (self merge: list1 with: list2 as: aCollectionClass)].
	(lists first species = aCollectionClass)
		ifTrue: [^lists first]
		ifFalse:
			[out _ aCollectionClass new: lists first size.
			 out addAll: lists first.
			 ^out].!

asScore
	"Return the merge of all my lists as a sorted Score."

	^self asClass: Score!

merge: list1 with: list2 as: aCollectionClass
	"Answer with an instance of aCollectionClass containing all the elements of list1 and list2 merged into it in the proper order."
	"NOTE: assumes list1 and list2 are already sorted in ascending order."

	| newList pos1 pos2 limit1 limit2 next1 next2 |
	newList _ aCollectionClass new: (list1 size + list2 size).

	(list1 isEmpty) ifTrue: [newList addAll: list2. ^newList].
	(list2 isEmpty) ifTrue: [newList addAll: list1. ^newList].

	pos1 _ 1.
	pos2 _ 1.
	limit1 _ list1 size.
	limit2 _ list2 size.
	next1 _ list1 at: pos1.
	next2 _ list2 at: pos2.

	"Merge list1 and list2 in ascending order until one of them is exhausted."
	[(pos1 <= limit1) & ( pos2 <= limit2)] whileTrue: [
		(next1 <= next2)
			ifTrue: [
				newList addLast: next1.
				pos1 _ pos1 + 1.
				(pos1 <= limit1) ifTrue: [next1 _ list1 at: pos1]]
			ifFalse: [
				newList addLast: next2.
				pos2 _ pos2 + 1.
				(pos2 <= limit2) ifTrue: [next2 _ list2 at: pos2]]
	].

	"Unless the two lists are the same size, one list will be exhausted first. One of the following
	two lines will quickly copy the remaining list into newList."
	(pos1 <= limit1) ifTrue: [newList addAll: (list1 copyFrom: pos1 to: limit1). ^newList].
	(pos2 <= limit2) ifTrue: [newList addAll: (list2 copyFrom: pos2 to: limit2). ^newList].

	^newList! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

MergeSorter class
	instanceVariableNames: ''!


!MergeSorter class methodsFor: 'instance creation'!

new

	^(super new) reset! !

OrderedCollection variableSubclass: #Score
	instanceVariableNames: 'nextIndex activeNotes scoreTime '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-MidiScores'!
Score comment:
'I am a subclass of OrderedCollection used to store collections of notes. I know how to perform the score I contain.

My instance variables are:
	nextIndex			index of the next note to play during a performance
	activeNotes		list of active notes (notes that have been struck but not released)
	scoreTime		a cache for the score duration (scoreTime = nil or the duration)
'!


!Score methodsFor: 'initializing'!

reset
	"Reset the next note pointer (nextIndex) and the active note list to prepare the score to be played from the beginning."

	nextIndex _ 1.
	activeNotes _ SortedCollection sortBlock: [:x :y | x offTime <= y offTime].! !

!Score methodsFor: 'accessing'!

computeScoreTime
	"Compute and answer my total playing time in centiseconds."

	| endTime noteEnd |
	endTime _ 0.
	self do: [: n |
		(n isTone)
			ifTrue:
				[ noteEnd _ n offTime.
				 (noteEnd > endTime) ifTrue: [endTime _ noteEnd]]
			ifFalse:
				[(n time > endTime) ifTrue: [endTime _ n time]]].
	^endTime!

mergedWith: aScore
	"Answer a new score that consists of all my elements merged with all the elements of aScore."

	| sorter |
	sorter _ MergeSorter new.
	self do: [: note | sorter add: note copy].
	sorter startNewSublist.
	aScore do: [: note | sorter add: note copy].
	^sorter asScore!

scoreTime
	"Answer my total playing time in centiseconds. This quantity is cached in my scoreTime instance variable and only recomputed if scoreTime is nil."

	scoreTime isNil
		ifTrue: [scoreTime _ self computeScoreTime].
	^scoreTime! !

!Score methodsFor: 'performance'!

done
	"Answer true if the performance is over."

	^(activeNotes isEmpty) and: [self size < nextIndex]!

play
	"Play from the beginning at normal speed."

	self playFrom: 0 rate: 1.!

playFrom: time rate: aRate
	"Plays the entire score at the given rate. aRate is 1 for normal speed, 2 for twice speed, 3/4 for three-quarter speed, etc."

	| mSecsAtStart currTime |
	"Scan for the first note at or after the given time, processing all control changes and program changes during this scanning to establish the proper synthesizer state."
	self reset; sendEventsThrough: time - 1 silent: true.

	mSecsAtStart _ (Delay millisecondClockValue) - (time * 10 // aRate).
	[self done | Sensor anyButtonPressed] whileFalse:
		[currTime _ 
			((Delay millisecondClockValue - mSecsAtStart) * aRate) // 10.
		self sendEventsThrough: currTime silent: false.
		Processor yield].
	MidiInterface allNotesOff.	"Turn off notes in case I was interrupted."!

testPlay
	"Play myself as fast as possible for testing. All midi events are sent but the clock is advanced as quickly as the processing time will allow."

	| currTime |
	self reset.
	currTime _ 0.
	[self done | Sensor anyButtonPressed]
		whileFalse: [
			self sendEventsThrough: currTime silent: false.
			currTime _ currTime + 10].
	MidiInterface allNotesOff.! !

!Score methodsFor: 'storing'!

maxVoice
	"Answer the number of the highest numbered voice in myself."

	| max |
	max _ 1.
	self do: [: scoreEl |
		((scoreEl isTone) & (scoreEl voice > max))
			ifTrue: [max _ scoreEl voice]].
	^max!

nextToneInVoice: voice
	"Answer the next tone of myself in the given voice at or after nextIndex. Non-Tone score elements are skipped. Set nextIndex to point to the next scoreElement. Answer nil if the end of the score is reached. Details: Uses the instance variable nextIndex to maintain state between invocations. The client should set nextIndex to 1 to start the enumeration."

	| scoreEl |
	[nextIndex <= self size] whileTrue:
		[scoreEl _ self at: nextIndex.
		 nextIndex _ nextIndex + 1.
		 ((scoreEl voice == voice) & (scoreEl isTone))
			ifTrue: [^scoreEl]].
	^nil!

storeAdagioOn: aStream
	"Write an Adagio representation of my Tones to the given stream."

	| maxVoice |
	aStream nextPutAll: '* Created on ', Time dateAndTimeNow printString; cr; cr.
	maxVoice _ self maxVoice.
	1 to: self maxVoice do:
		[: voice |
			aStream nextPutAll: '* Voice ', voice printString; cr.
			self storeVoice: voice on: aStream.
			(voice ~~ maxVoice)
				ifTrue: [aStream cr; cr]].!

storeAdagioOnFile: aFileName
	"Write an Adagio representation of my Tones to a file with the given name. Note: Only the Tones of the score are recorded. This could be extended to store control changes or other kinds of score elements but so far that has not been needed."

	| aStream |
	aStream _ FileStream newFileNamed: aFileName, '.gio'.
	aStream nextPutAll: '* ', aFileName; cr.
	self storeAdagioOn: aStream.
	aStream close.!

storeVoice: voice on: aStream
	"Write an Adagio representation of the Tones of the given voice to the given stream."

	| oldNextIndex previousTone thisTone nextTone |
	"save nextIndex (used for enumeration) and start voice enumeration"
	oldNextIndex _ nextIndex.	
	nextIndex _ 1.

	"get first three tones"
	previousTone _ self nextToneInVoice: voice.
	thisTone _ self nextToneInVoice: voice.
	nextTone _ self nextToneInVoice: voice.

	"store first tone, if there is one (the voice could be empty)"
	(previousTone notNil)
		ifTrue:
			[previousTone
				storeAdagioOn: aStream
				previous: nil next: thisTone].

	"store the rest"
	[thisTone notNil]
		whileTrue:
			[thisTone
				storeAdagioOn: aStream
				previous: previousTone next: nextTone.
			 previousTone _ thisTone.
			 thisTone _ nextTone.
			 nextTone _ self nextToneInVoice: voice].

	nextIndex _ oldNextIndex.		"restore nextIndex"! !

!Score methodsFor: 'editing'!

edit
	"Open a piano-roll editor on myself."

	PianoRollView openOn: self.! !

!Score methodsFor: 'private'!

sendEventsThrough: time silent: silentFlag
	"Processes the score up through the given time. If silentFlag is true, then sounding notes are skipped but all control and program changes are performed."

	| event |
	self turnOffNotesAt: time.
	[(nextIndex <= self size) and:
	  [(event _ self at: nextIndex) time <= time]]
		whileTrue:
			[(event isTone)
				ifTrue:
					[silentFlag ifFalse:
						[event perform.
						 activeNotes add: event]]
				ifFalse: [event perform].
			 nextIndex _ nextIndex + 1].!

setIndices
	"Override OrderedCollection's method for this initialization method since we anticipate adding things at the end. (We most frequently use addLast: rather than addFirst: to add to Score objects.)"

	firstIndex _ 1.
	lastIndex _ 0.!

turnOffNotesAt: time
	"Turns off all notes in the active list with times at or before the given time."

	| note |
	[activeNotes isEmpty ifTrue: [^self].
	 note _ activeNotes first.
	 ((note offTime) <= time)]
		whileTrue:
			[MidiInterface noteOff: (note pitch) chan: (note voice).
			 activeNotes removeFirst].! !

!Score methodsFor: 'converting'!

asEventList
	"return an eventList version of the receiver score"

	| list |
	list _ EventList newNamed: EventList nextName.
	self do: [ :event |
		list add: (NoteEvent dur: (event dur * 10) 
				pitch: (event pitch) 
				ampl: (event velocity) 
				voice: (event voice))
			at: event time * 10].
	^list! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Score class
	instanceVariableNames: ''!


!Score class methodsFor: 'instance creation'!

new
	"Answer with a new, empty score."

	^(super new) reset!

new: initialSize
	"Answer with a new, empty score with room for initialSize elements before growth is needed."

	^(super new: initialSize) reset! !

!Score class methodsFor: 'examples'!

example
	"Score example"

	| score startTime currTime |
	score _ Score new.
	score add: (NoteElement new: 60 at: 0 dur: 40).
	score add: (NoteElement new: 62 at: 50 dur: 40).
	score add: (NoteElement new: 64 at: 100 dur: 40).
	score add: (NoteElement new: 65 at: 150 dur: 40).
	score add: (NoteElement new: 67 at: 200 dur: 25).
	score add: (NoteElement new: 65 at: 225 dur: 25).
	score add: (NoteElement new: 64 at: 250 dur: 25).
	score add: (NoteElement new: 62 at: 275 dur: 25).
	score add: (NoteElement new: 60 at: 300 dur: 100).
	^score
"
	startTime _ Time millisecondClockValue.
	[score done] whileFalse:
		[currTime _ (Time millisecondClockValue - startTime) // 10.
		 score sendEventsThrough: currTime silent: false].
"! !

Object subclass: #AdagioParser
	instanceVariableNames: 'typeTable source buffer hereChar tokenType notes tempo rate synchPoint time duration nextTimeDelta pitchClass modifier octave loudness voice timbre restFlag hadAttributeFlag '
	classVariableNames: 'AdagioTypeTable DurationsTable DynamicsTable PitchTable StopChar '
	poolDictionaries: ''
	category: 'Music-MidiScores'!
AdagioParser comment:
'I am a parser for the Adagio score representation language as described in the CMU Midi Toolkit handbook. I am  invoked by "AdagioParser parse: stream" where the stream usually comes from a string or file. I operate according to a table of token types, TypeTable. (See my class initialization message.) I work by collecting note attributes and flags until a delimiter is reached, at which point I output a note and update the current time according to the current set of attributes. Delimiters are comma, semi-colon, and the end of a line.'!


!AdagioParser methodsFor: 'initialize-release'!

initAdagioDefaults
	"Initialize all Adagio state variables to their default values."

	tempo _ 100.0.
	rate _ 100.0.
	synchPoint _ 0.0.
	time _ 0.0.
	duration _ DurationsTable at: $q.
	nextTimeDelta _ nil.
	pitchClass _ PitchTable at: $c.
	modifier _ 0.
	octave _ 4.
	loudness _ 127.
	voice _ 1.
	timbre _ 1.
	restFlag _ false.
	hadAttributeFlag _ false.!

parse: aStream
	"Parse the Adagio score in aStream and answer the resulting Score object."

	typeTable _ AdagioTypeTable.
	source _ aStream.
	buffer _ BufferStream on: (String new: 40).
	self nextChar.	"prime the pump"
	notes _ MergeSorter new.
	self initAdagioDefaults.
	[self doNext] whileFalse: [].	"parse until done"
	^notes asScore! !

!AdagioParser methodsFor: 'attribute parsing'!

parseDur: aStream
	"Parse the symbolic duration contained in aStream (e.g. 'q.t' is a dotted quarter triplet). Answer the duration in 100ths of a second if this succeeds, answer nil if it fails."

	| dur ch multiplier |

	dur _ DurationsTable
			at: aStream next
			ifAbsent:
				[self badAttribute: aStream contents.
				^nil].

	[(aStream atEnd not) and: [aStream peek isDigit not]]
		whileTrue: [
			ch _ aStream next.
			(ch == $.) ifTrue: [dur _ dur * 1.5].
			(ch == $t) ifTrue: [dur _ dur * 0.66666666].
			((ch == $t) | (ch == $.)) ifFalse:
				[self badAttribute: aStream contents.
				 ^nil]].		"bad attribute; abort"

	(aStream atEnd)
		ifTrue: [^dur]
		ifFalse:
			[multiplier _ self convertToNumber: aStream.
			 multiplier isNil ifTrue: [^nil].	"bad number; abort"
			 ^dur * multiplier].!

parseRate
	"Parse an Adagio rate command (e.g. '!!rate 120')."

	| newRate |
	self skipWhiteSpace.
	self scanAttribute.
	newRate _ self convertToNumber: buffer.
	newRate isNil
		ifTrue: [
			"bad number; abort"
			^self error: 'bad argument to !!rate', self token].
	rate _ newRate asFloat.
	synchPoint _ time.
	notes startNewSublist.
	self specialCheck.!

parseTempo
	"Parses an Adagio tempo command (e.g. '!!tempo 120')."

	| newTempo |
	self skipWhiteSpace.
	self scanAttribute.
	newTempo _ self convertToNumber: buffer.
	newTempo isNil
		ifTrue: [
			"bad number; abort"
			^self error: 'bad argument to !!tempo', self token].
	tempo _ newTempo asFloat.
	synchPoint _ time.
	notes startNewSublist.
	self specialCheck.!

processPitchModifier: ch
	"Process a pitch modifer character. Answer nil if it is bad."

	(ch == $s)
		ifTrue:
			["sharp"
			 modifier notNil ifTrue: [^nil].
			 ^modifier _ 1].
	(ch == $f)
		ifTrue:
			["flat"
			 modifier notNil ifTrue: [^nil].
			 ^modifier _ -1].
	(ch == $n)
		ifTrue:
			["natural"
			 modifier notNil ifTrue: [^nil].
			 ^modifier _ 0].
	(ch isDigit)
		ifTrue:
			[octave notNil ifTrue: [^nil].
			 ^octave _ ch digitValue].

	"if we get here, we have encountered a bad pitch modifer"
	^nil!

specialCheck
	"Special commands cannot be combined with other attributes. This method is to be called after parsing the arguments of a special command to be sure that no other attributes occur before the next separator."

	[(self peekAhead == #separator) or:
	 [self peekAhead == #end]]
		whileFalse:
			[self skipWhiteSpace.
			 self scanAttribute.
			 self error: 'Attributes not allowed with special commands', self token].!

xAbsDuration
	"Parses absolute durations (e.g. 'u60' is a note lasting 60 100ths of a second)."

	| dur |
	self scanAttribute.
	buffer next.
	(buffer atEnd) ifTrue: [^self badAttribute].
	dur _ self convertToNumber: buffer.
	dur isNil
		ifTrue: [^self].	"bad number; abort"
	duration _ dur asFloat * (tempo / 100.0).		"scale to current tempo."
	hadAttributeFlag _ true.!

xAbsPitch
	"Parses absolute pitches (e.g. 'p60' is middle C)."

	| p |
	self scanAttribute.
	buffer next.
	(buffer atEnd) ifTrue: [^self badAttribute].
	p _ self convertToNumber: buffer.
	(p isNil) ifTrue: [^self].	"bad number; abort"
	(p < 1) ifTrue:
		[self range: 'pitch' value: p using: 1.
		 p _ 1].
	(p > 127) ifTrue:
		[self range: 'pitch' value: p using: 127.
		 p _ 127].
	pitchClass _ p rem: 12.
	modifier _ 0.
	octave _ p // 12.
	hadAttributeFlag _ true.!

xComment
	"A comment starts with a '*' character and runs through the end of the line."

	self skipRestOfLine.!

xDuration
	"Parses symbolic durations (e.g. 'q.t' is a dotted quarter triplet)."

	| dur |
	self scanAttribute.
	dur _ self parseDur: buffer.
	dur isNil
		ifTrue: [^self].		"bad attribute; abort"
	duration _ dur.
	hadAttributeFlag _ true.!

xLoudness
	"Parses loudness attributes, either absolute or symbolic (e.g. 'L58' = 'Lmf', both of which  mean to play with a loudness of mezzo forte)."

	| stream loud |
	self scanAttribute.
	buffer next.
	(buffer atEnd) ifTrue: [^self badAttribute].
	(buffer peek isDigit)
		ifTrue:
			[loud _ self convertToNumber: buffer.
			 loud isNil ifTrue: [^self].	"bad number; abort"
			 (loud > 127)
				ifTrue:
					[self range: 'loudness' value: loud using: 127.
					 loud _ 127].
			 (loud < 1)
				ifTrue:
					[self range: 'loudness' value: loud using: 1.
					 loud _ 1]]
		ifFalse:
			[loud _ DynamicsTable
						at: buffer throughEnd asSymbol
						ifAbsent: [^self badAttribute]].
	loudness _ loud.
	hadAttributeFlag _ true.!

xNextTime
	"Parses a next time attribute. The next time value may be a positive number or it may be a symbolic duration. For example, 'n60' = 'nq' at a tempo of 100; both cause the next note to be played a quarter note amount of time after the current note, or 60/100ths of a second."

	| n |
	self scanAttribute.
	buffer next.
	(buffer atEnd) ifTrue: [^self badAttribute].
	(buffer peek isDigit)
		ifTrue:
			[n _ (self convertToNumber: buffer) * (100.0 / rate).
			 n isNil ifTrue: [^self]]	"bad number; abort"
		ifFalse:
			[n _ (self parseDur: buffer) * (100.0 / tempo) * (100.0 / rate).
			 n isNil ifTrue: [^self]].	"bad duration; abort"
	nextTimeDelta _ n.
	hadAttributeFlag _ true.!

xPitch
	"Parses a pitch attribute starting with a pitch name. See xAbsPitch for absolute pitch parsing."

	| oldPitchClass oldModifier oldOctave oldPClass newPClass pitchDif |
	self scanAttribute.
	oldPitchClass _ pitchClass.
	oldModifier _ modifier.
	oldOctave _ octave.
	pitchClass _ PitchTable at: (buffer next) ifAbsent: [^self badAttribute].
	modifier _ octave _ nil.
	[buffer atEnd]
		whileFalse:
			["process pitch modifiers: octave, sharp, flat and natural"
			 (self processPitchModifier: (buffer next)) isNil
				ifTrue:
					["nil indicates a bad pitch modifier"
					 pitchClass _ oldPitchClass.
					 modifier _ oldModifier.
					 octave _ oldOctave.
					 ^self badAttribute]].
	hadAttributeFlag _ true.
	modifier isNil ifTrue: [modifier _ 0].	"no modifier specified"
	(octave isNil)
		ifTrue:
			["if no octave was specified, use the one which minimizes the leap from the previous note"
			 oldPClass _ (oldPitchClass + oldModifier) \\ 12.
			 newPClass _ (pitchClass + modifier) \\ 12.
			 pitchDif _ newPClass - oldPClass.	"compare with last pitch"
			 (pitchDif > 5)
				ifTrue: [^octave _ oldOctave - 1 max: 1].
			 (pitchDif < -6)
				ifTrue: [^octave _ oldOctave + 1 min: 9].
			octave _ oldOctave].!

xProgram
	"Parses a program number attribute (e.g. 'z48' sets the synthesizer to timber 48)."

	| prog |
	self scanAttribute.
	buffer next.
	(buffer atEnd) ifTrue: [^self badAttribute].
	prog _ self convertToNumber: buffer.
	(prog isNil)
		ifTrue: [^self].	"bad number; abort"
	(prog < 1)
		ifTrue:
			[self range: 'program number' value: prog using: 1.
			 prog _ 1].
	(prog > 127)
		ifTrue:
			[self range: 'program number' value: prog using: 127.
			 prog _ 127].
	timbre _ prog.
	hadAttributeFlag _ true.!

xRest
	"Sets a flag indicating that this was a rest. This flag suppresses generation of a note, although other attributes such as voice, timbre, and time are updated."

	| restCmd |
	self scanAttribute.
	restCmd _ self token.
	((restCmd = 'r') or: [restCmd = 'R'])
		ifFalse: [^self badAttribute].
	hadAttributeFlag _ restFlag _ true.!

xSpecial
	"Parses an Adagio special command. Special commands start with an exclaimation point (!!) and may not be combined on a line with other attributes."

	| cmd |
	self scanAttribute.
	cmd _ self token.
	(cmd = '!!tempo') ifTrue: [^self parseTempo].
	(cmd = '!!rate') ifTrue: [^self parseRate].
	(cmd = '!!endscore')
		ifTrue:
			[hadAttributeFlag _ false.
			 tokenType _ #end.
			 ^source setToEnd].
	"unknown special command"
	self error: 'bad special command', cmd.
	self specialCheck.!

xTime
	"Parses a time attribute. The time may be a positive number or it may be a symbolic duration. For example, 't60' = 'tq' at a tempo of 100; both cause the time to be set to one quarter note, or 60/100ths of a second, after the last synch point."

	| t |
	self scanAttribute.
	buffer next.
	(buffer atEnd) ifTrue: [^self badAttribute].
	(buffer peek isDigit)
		ifTrue:
			[t _ (self convertToNumber: buffer) * (100.0 / rate).
			 t isNil ifTrue: [^self]]	"bad number; abort"
		ifFalse:
			[t _ (self parseDur: buffer) * (100.0 / tempo) * (100.0 / rate).
			 t isNil ifTrue: [^self]].	"bad duration; abort"
	time _ synchPoint + t.
	notes startNewSublist.
	hadAttributeFlag _ true.!

xVoice
	"Parses a voice attribute (e.g. 'v3' sets the current voice to 3)."

	| v |
	self scanAttribute.
	buffer next.
	(buffer atEnd) ifTrue: [^self badAttribute].
	v _ self convertToNumber: buffer.
	(v isNil)
		ifTrue: [^self].	"bad number; abort"
	(v < 1) ifTrue:
			[self range: 'voice' value: v using: 1.
			 v _ 1].
	(v > 16) ifTrue:
			[self range: 'voice' value: v using: 16.
			 v _ 16].
	voice _ v.
	hadAttributeFlag _ true.! !

!AdagioParser methodsFor: 'scanning'!

nextChar
	"Read the next character of the source into hereChar. hereChar is set to EndChar when the source is exhausted."

	(hereChar _ source next) isNil
		ifTrue: [hereChar _ StopChar].!

nextToken
	"Read the next token from source, which is determined by looking at the first non-white space character. Set tokenType to indicate the type of the token just read."

	self skipWhiteSpace.
	tokenType _ typeTable at: hereChar asciiValue.
	"if x is the first letter of tokenType:"
	(tokenType at: 1) == $x
		ifTrue:
			["then perform it to scan a note attribute token"
			 self perform: tokenType]
		ifFalse:
			["else this tokenType needs no further processing"
			 self nextChar].!

peekAhead
	"Skip white space and peek at the type of the next character type."

	self skipWhiteSpace.
	^typeTable at: hereChar asciiValue!

scanAttribute
	"Read a string of contiguous characters not containing spaces or delimiters."

	| cType |
	buffer resetAll.
	[cType _ typeTable at: hereChar asciiValue.
	 (cType == #spacer) or:
	  [(cType == #separator) or:
	   [(cType == #semicolon) or:
	    [(cType == #comma) or:
	     [cType == #end]]]]]
			whileFalse:
				[buffer nextPut: hereChar asLowercase.
				 self nextChar].
	buffer reset.
	tokenType _ #attribute.!

skipRestOfLine
	"Skip the remaining characters of this line."

	[((typeTable at: hereChar asciiValue) == #separator) or:
	 [hereChar == StopChar]]
		whileFalse: [self nextChar].
	self nextChar.!

skipWhiteSpace
	"Skip white space."

	[(typeTable at: hereChar asciiValue) == #spacer]
		whileTrue: [self nextChar].! !

!AdagioParser methodsFor: 'private'!

badAttribute

	self badAttribute: self token.!

badAttribute: attributeString

	self report: 'Bad Adagio attribute: ', attributeString.!

convertToNumber: aStream
	"Converts aStream to a number. Complain and answer nil if it contains non-digits."
	"Details: If aStream contains non-digits, 'Integer readFrom:' will not consume the entire stream."

	| val |
	val _ Integer readFrom: aStream radix: 10.
	aStream atEnd
		ifFalse:
			["error parsing the contents of the stream; didn't read it all"
			 self report: 'Only digits are expected here: ', self token.
			 ^nil].	"return nil to indicate an error even if the user proceeds"
	(val < 0)
		ifTrue:
			["only positive integers are allowed"
			 self report: 'Only positive numbers are allowed here: ', self token.
			 ^nil].	"return nil to indicate an error even if the user proceeds"
	^val!

curDuration
	"A tempo of 100 means 100 quarter notes per minute, or 60 hundredths of a second per quarter note. A duration must be scaled by both the tempo and the rate. A rate of  '100' is normal, '200' is twice normal, etc. If the rate is doubled, the duration of each note must be cut in half."

	^duration * (100.0 / tempo) * (100.0 / rate)!

curPitch

	^(octave + 1) * 12 + pitchClass + modifier!

doNext
	"Scan the next token from the input and take appropriate action. Answer true when input is exhausted."

	self nextToken.	"sets tokenType"
	(tokenType == #attribute) ifTrue:
		[^false].
	(tokenType == #separator) ifTrue:
		["Generate an event and advance time."
		 self generateEvent: true.
		 ^false].
	(tokenType == #semicolon) ifTrue:
		["Generate an event and advance time."
		 self generateEvent: true.
		 ^false].
	(tokenType == #comma) ifTrue:
		["Generate an event but do not advance time (comma separator)."
		 self generateEvent: false.
		 ^false].
	(tokenType == #xComment) ifTrue:
		[^false].
	(tokenType == #end) ifTrue:
		[self generateEvent: false.	"Generate the last event. Don't bother advancing time."
		 ^true].
	"if we get here, we have encountered an unexpected character"
	self report: 'Unexpected character ', hereChar asciiValue printString.
	self skipRestOfLine.!

generateEvent: advanceFlag

	| note |
	hadAttributeFlag ifTrue:
		[restFlag ifFalse:
			[note _ NoteElement
				new: (self curPitch)
				at: (time truncated)
				dur: (self curDuration truncated)
				vel: loudness
				voice: voice.
			 notes add: note].
		 (nextTimeDelta notNil)
			ifTrue:
				["always advance time if a nextTimeDelta attribute was given"
				time _ time + nextTimeDelta.
				nextTimeDelta _ nil]
			ifFalse:
				["advance time even if nextTimeDelta wasn't given, using the duration of the current note"
				 advanceFlag ifTrue: [time _ time + self curDuration]]].

	hadAttributeFlag _ restFlag _ false.!

range: attributeName value: offendingValue using: substitutedValue
	"Post a range error and correction notification."

	self report:
		offendingValue printString, ' is out of range for the ',
		attributeName, ' attribute; substituting ', substitutedValue printString.!

report: aString
	"Report an error."

	self error: aString.!

token

	^buffer contents! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

AdagioParser class
	instanceVariableNames: ''!


!AdagioParser class methodsFor: 'class initialization'!

initialize
	"This table drives the scanner. The first letter of each attribute is looked up here to decide what to do with the rest of the attribute. In the case of spaces and punctuation, no further scanning is required."
	"AdagioParser initialize."

	| newTable |
	StopChar _ 255 asCharacter.
	newTable _ Array new: 256 withAll: #undefined.	"default entry"
	newTable atAll: #(9 32) put: #spacer.				"tab space"
	newTable atAll: #(10 12 13) put: #separator.		"lf ff cr"
	newTable at: StopChar asciiValue put: #end.		"StopChar"
	newTable at: $; asciiValue put: #semicolon.
	newTable at: $, asciiValue put: #comma.
	newTable atAll: ($0 asciiValue to: $9 asciiValue) put: #digit.
	newTable at: $* asciiValue put: #xComment.
	newTable at: $!! asciiValue put: #xSpecial.
	newTable
		atAll: (#($a $b $c $d $e $f $g $A $B $C $D $E $F $G) collect: [: letter | letter asciiValue])
 		put: #xPitch.
	newTable
		atAll: (#($w $h $q $i $s $W $H $Q $I $S) collect: [: letter | letter asciiValue])
 		put: #xDuration.
	newTable atAll: (#($l $L) collect: [: letter | letter asciiValue]) put: #xLoudness.
	newTable atAll: (#($n $N) collect: [: letter | letter asciiValue]) put: #xNextTime.
	newTable atAll: (#($p $P) collect: [: letter | letter asciiValue]) put: #xAbsPitch.
	newTable atAll: (#($r $R) collect: [: letter | letter asciiValue]) put: #xRest.
	newTable atAll: (#($t $T) collect: [: letter | letter asciiValue]) put: #xTime.
	newTable atAll: (#($u $U) collect: [: letter | letter asciiValue]) put: #xAbsDuration.
	newTable atAll: (#($v $V) collect: [: letter | letter asciiValue]) put: #xVoice.
	newTable atAll: (#($z $Z) collect: [: letter | letter asciiValue]) put: #xProgram.
	AdagioTypeTable _ newTable.
	self initializeOtherTables.!

initializeOtherTables

	PitchTable _ Dictionary new.
	PitchTable at: $c put: 0.
	PitchTable at: $d put: 2.
	PitchTable at: $e put: 4.
	PitchTable at: $f put: 5.
	PitchTable at: $g put: 7.
	PitchTable at: $a put: 9.
	PitchTable at: $b put: 11.

	DynamicsTable _ Dictionary new.
	DynamicsTable at: #ppp put: 20.
	DynamicsTable at: #pp put: 26.
	DynamicsTable at: #p put: 34.
	DynamicsTable at: #mp put: 44.
	DynamicsTable at: #mf put: 58.
	DynamicsTable at: #f put: 75.
	DynamicsTable at: #ff put: 98.
	DynamicsTable at: #fff put: 127.

	DurationsTable _ Dictionary new.
	DurationsTable at: $w put: 240.0.
	DurationsTable at: $h put: 120.0.
	DurationsTable at: $q put: 60.0.
	DurationsTable at: $i put: 30.0.
	DurationsTable at: $s put: 15.0.! !

!AdagioParser class methodsFor: 'parsing'!

parse: aStream

	^(AdagioParser new) parse: aStream! !

AdagioParser initialize!


Magnitude subclass: #ScoreElement
	instanceVariableNames: 'time '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-MidiScores'!
ScoreElement comment:
'ScoreElement is an abstract superclass for things which an be placed in scores, such as Notes, ProgramChanges, ControlChanges, and so forth. The salient feature of all ScoreElements is that they occur at some time and can be sorted into order accordingly. They also obey the Link protocol, so they may be stored in LinkedLists.'!


!ScoreElement methodsFor: 'accessing'!

time
	"Answer the performance time of this ScoreElement in centiseconds."

	^time!

time: newTime
	"Set the performance time of this ScoreElement to the given time."

	time _ newTime.! !

!ScoreElement methodsFor: 'testing'!

isNote

	^false!

isTone

	^false! !

!ScoreElement methodsFor: 'comparing'!

< aScoreElement
	"Answer true if I am at an earlier time than aScoreElement."

	^self time < aScoreElement time!

<= aScoreElement
	"Answer true if I am at an earlier or the same time as aScoreElement."

	^self time <= aScoreElement time!

= aScoreElement
	"Answer true if I am at the same time as aScoreElement."

	^self time = aScoreElement time!

> aScoreElement
	"Answer true if I am at a later time than aScoreElement."

	^self time > aScoreElement time!

>= aScoreElement
	"Answer true if I am at a later or the same time as aScoreElement."

	^self time >= aScoreElement time!

hash
	"Answer the hash of my time."

	^self time hash!

~= aScoreElement
	"Answer true if I am not at the same time as aScoreElement."

	^self time ~= aScoreElement time! !

!ScoreElement methodsFor: 'performing'!

perform
	"Perform myself. This is a placeholder that subclasses override to actually do something."! !

ReadWriteStream subclass: #BufferStream
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-MidiScores'!
BufferStream comment:
'A BufferStream is a re-usable ReadWriteStream. The message resetAll will reset both the readLimit and the current position (making it empty). Used by the Adagio parser.'!


!BufferStream methodsFor: 'special'!

buffer
	"Answer my buffer (collection) to allow direct access for high performace I/O. Experts only!!!!"

	^collection!

resetAll
	"Resets both the read and write positions so that the BufferStream becomes empty and may be re-used."

	readLimit _ position _ 0.!

setReadLimit: newLimit
	"Sets the readLimit to newLimit. No sanity check is performed, so be sure you know what you are doing!!"

	readLimit _ newLimit!

throughEnd
	"Answer a sub-collection containing all the unread characters in the buffer."

	^collection copyFrom: position + 1 to: readLimit! !

Object subclass: #MidiInterface
	instanceVariableNames: ''
	classVariableNames: 'Debug InBuf PitchBends Port PrimOn ProgramNumbers '
	poolDictionaries: ''
	category: 'Music-MidiScores'!
MidiInterface comment:
'See my class comment.'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

MidiInterface class
	instanceVariableNames: ''!
MidiInterface class comment:
'MidiInterface - Musical Instrument Digital Interface

This class allows Smalltalk programs to control synthesizers and other musical devices attached via the RS232 port and a midi interface box. Because the Midi port(s) are a shared resource, there are no instances of this class. Instead, Midi control messages are directed to the class itself. You can use the "class refs" facility to find all the users of Midi operations in the system.
-----------------------------------------------------------------------
Notes on the Smalltalk Midi Primitive

1. Be sure to have a Midi interface attached the port you are using. The interface should have a clock speed of 1 Mhz (most do, including the one from Apple). It is possible to hang the system if you open a port that does not have a Midi interface attached (and turned on) because the port depends on a clock signal supplied by the Midi interface -- no clock, no action!!!!

2. The Midi driver is interrupt driven and the interrupt code is part of the Smalltalk interpreter. Thus, if you quit from Smalltalk with one of the Midi ports open and an interrupt occurs (due to a character arriving, say) then your Mac will crash because the drivers aren''''t there anymore. Included in this filein is a modification of Smalltalk''''s quit method that closes all Midi ports. (Closing a non-open port is harmless.) It is recommended that you include a similar mechanism in any applications you develop yourself. You will have to re-open the Midi port(s) you were using after restarting the image.

3. There is a limit of 1000 bytes per call to the Midi put primitive; you can pass it a bigger buffer but the extra bytes will not be sent.

4. The Midi put primitive puts its bytes in an internal buffer until all previous bytes have been sent. If there is not enough room in the buffer, the put primitive waits until there is room. If there IS room in in the buffer, the bytes of the current call are buffered and the call returns to Smalltalk immediately even though not all the bytes have actually been sent.

5. If the transmit buffer is completely full, it takes nearly three seconds to empty itself. (Midi data is sent at 3125 bytes/second, so you have to send a lot of data fast to actually fill up the 8192 byte transmit buffer.) If you close the Midi port before the buffer is empty, some bytes will never be sent. My practice is to keep the Midi port open continuously during a Smalltalk session but if your application opens and closes the port you may wish to put a three second delay before each close operation to absolutle guarantee that all the bytes have been sent.
'!


!MidiInterface class methodsFor: 'initialize-debug'!

debugOff
	"MidiInterface debugOff"

	Debug _ false.!

debugOn
	"When the debugging switch is turned on, all midi output will be printed to the Transcript rather than actually sent. Debugging also enables range checks on the parameters of the outgoing Midi messages."
	"MidiInterface debugOn"

	Debug _ true.!

initialize
	"MidiInterface initialize"

	Debug _ false.
	PrimOn  _ false.
	InBuf _ BufferStream on: (ByteArray new: 100).
	Port _ nil.
"	self primClose: 0.
	self primClose: 1.
"	Smalltalk removeDependent: self.!

primOff
	"MidiInterface PrimOff"

	PrimOn  _ false.!

primOn
	"MidiInterface PrimOn"

	PrimOn  _ true.! !

!MidiInterface class methodsFor: 'port control'!

closePort
	"Close the midi port."
	"MidiInterface closePort"

	Smalltalk removeDependent: self.
	(Port isNil) ifFalse: [self primClose: Port].
	Port _ nil.!

closePortPrim
	"Close the midi port."
	"MidiInterface closePort"

	Smalltalk removeDependent: self.
	(Port isNil) ifFalse: [self primClose: Port].
	Port _ nil.!

openPort
	"Open the RS-232C port unless debugging is turned on."
	"MidiInterface openPort."

	(Debug not & Port isNil)
		ifTrue:
			[Port _ RS232Port open: 1.
			 Port isNil ifFalse:
					[Port setRawMode; setMidiMode]]!

openPort: portNumber
	"Open the given midi port. In case the port is currently open, close it first."
	"MidiInterface openPort: 0"

	self closePort.
	Port _ portNumber.
	Smalltalk addDependent: self.
	self primOpen: portNumber.!

openPortPrim: portNumber
	"Open the given midi port. In case the port is currently open, close it first."
	"MidiInterface openPort"

	self closePort.
	Port _ portNumber.
	Smalltalk addDependent: self.
	self primOpen: portNumber.!

openPortStandard
	"Open the RS-232C port unless debugging is turned on."
	"MidiInterface openPortStandard."

	(Debug not & Port isNil)
		ifTrue:
			[Port _ RS232Port open: 1.
			 Port setRawMode; setMidiMode].
"
	SendBuf _ BufferStream on: (String new: 100)."!

update: parameter
	"Close the port if the user quits Smalltalk to avoid hanging the system."

	((parameter == #aboutToSnapshotAndQuit) | (parameter == #aboutToQuit))
		ifTrue: [self closePort].! !

!MidiInterface class methodsFor: 'midi output'!

allNotesOff
	"Turns off all midi notes on all midi channels."

	1 to: 16 do: [: ch | self control: 123 value: 0 chan: ch].!

control: control value: value chan: channel
	"Sets the given control to the given new value."

	self send: 176 with: control with: value on: channel.!

noteOff: key chan: channel
	"Turns off the given note."
	"Detail: Send a 'noteOn' command with zero velocity since not all synthesizers recognize the 'noteOff' command."

	self send: 144 with: key with: 0 on: channel.!

noteOn: key vel: velocity chan: channel
	"Turns on the given note."

	self send: 144 with: key with: velocity on: channel.!

pitchBend: bend chan: channel
	"Updates the pitch bend on this channel to the given new value."

	self
		send: 224
		with: (bend bitAnd: 127)
		with: (bend bitShift: -7)
		on: channel.!

program: program chan: channel
	"Changes the synthesizer's 'program number' (or 'instrument' or 'preset') for this channel to the given new value."

	self send: 192 with: (program - 1) on: channel.!

synthInit
	"Resets the synthesizer. This may need to be changed since devices differ; the idea is to put the synthesizer into a known state for all parameters which might be changed through Midi."

	1 to: 16 do:
		[: channel |
			self control: 123 value: 0 chan: channel.	"All notes off on this channel."
			self pitchBend: 8192 chan: channel.
			self program: 1 chan: channel.
			"You may wish to reset various other control values here."].! !

!MidiInterface class methodsFor: 'midi monitor'!

monitor
	"This is a simple midi monitor program which deciphers incoming
	midi commands and prints them to the transcript until a mouse
	button is pressed."
	"MidiInterface monitor"

	self monitorAllBut: #().!

monitorAllBut: cmdList
	"Print midi messages to the transcript until any mouse button is pressed.
	Ignore the given collection of command bytes."

	| recorder |
	recorder _ MidiRecorder new.
	cmdList do: [: cmd | recorder ignoreCmd: cmd].	"ignore these commands"
	recorder monitor.!

monitorRaw
	"Shows the raw midi data as decimal bytes as it arrives. Suppresses printing of 'active status' bytes."
	"MidiInterface monitorRaw"

	| c |
	self flushInput.
	[Sensor anyButtonPressed] whileFalse:
		[c _ self getByte.
		 ((c ~= #NoInput) & (c ~= 254))
			ifTrue: [Transcript show: c printString, ' ']].! !

!MidiInterface class methodsFor: 'low level - output'!

send: midiCode with: arg on: channel
	"Sends a one-argument midi channel message. Checks midiCode, 
	argument, and channel number if Debug is true."

	| cmd |
	cmd _ midiCode bitOr: (channel - 1).
	Debug
		ifTrue:
			[((midiCode >= 128) & (midiCode <= 255))
				ifFalse: [self error: 'bad Midi command'].
			 ((channel >= 1) & (channel <= 16))
				ifFalse: [self error: 'bad channel number'].
			 ((arg >= 0) & (arg <= 127))
				ifFalse: [self error: 'bad Midi argument'].
			 Transcript show:
				cmd printString, ' ',
				arg printString; cr]
		ifFalse:
			[(Port isNil) ifTrue: [^self error: 'Midi port not open'].
			PrimOn
				ifTrue: [self
						primPut: Port
						bytes: (ByteArray with: cmd with: arg)
						count: 2]
				ifFalse: [Port sendBuffer: (ByteArray with: cmd with: arg)]]!

send: midiCode with: arg1 with: arg2 on: channel 
	"Sends a two-argument midi channel message. Checks midiCode, 
	arguments, and channel number if Debug is true."

	| cmd |
	cmd _ midiCode bitOr: channel - 1.
	Debug
		ifTrue: 
			[midiCode >= 128 & (midiCode <= 255) 
						ifFalse: [self error: 'bad Midi command'].
			channel >= 1 & (channel <= 16) 
						ifFalse: [self error: 'bad channel number'].
			arg1 >= 0 & (arg1 <= 127) 
						ifFalse: [self error: 'bad Midi argument'].
			arg2 >= 0 & (arg2 <= 127) 
						ifFalse: [self error: 'bad Midi argument'].
			Transcript show: cmd printString , ' ' , 
						arg1 printString , ' ' , 
						arg2 printString; cr]
		ifFalse: 
			[Port isNil ifTrue: [^self error: 'Midi port not open'].
			PrimOn
				ifTrue: [self
						primPut: Port
						bytes: (ByteArray with: cmd with: arg1 with: arg2)
						count: 3]
				ifFalse: [Port sendBuffer: (ByteArray with: cmd with: arg1 with: arg2)]]! !

!MidiInterface class methodsFor: 'low level - input'!

flushInput
	"Empty the midi port's input buffer."

	self getBuffer.
	[InBuf atEnd] whileFalse:
		["suck in buffers until there is no more data in the queue"
		 self getBuffer].!

getBuffer
	"Refills InBuf from the midi port."

	| buf count |
	(Port isNil) ifTrue: [^self error: 'Midi port not open'].
	InBuf resetAll.
	buf _ InBuf buffer.
	count _ self primGet: Port bytes: buf size: buf size.
	InBuf setReadLimit: count.!

getByte
	"Gets next input byte without waiting. Answers #NoInput if there isn't any."

	(Debug | Port isNil)
		ifTrue: [^#NoInput]
		ifFalse:
			[(InBuf atEnd) ifTrue: [self getBuffer].	"get more input data"
			 (InBuf atEnd)	"nothing?"
				ifTrue:  [^#NoInput]
				ifFalse: [^InBuf next]].!

readStream

	self getBuffer.
	^InBuf! !

!MidiInterface class methodsFor: 'primitive tests'!

noopTest
	"MidiPrimTests noopTest"

	^Time millisecondsToRun:
		[1000 timesRepeat: [self primNoop]]!

openCloseTest
	"MidiPrimTests openCloseTest"

	self primOpen: 0.
	self primClose: 0.!

receiveTest
	"Receive for 5 seconds from to midi port 0."
	"MidiPrimTests receiveTest"

	| bytes count startTime |
	bytes _ ByteArray new: 1000.
	count _ 0.
	startTime _ Time millisecondClockValue.
	self primOpen: 0.
	[(Time millisecondClockValue - startTime) < 5000]
		whileTrue:
			[count _ count +
				(self primGet: 0 bytes: bytes size: 1000)].
	self primClose: 0.
	^count!

sendTest
	"Send 10000 bytes to midi port 0."
	"MidiPrimTests sendTest"

	| bytes time |
	bytes _ ByteArray new: 1000 withAll: 248. "Midi clock byte"
	self primOpen: 0.
	time _ Time millisecondsToRun: [
		1000 timesRepeat:
			[self primPut: 0 bytes: bytes count: 10]].
	self primClose: 0.
	^time! !

!MidiInterface class methodsFor: 'primitives'!

primClose: portNumber

	<primitive: -2>
	self primitiveFailed.!

primGet: portNumber bytes: aByteArray size: aNumber

	<primitive: -4>
	self primitiveFailed.!

primNoop

	<primitive: -5>
	self primitiveFailed.!

primOpen: portNumber

	<primitive: -1>
	self primitiveFailed.!

primPut: portNumber bytes: aByteArray count: aNumber

	<primitive: -3>
	self primitiveFailed.! !

MidiInterface initialize!


Object subclass: #MidiRecorder
	instanceVariableNames: 'midiTable times cmds arg1s arg2s next limit state lastCmd lastSelector activeNotes pedalDown pedalNotes port inBuf '
	classVariableNames: 'DefaultMidiTable InitialSize '
	poolDictionaries: ''
	category: 'Music-MidiScores'!
MidiRecorder comment:
'I am a recorder for Midi events. I can be used to record a score in real time.'!


!MidiRecorder methodsFor: 'initialize-release'!

reset
	"Reset to a state of being ready to record Midi events."

	(times isNil)
		ifTrue:
			[limit _ InitialSize.
			 times _ Array new: InitialSize.
			 cmds _ Array new: InitialSize.
			 arg1s _ Array new: InitialSize.
			 arg2s _ Array new: InitialSize]
		ifFalse:
			[1 to: limit do:
				[: index |
					 times at: index put: nil.
					 cmds at: index put: nil.
					 arg1s at: index put: nil.
					 arg2s at: index put: nil].
			].

	next _ 1.
	state _ #idle.
	lastCmd _ nil.
	lastSelector _ nil.
	activeNotes _ nil.
	pedalNotes _ nil.
	inBuf _ ByteArray new: 4.!

resetMidiTable
	"Resets the Midi table to default table from my class."

	midiTable _ DefaultMidiTable deepCopy.! !

!MidiRecorder methodsFor: 'midi filtering'!

ignoreChan: chan
	"Don't record events on the given midi channel."

	((chan isInteger not) | (chan < 1) | (chan > 16))
		ifTrue: [^self error: 'bad Midi channel specification'].

	"two-arg channel messages"
	#(128 144 160 176 224)
		do: [: msg |
			midiTable at: (msg bitOr: chan - 1) put: #ignoreTwo:].

	"one-arg channel messages"
	#(192 208)
		do: [: msg |
			midiTable at: (msg bitOr: chan - 1) put: #ignoreOne:].!

ignoreCmd: midiCmd
	"Ignore the given midi command on all channels."

	| cmd | 

	((midiCmd isInteger not) | (midiCmd < 128) | (midiCmd > 255))
		ifTrue: [^self error: 'bad Midi command'].

	(#(240 241 244 245 247 249 253) includes: midiCmd)
		ifTrue:		"can't ignore any of these commands"
			[^self error: 'You can''t ignore this Midi command: ', midiCmd printString].

	(midiCmd < 240)
		ifTrue:		"its a midi channel command"
			[cmd _ midiCmd bitAnd: 2r11110000.
			 1 to: 16 do: [: chan |
				midiTable
					at: (cmd bitOr: chan - 1)
					put: ((#(128 144 160 176 224) includes: cmd)
								ifTrue: [#ignoreTwo:]
								ifFalse: [#ignoreOne:])].
			^self
		].

	(#(246 248 250 251 252 254 255) includes: midiCmd)
		ifTrue:		"zero-arg commands"
			[midiTable at: midiCmd put: #ignore.
			 ^self].

	(midiCmd = 243)
		ifTrue:		"one-arg command"
			[midiTable at: midiCmd put: #ignoreOne:.
			 ^self].

	(midiCmd = 242)
		ifTrue:		"two-arg command"
			[midiTable at: midiCmd put: #ignoreTwo:.
			 ^self].

	"we should not get here"
	self error: 'implementation error'.!

recordOnly: channelList
	"Turn off recording of all but the midi channels given.
	 Midi channels are specified by numbers in the range 1-16."

	channelList do: [: chan |
		((chan isInteger not) | (chan < 1) | (chan > 16))
			ifTrue: [^self error: 'bad Midi channel specification: ', chan printString]].

	1 to: 16 do: [: chan |
		(channelList includes: chan)
			ifFalse: [self ignoreChan: chan].
	].! !

!MidiRecorder methodsFor: 'midi monitor'!

echo: cmd with: arg1 with: arg2
	"Echo the note a third above."

	| cmdType chan |
	cmdType _ cmd bitAnd: 2r11110000.
	chan _ (cmd bitAnd: 2r00001111) + 1.
	(cmdType == 128)
		ifTrue: [^Midi noteOff: arg1 + 7 chan: chan].
	(cmdType == 144)
		ifTrue: [^Midi noteOn: arg1 + 7 vel: arg2 chan: chan].!

funky
	"Play in thirds with the keyboard."
	"MidiRecorder new funky."

	| i |
	self reset.
	self start.
	i _ 2.
	[Sensor anyButtonPressed]
		whileFalse:
			[self recordData.
			 (next > i)
				 ifTrue:
					[self echo: (cmds at: i) with: (arg1s at: i) with: (arg2s at: i).
					 i _ i + 1].
			 ((state == #idle) & (next == i) & (next > 2))
				ifTrue: [next _ i _ 2]].
	self stop.!

monitor
	"Print midi messages to the transcript until any mouse button is pressed. Details: Set ourselves up to record as usual, but every time through the loop print the next command in the recording buffers. If the printing process catches up to the 'record head' (next), and we are in an idle state, reset the record head and i to the beginning."

	| i |
	self reset.
	self start.
	i _ 2.
	[Sensor anyButtonPressed]
		whileFalse:
			[self recordData.
			 (next > i)
				 ifTrue:
					[self printCmd: (cmds at: i) with: (arg1s at: i) with: (arg2s at: i).
					 i _ i + 1].
			 ((state == #idle) & (next == i) & (next > 2))
				ifTrue: [next _ i _ 2]].
	self stop.!

printCmd: cmd with: arg1 with: arg2
	"print the given midi command"

	| cmdType chan |

	cmdType _ cmd bitAnd: 2r11110000.
	chan _ (cmd bitAnd: 2r00001111) + 1.

	(cmdType == 128)
		ifTrue: [^Transcript show: ('key up: ', arg1 printString, ' vel: ', arg2 printString, ' chan: ', chan printString); cr].
	(cmdType == 144)
		ifTrue: [^Transcript show: ('key down: ', arg1 printString, ' vel: ', arg2 printString, ' chan: ', chan printString); cr].
	(cmdType == 160)
		ifTrue: [^Transcript show: ('key after touch: ', arg1 printString, ' vel: ', arg2 printString, ' chan: ', chan printString); cr].
	(cmdType == 176)
		ifTrue: [^Transcript show: ('controller: ', arg1 printString, ' val: ', arg2 printString, ' chan: ', chan printString); cr].
	(cmdType == 192)
		ifTrue: [^Transcript show: ('program change: ', arg1 printString, ' chan: ', chan printString); cr].
	(cmdType == 208)
		ifTrue: [^Transcript show: ('channel after touch ', arg1 printString, ' chan: ', chan printString); cr].
	(cmdType == 224)
		ifTrue: [^Transcript show: ('pitch bend: ', arg1 printString, ' ', arg2 printString, ' chan: ', chan printString); cr].
	(cmd == 240)
		ifTrue: [^Transcript show: ('system exclusive: ', (arg1 at: 1) printString, ' (', arg1 size printString, ' bytes)'); cr].

	Transcript show: 'cmd: ', cmd printString, ' arg1: ', arg1 printString, ' arg2: ', arg2 printString; cr.!

testmonitor: aStream
	"Print midi messages to the transcript until any mouse button is pressed. Use aStream as the source of Midi data. This method is used for debugging."

	| i |
	self reset.
	self start.
	i _ 2.
	[aStream atEnd or: [Sensor anyButtonPressed]]
		whileFalse:
			[self processByte: (aStream next) asciiValue.
			 (next > i)
				 ifTrue:
					[self printCmd: (cmds at: i) with: (arg1s at: i) with: (arg2s at: i).
					 i _ i + 1].
			 ((state == #idle) & (next == i) & (next > 2))
				ifTrue: [next _ i _ 2]].
	self stop.! !

!MidiRecorder methodsFor: 'record/overdub'!

echoTest
	"This test records your playing until a mouse button is pressed, then plays what you played back to you. Answers the recorded score."

	| score |
	self reset.
	score _ self record.
	[Sensor anyButtonPressed] whileTrue: ["wait for button to be released"].
	score playFrom: 0 rate: 1.
	^score!

overDub: aScore
	"Record a new score while performing aScore from the beginning at normal speed. Stop when a mouse button is pressed. Answer the newly recorded 'track'."

	^self overDub: aScore playFrom: 0 rate: 1!

overDub: aScore playFrom: startTime
	"Record a new score while performing aScore from the given starting point at normal speed. Stop when a mouse button is pressed. Answer the newly recorded 'track'."

	^self overDub: aScore playFrom: startTime rate: 1!

overDub: aScore playFrom: startTime rate: rate
	"Record a new score while performing aScore from the given point at the given rate. Stop when a mouse button is pressed. Answer the newly recorded 'track'."

	| mSecsAtStart currTime |
	"ignore key after-pressure, program change, channel pressure (after-touch), and pitch wheel change commands:"
	#(160 192 208 224) do: [: cmd | self ignoreCmd: cmd].
	self reset.

	"Scan for the first note at or after the given time, processing all control changes and program changes during this scanning to establish the proper synthesizer state."
	aScore reset.
	aScore sendEventsThrough: startTime - 1 silent: true.

	"play aScore while recording"
	self start.

	mSecsAtStart _ (Delay millisecondClockValue) - (startTime * 10 // rate).
	[Sensor anyButtonPressed]
		whileFalse:
			[currTime _ ((Delay millisecondClockValue - mSecsAtStart) * rate) // 10.
			 aScore sendEventsThrough: currTime silent: false.
			 self recordData].
	self stop.
	self convertTimes: true rate: rate.
	^self asScore!

record
	"Record a score from the Midi device until a mouse button is pressed."

	^self overDub: Score new! !

!MidiRecorder methodsFor: 'private - recording'!

recordData

	| stream |
	"try to read and process a block of characters from fileDescriptor"
	self recordData: MidiInterface readStream.!

recordData: bufStream
	"Process the contents of bufStream as if it were input."

	[bufStream atEnd]
		whileFalse:
			[self processByte: bufStream next].!

start
	"Start Midi recording."	

	| time |
	"reset  and flush midi port"
	MidiInterface flushInput.

	"just assume there is enough space for the first entry"
	time _ LargePositiveInteger new: 4.
 	Time millisecondClockInto: time.
	times at: 1 put: time.
	cmds at: 1 put: #start.
	next _ 2.!

stop
	"Stop Midi recording."	

	| time |
	time _ LargePositiveInteger new: 4.
 	Time millisecondClockInto: time.
	(next > limit) ifTrue: [self growAll].
	times at: next put: time.
	cmds at: next put: #stop.
	next _ next + 1.! !

!MidiRecorder methodsFor: 'private - table msgs'!

badTableEntry: aByte
	"Implementation error: there is a bad Midi table entry."

	self error: 'implementation error'.!

data: dataByte
	"Implementation error: you should not be using a data byte to index into the Midi action table."

	self error: 'implementation error'.!

endSysExclusive: cmdByte
	"We have received an unexpected 'end system exclusive' command."

	self error: 'unexpected ''End of System Exclusive'' command'.!

ignore: cmdByte
	"Ignore a zero-arg midi command."	

	"do nothing"!

ignoreOne: cmdByte
	"Ignore a one-arg midi command."	

	lastCmd _ cmdByte.
	lastSelector _ #ignoreOne:.
	state _ #ignore1.!

ignoreTwo: cmdByte
	"Ignore a two-arg midi command."	

	lastCmd _ cmdByte.
	lastSelector _ #ignoreTwo:.
	state _ #ignore2.!

recordOne: cmdByte
	"Record a one-arg midi command at the current time."	

	| time |
	(next > limit) ifTrue: [self growAll].
	time _ LargePositiveInteger new: 4.
 	Time millisecondClockInto: time.
	times at: next put: time.
	cmds at: next put: cmdByte.
	lastCmd _ cmdByte.
	lastSelector _ #recordOne:.
	state _ #want1only.!

recordSysExclusive: cmdByte
	"The beginning of a variable length 'system exclusive' command."

	| time |
	(next > limit) ifTrue: [self growAll].
	time _ LargePositiveInteger new: 4.
 	Time millisecondClockInto: time.
	times at: next put: time.
	cmds at: next put: cmdByte.
	arg1s at: next put: (OrderedCollection new: 1000).
		"Make a place to store the system exclusive message. The first byte of this collection will be the manufacture's ID number."
	arg2s at: next put: nil.
	lastCmd _ nil.
	lastSelector _ nil.
	state _ #sysExclusive.!

recordTwo: cmdByte
	"Record a two arg midi command at the current time."	

	| time |
	(next > limit) ifTrue: [self growAll].
	time _ LargePositiveInteger new: 4.
 	Time millisecondClockInto: time.
	times at: next put: time.
	cmds at: next put: cmdByte.
	lastCmd _ cmdByte.
	lastSelector _ #recordTwo:.
	state _ #want1of2.!

undefined: cmdByte
	"We have received an undefined command."

	self error: 'undefined midi command ', cmdByte printString.! !

!MidiRecorder methodsFor: 'private - other'!

asScore

	| score cmd chan note |
	score _ Score new.
	activeNotes _ OrderedCollection new.
	1 to: (next - 1) do: [: i |
		((cmds at: i) < 240)
			ifTrue:
				[cmd _ (cmds at: i) bitAnd: 2r11110000.
				 chan _ (cmds at: i) bitAnd: 2r00001111]
			ifFalse:
				[cmd _ cmds at: i.
				 chan _ nil].
		(cmd == 144)	"note on"
			ifTrue:
				[((arg2s at: i) > 0)	"non-zero velocity"
					ifTrue:
						[note _ NoteElement
							new: (arg1s at: i)
							at: ((times at: i) asFloat / 10.0) rounded
							dur: 60	"just a guess for now"
							vel: (arg2s at: i)
							voice: (chan + 1).
						score add: note.
						activeNotes add: note]
					ifFalse:
						[self setDur: (arg1s at: i) endTime: ((times at: i) asFloat / 10.0) rounded]]].
	^score!

convertTimes: absFlag
	"Convert the absolute time in ticks to time since start of recording for all recorded events. If absFlag is true, then times are absolute. Otherwise, the starting time of the first recorded event is taken to be time zero."

	| timeAtStart |
	timeAtStart _ absFlag ifTrue: [times at: 1] ifFalse: [times at: 2].
	times at: 1 put: 0.
	2 to: times size do:
		[: index |
			times at: index put: ((times at: index) - timeAtStart).
			((cmds at: index) == #stop) ifTrue: [^self]].!

convertTimes: absFlag rate: rate
	"Convert the absolute time in ticks to time since start of recording for all recorded events. If absFlag is true, then times are absolute. Otherwise, the starting time of the first recorded event is taken to be time zero."

	| timeAtStart |
	timeAtStart _ absFlag ifTrue: [times at: 1] ifFalse: [times at: 2].
	times at: 1 put: 0.
	2 to: times size do:
		[: index |
			times at: index put: (((times at: index) - timeAtStart) * rate) rounded.
			((cmds at: index) == #stop) ifTrue: [^self]].!

growAll
	"Grow my recording space and set new limit."	

	times grow.
	cmds grow.
	arg1s grow.
	arg2s grow.
	limit _ times size.!

processByte: aByte
	"Process the given incoming Midi byte and record completed commands."
	"Details: Because this must be fast, it has been hand-tuned. Be careful!!"

	(aByte == 254) ifTrue: [^self].		"filter out active status messages"
	state == #idle
		ifTrue: [
			(aByte < 128)
				ifFalse: [^self perform: (midiTable at: aByte) with: aByte]
				ifTrue: [
					(lastCmd notNil) ifTrue:
						["running status: process this data as if it had
							 a command byte like the lastCmd in front of it."
						 self perform: lastSelector with: lastCmd.
						 "the previous line has put us into a new state;
						 we now 'fall through' this case to process the
						data byte given this new state. For this reason,
						the (state==idle) case must be first."]]].

	state == #ignore1
		ifTrue: [^state _ #idle].

	state == #want2of2
		ifTrue: [
			arg2s at: next put: aByte.
			next _ next + 1.
			^state _ #idle].

	state == #want1of2
		ifTrue: [
			arg1s at: next put: aByte.
			^state _ #want2of2].

	state == #ignore2
		ifTrue: [^state _ #ignore1].

	state == #want1only
		ifTrue: [
			arg1s at: next put: aByte.
			next _ next + 1.
			^state _ #idle].

	state == #sysExclusive
		ifTrue: [
			(aByte < 128)
				ifTrue: [
					"record a system exclusive data byte"
					(arg1s at: next) addLast: aByte]
				ifFalse: [
					"a system exclusive message is terminated by any command byte"
					next _ next + 1.
					state _ #idle.
					aByte == 247
						ifTrue: [^self]							"if endSysExclusive command, nothing left to do"
						ifFalse: [^self processByte: aByte]]].	"otherwise, we must handle this new command"!

setDur: pitch endTime: endTime
	"Set the duration for the note(s) which just ended."

	| notes |
	notes _ activeNotes removeAllSuchThat: [: note | (note pitch == pitch)].
	notes do: [: note | note dur: (endTime - note time)].! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

MidiRecorder class
	instanceVariableNames: ''!


!MidiRecorder class methodsFor: 'class initialization'!

initialize
	"MidiRecorder initialize."

	InitialSize _ 5000.

	"Build the default midi action table. This table is a 255 element array of actions to be performed when a byte whose value is the entries index is received. Note that DefaultMidiTable at: 0 is not defined; however, we never actually use a data byte (< 128) to index into the table."

	DefaultMidiTable _ Array new: 255 withAll: #badTableEntry:.
	1 to: 127 do: [: i | DefaultMidiTable at: i put: #data:].

	128 to: 143 do: [: i | DefaultMidiTable at: i put: #recordTwo:].		"Key off"
	144 to: 159 do: [: i | DefaultMidiTable at: i put: #recordTwo:].		"Key on"
	160 to: 175 do: [: i | DefaultMidiTable at: i put: #recordTwo:].		"Key pressure (after-touch)"
	176 to: 191 do: [: i | DefaultMidiTable at: i put: #recordTwo:].		"Control change"
	192 to: 207 do: [: i | DefaultMidiTable at: i put: #recordOne:].		"Program change"
	208 to: 223 do: [: i | DefaultMidiTable at: i put: #recordOne:].		"Channel pressure"
	224 to: 239 do: [: i | DefaultMidiTable at: i put: #recordTwo:].		"Pitch wheel change"

	DefaultMidiTable at: 240 put: #recordSysExclusive:.	"Start of system exclusive block"
	DefaultMidiTable at: 241 put: #undefined:.
	DefaultMidiTable at: 242 put: #ignoreTwo:.			"Song position select"
	DefaultMidiTable at: 243 put: #ignoreOne:.			"Song select"
	DefaultMidiTable at: 244 put: #undefined:.
	DefaultMidiTable at: 245 put: #undefined:.
	DefaultMidiTable at: 246 put: #ignore:.				"Tune request"
	DefaultMidiTable at: 247 put: #endSysExclusive:.	"End of system exclusive block"
	DefaultMidiTable at: 248 put: #ignore:.				"Timing clock"
	DefaultMidiTable at: 249 put: #undefined:.
	DefaultMidiTable at: 250 put: #ignore:.				"Start"
	DefaultMidiTable at: 251 put: #ignore:.				"Continue"
	DefaultMidiTable at: 252 put: #ignore:.				"Stop/Clock"
	DefaultMidiTable at: 253 put: #undefined:.
	DefaultMidiTable at: 254 put: #ignore:.				"Active sensing"
	DefaultMidiTable at: 255 put: #ignore:.				"System reset"! !

!MidiRecorder class methodsFor: 'instance creation'!

new

	^(super new) resetMidiTable; reset; yourself! !

MidiRecorder initialize!


ScoreElement subclass: #NoteElement
	instanceVariableNames: 'velocityPitchVoice duration '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-MidiScores'!
NoteElement comment:
'I am a subclass of ScoreElement which is used to represent a soundable tone. In order to allow one to fit larger scores in memory, my space is optimized at a very slight cost in additional access time. I pack three of my parameters into one instance SmallInteger instance variable as follows:

	my voice (0-31) -- lowest 5 bits of velocityPitchVoice
	my pitch (0-127) -- next 7 bits of velocityPitchVoice
	my velocity (0-127) -- next 7 bits of velocityPitchVoice
'!


!NoteElement methodsFor: 'initialize-release'!

initialize
	"Initialize myself with reasonable default parameters."

	time _ 0.
	velocityPitchVoice _ 309121.

	"the above is equivalent to:
		velocityPitchVoice _ 0
		self time: 0; pitch: 60; velocity: 75; dur: 60; voice: 1."!

vel: vel pitch: pitch voice: voice
	"Initialize myself with the given velocity, pitch, and voice and a reasonable default time and duration."

	time _ 0.
	duration _ 60.
	velocityPitchVoice _
		(((vel bitAnd: 127) bitShift: 12) bitOr:
		 ((pitch bitAnd: 127) bitShift: 5)) bitOr:
		  (voice bitAnd: 31).

	"the above is equivalent to:
		self time: 0; pitch: 60; velocity: 75; dur: 60; voice: 1."! !

!NoteElement methodsFor: 'accessing'!

dur

	^duration!

dur: newDur

	duration _ newDur.!

offTime

	^time + duration!

pitch

	^(velocityPitchVoice bitShift: -5) bitAnd: 127!

pitch: newPitch

	velocityPitchVoice _
		((newPitch bitAnd: 127) bitShift: 5) bitOr:
			(velocityPitchVoice bitAnd: 520223 "(127 bitShift: 12) + 31").!

velocity

	^velocityPitchVoice bitShift: -12!

velocity: newVelocity

	velocityPitchVoice _
		((newVelocity bitAnd: 127) bitShift: 12) bitOr:
			(velocityPitchVoice bitAnd: 4095 "low 7+5 bits").!

voice

	^velocityPitchVoice bitAnd: 31!

voice: newVoice

	velocityPitchVoice _
		(newVoice bitAnd: 31) bitOr:
			(velocityPitchVoice bitAnd: -32 "all but lowest 5 bits").! !

!NoteElement methodsFor: 'testing'!

isTone

	^true! !

!NoteElement methodsFor: 'performing'!

perform
	"Perform myself."

	MidiInterface
		noteOn: (self pitch)
		vel: (self velocity)
		chan: (self voice).! !

!NoteElement methodsFor: 'print/store'!

printOn: aStream

	aStream nextPutAll:
		('time: ', self time printString,
		 ' pitch: ', self pitch printString,
		 ' dur: ', self dur printString,
		 ' vel: ', self velocity printString,
		 ' voice: ', self voice printString, '\') withCRs.!

storeAdagioOn: aStream previous: previousTone next: nextTone

	| deltaTime |
	(previousTone isNil) ifTrue:
		[aStream nextPutAll:
			't', self time printString,
			' p', self pitch printString,
			' u', self dur printString,
			' l', self velocity printString,
			' v', self voice printString]
		ifFalse: 
			[aStream nextPutAll: 'p', self pitch printString.
			 (self dur ~= previousTone dur)
				ifTrue: [aStream nextPutAll: ' u', self dur printString].
			 (self velocity ~= previousTone velocity)
				 ifTrue: [aStream nextPutAll: ' l', self velocity printString].
			 (self voice ~= previousTone voice)
				ifTrue: [aStream nextPutAll: ' v', self voice printString]].
	(nextTone isNil)
		ifTrue: [aStream cr]
		ifFalse:
			[deltaTime _ nextTone time - self time.
			 (deltaTime == 0)
				ifTrue: [aStream nextPutAll: ', ']
				ifFalse:
					[(deltaTime ~~ self dur)
						ifTrue:
							[aStream nextPutAll: ' n', deltaTime printString.
							 aStream cr]
						ifFalse: [aStream cr]]].! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

NoteElement class
	instanceVariableNames: ''!


!NoteElement class methodsFor: 'instance creation'!

new
	"Creates a new note."

	^super new initialize!

new: pitch at: time dur: duration
	"Creates a new instance of me with the given pitch starting at the given starting time and lasting for the given duration. Other attributes are given reasonable defaults."

	^(self basicNew)
		vel: 75 pitch: pitch voice: 1;
		time: time;
		dur: duration!

new: pitch at: time dur: duration vel: velocity voice: voice
	"Creates a new instance of me with the given pitch, velocity, and voice starting at the given starting time and lasting for the given duration."

	^(self basicNew)
		vel: velocity pitch: pitch voice: voice;
		time: time;
		dur: duration! !

View subclass: #PianoRollView
	instanceVariableNames: 'timeOffset visibleNotes rate selected leftSelTime rightSelTime '
	classVariableNames: 'LeftMarker PitchOffsets RightMarker StaffLeft StaffRight StaffSpace VOffset WindowLeft WindowRight '
	poolDictionaries: ''
	category: 'Music-MidiScores'!
PianoRollView comment:
'I represent a view of a PianoRoll.  I display pin placement with bold dots and wire routing as thin lines.'!


!PianoRollView methodsFor: 'initialize-release'!

initializeView
	"Initialize my instance variables."

	| topView |

	timeOffset _ 0.
	visibleNotes _ OrderedCollection new.
	rate _ 100.0.
	selected _ IdentitySet new.
	leftSelTime _ (model isEmpty) ifTrue: [0] ifFalse: [model first time].
	rightSelTime _ model scoreTime.!

open
	"Open a new view of myself."

	| topView |

	topView _ StandardSystemView new
		model: model;
		label: 'Piano Roll Editor';
		addSubView: self.
	topView controller open.! !

!PianoRollView methodsFor: 'accessing'!

cacheUpdate

	self timeOffset: timeOffset!

rate

	^rate!

rate: newRate

	rate _ newRate!

timeOffset

	^timeOffset!

timeOffset: aNumber

	timeOffset _ aNumber.
	visibleNotes _ visibleNotes species new: (visibleNotes size).
	model do: [: note |
		((note time >= timeOffset) | ((note offTime) >= timeOffset))
			ifTrue: [visibleNotes add:
						(Association
							key: note
							value: (self displayTransform: (self noteBox: note)) rounded)].
		(note time >= (timeOffset + StaffRight))
			ifTrue: [^self].
	].!

windowEndTime

	^timeOffset + ((StaffRight - StaffLeft) // 10)! !

!PianoRollView methodsFor: 'selections'!

addToSelection: aNote

	selected add: aNote.
	self displayNote: aNote selected: true.!

clearSelection

	selected _ selected species new.
	self displayView.!

nearestNote: aPoint

	| possibilities closest distToClosest noteAssociation noteBox |

	possibilities _ OrderedCollection new.
	visibleNotes do:
		[: noteAssociation |
			noteBox _ noteAssociation value deepCopy.
			noteBox origin: (noteBox origin - (2@4)); corner: (noteBox corner + (2@4)).
			(noteBox containsPoint: aPoint)
				ifTrue: [possibilities add: noteAssociation]].

	possibilities isEmpty ifTrue: [^nil].	"no nearby notes"

	closest _ possibilities first.
	distToClosest _ (closest value center y - aPoint y) abs.

	possibilities do:
		[: noteAssociation |
			noteBox _ noteAssociation value.
			(noteBox containsPoint: aPoint)
				ifTrue: [^noteAssociation key].
			((noteBox center y - aPoint y) abs < distToClosest)
				ifTrue:
					[distToClosest _ (noteBox center y - aPoint y) abs.
					 closest _ noteAssociation].
			].

	^closest key!

noteIsSelected: aNote
	"Answer true iff the given note is in the selection."

	^selected includes: aNote!

removeFromSelection: aNote

	selected remove: aNote ifAbsent: [].
	self displayNote: aNote selected: false.!

selected

	^selected!

timeRangeEnd

	^rightSelTime!

timeRangeStart

	^leftSelTime! !

!PianoRollView methodsFor: 'displaying'!

displayNote: aNote selected: noteSelected
	"Display a note as a bar in a piano roll."

	| noteBox |

	noteBox _ (self displayTransform: (self noteBox: aNote)) rounded.
	self displayNoteBox: noteBox selected: noteSelected.!

displayNoteBox: noteBox selected: noteSelected
	"Display a note box with the given shade of gray."

	"note bars are four bits thick"
	noteSelected
		ifTrue:
			[Display fill: (noteBox expandBy: 0@3) mask: Form black.
			 Display fill: (noteBox expandBy: -1@2) mask: Form gray]
		ifFalse:
			[Display fill: (noteBox expandBy: 0@3) mask: Form white.
			 Display fill: (noteBox expandBy: 0@2) mask: Form black].!

displaySelectionBar
	"Displays the selection bar along the bottom of a piano roll."

	| selBar |

	selBar _ self selectionBar.
	Display border: selBar width: 1.
	self displaySelBarInsides: selBar.!

displayStaves
	"Displays the staff lines for a piano roll."

	| vPos1 vPos2 tempRect |

	(0 to: 4) do: [: i |
		vPos1 _ VOffset + (StaffSpace * i).
		vPos2 _ vPos1 + (StaffSpace * 6).

		"draw a line in the top staff"
		tempRect _ StaffLeft@vPos1 corner: StaffRight@vPos1.	"zero height"
		tempRect _ (self displayTransform: tempRect) rounded.
		tempRect bottom: (tempRect bottom + 1).		"staff lines are always one pixel wide"
		Display fill: tempRect mask: Form black.

		"draw similar line in the bottom staff"
		tempRect _ StaffLeft@vPos2 corner: StaffRight@vPos2.	"zero height"
		tempRect _ (self displayTransform: tempRect) rounded.
		tempRect bottom: (tempRect bottom + 1).		"staff lines are always one pixel wide"
		Display fill: tempRect mask: Form black.
	].!

displayView
	"Display a score in piano-roll notation."

	| color |

	self timeOffset: timeOffset.
	self clearInside.
	self displayStaves.
	self displaySelectionBar.
	visibleNotes do:
		[: noteAssociation |
			self
				displayNoteBox: (noteAssociation value)
				selected: (self noteIsSelected: noteAssociation key)].! !

!PianoRollView methodsFor: 'time selection bar'!

displaySelBarInsides: selBar
	"Displays the selected time range in the selection bar along the bottom of a piano roll."

	| leftRect rightRect |

	"be sure that leftSelTime <= rightSelTime"
	(leftSelTime > rightSelTime) ifTrue: [rightSelTime _ leftSelTime].

	"compute marker box locations; nil indicates that the time range selection is not visible"
	(leftRect _ self leftMarkerRect: selBar) isNil
		ifTrue: [^self].
	(rightRect _ self rightMarkerRect: selBar) isNil
		ifTrue: [^self].

	"display marker boxes"
	LeftMarker
		displayOn: Display
		at: leftRect topLeft
		clippingBox: leftRect
		rule: (Form over)
		mask: (Form black).
	RightMarker
		displayOn: Display
		at: rightRect topLeft
		clippingBox: rightRect
		rule: (Form over)
		mask: (Form black).

	"highlight the time range selection"
	Display
		fill: ((selBar topLeft + 1) corner: (leftRect bottomLeft))
		mask: Form white.
	Display
		fill: ((rightRect topRight) corner: (selBar bottomRight - 1))
		mask: Form white.
	Display
		fill: ((leftRect topRight) corner: (rightRect bottomLeft))
		mask: Form gray.!

dragTimeMarker
	"Pick the time marker to move and drag it."

	| selBar scale cursorTime |

	selBar _ self selectionBar.
	scale _ 1000.0 / selBar width.

	 cursorTime _ timeOffset + ((Sensor cursorPoint x - selBar left) * scale) truncated.
	 ((cursorTime <= leftSelTime) or: [Sensor leftShiftDown])
		ifTrue: [self dragTimeMarker: #left]
		ifFalse: [self dragTimeMarker: #right].!

dragTimeMarker: which
	"Drag the time marker specified by which, subject to various constraints on its movement."

	| selBar scale markerWidthInTimeUnits leftLimit rightLimit cursorX lastCursorX cursorTime time constrainedX lastConstrainedX |

	selBar _ self selectionBar.
	scale _ 1000.0 / selBar width.
	markerWidthInTimeUnits _  (((RightMarker extent x) * scale) + 0.5) rounded.

	(which == #left)
		ifTrue: [
			leftLimit _ timeOffset + markerWidthInTimeUnits + 1.
			(self rightMarkerVisible: selBar)
				ifTrue: [rightLimit _ timeOffset + 999 - markerWidthInTimeUnits]
				ifFalse: [rightLimit _ timeOffset + 999]].
	(which == #right)
		ifTrue: [
			rightLimit _ timeOffset + 999 - markerWidthInTimeUnits.
			(self leftMarkerVisible: selBar)
				ifTrue: [leftLimit _ timeOffset + markerWidthInTimeUnits + 1]
				ifFalse: [leftLimit _ timeOffset + 1]].

	[Sensor redButtonPressed]
		whileTrue:
			[cursorX _ Sensor cursorPoint x.
			 (cursorX ~= lastCursorX)
				ifTrue:
					[lastCursorX _ cursorX.
					 cursorTime _ timeOffset + ((cursorX - selBar left) * scale) truncated.
					 (cursorTime <= leftLimit) ifTrue: [time _ leftLimit].
					 (cursorTime >= rightLimit) ifTrue: [time _ rightLimit].
					 ((cursorTime > leftLimit) & (cursorTime < rightLimit))
						ifTrue:
							[time _ cursorTime].
					 time _ time min: model scoreTime.
					 constrainedX _ ((time - timeOffset) / scale) + selBar left.
					 (lastConstrainedX notNil)	"not the first time"
						ifTrue: [self showLine: lastConstrainedX; showLine: constrainedX]
						ifFalse: [self showLine: constrainedX].
					 lastConstrainedX _ constrainedX.
					 (which == #left)
						ifTrue:
					 		[leftSelTime _ time.
							 rightSelTime _ time max: rightSelTime].
					 (which == #right)
						ifTrue:
					 		[rightSelTime _ time.
							 leftSelTime _ time min: leftSelTime].
					 self displaySelBarInsides: selBar]].

	self showLine: lastConstrainedX.!

leftMarkerRect: selBar
	"Answer a rectangle, in physical coordinates, in which to display the left time selection marker. Answer the left edge if the left marker itself is not visible but the selection is. Answer nil if no part of the selection is visible."

	| top leftEdge rightEdge leftSel |

	"compute all quantities in physical coordinates"
	top _ selBar top + 1.
	leftEdge _ selBar left.
	rightEdge _ selBar right.
	leftSel _ (self displayTransform: (self timeToX: leftSelTime)@0) rounded x.

	((leftSel > leftEdge) & (leftSel < rightEdge))
		ifTrue:
			["the left selection marker is visible"
			 leftSel _ leftSel - (LeftMarker extent x) max: leftEdge.
			 ^leftSel@top extent: LeftMarker extent]
		ifFalse:
			[leftSel >= rightEdge
				ifTrue:
					["The right selector is to the right of leftSel and leftSel is to the right of the right edge, thus no part of the selection is visible."
					 ^nil]
				ifFalse:
					["The selection is visible even though the selection marker is not."
					 ^((leftEdge + 1)@top extent: 0@(selBar height - 2))]].!

leftMarkerVisible: selBar
	"Answer true if the left marker is visible."

	| marker |
	marker _ self leftMarkerRect: selBar.
	^marker notNil and: [marker extent x > 0]!

rightMarkerRect: selBar
	"Answer a rectangle, in physical coordinates, in which to display the right time selection marker. Answer the right edge if the right marker itself is not visible but the selection is. Answer nil if no part of the selection is visible."

	| top leftEdge rightEdge rightSel |

	"compute all quantities in physical coordinates"
	top _ selBar top + 1.
	leftEdge _ selBar left.
	rightEdge _ selBar right.
	rightSel _ (self displayTransform: (self timeToX:rightSelTime)@0) rounded x.

	((rightSel > leftEdge) & (rightSel < rightEdge))
		ifTrue:
			["the right selection marker is visible"
			 rightSel _ rightSel min: rightEdge - (RightMarker extent x).
			 ^rightSel@top extent: RightMarker extent]
		ifFalse:
			[rightSel <= leftEdge
				ifTrue:
					["The left selector is to the left of rightSel and rightSel is to the left of the left edge, thus no part of the selection is visible."
					 ^nil]
				ifFalse:
					["The selection is visible even through the selection marker is not."
					 ^((rightEdge - 1)@top extent: 0@(selBar height - 2))]].!

rightMarkerVisible: selBar
	"Answer true if the left marker is visible."

	| marker |
	marker _ self rightMarkerRect: selBar.
	^marker notNil and: [marker extent x > 0]!

selectionBar
	"Answers the selection bar rectangle in physical coordinates."

	| rect |

	rect _
		(self displayTransform:
			(StaffLeft@1200 corner: StaffRight@1200)) rounded.	"zero height"
	rect bottom: (rect bottom - 5).
	rect top: (rect bottom - 17).
	^rect!

showLine: XPosition
	"Show a vertical line over the current time selection."

	| top bottom rect |

	top _ self insetDisplayBox top + 5.
	bottom _ self selectionBar top - 5.
	rect _ XPosition@top corner: (XPosition + 1)@bottom.

	Display reverse: rect.! !

!PianoRollView methodsFor: 'scrolling'!

scrollTo: time

	self timeOffset: ((time min: model scoreTime) max: 0).
	self display.! !

!PianoRollView methodsFor: 'controller access'!

defaultControllerClass
	"Answer the class of the default controller for the receiver."

	^PianoRollController! !

!PianoRollView methodsFor: 'window access'!

defaultWindow
	"Answer a Rectangle of specified by the user."

	^0@0 corner: 1100@1200! !

!PianoRollView methodsFor: 'mapping'!

noteBox: aNote
	"Answer the given note's display box in virtual coordinates, clipped to the virtual viewing rectangle."

	| startX endX vOffset |

	startX _ (self timeToX: aNote time) max: StaffLeft.
	endX _ (self timeToX: (aNote offTime)) min: StaffRight.
	vOffset _ self pitchToY: aNote pitch.

	^startX@vOffset corner: endX@vOffset	"zero height rectangle"!

pitchToY: pitch
	"Convert from a pitch between 0 and 127 to a y-offset."

	| yForPitchZero octaveOffset noteOffset |

	yForPitchZero _ VOffset + (22.5 * StaffSpace) truncated.
	octaveOffset _ (pitch // 12) * (3.5 * StaffSpace) truncated.
	noteOffset _ PitchOffsets at: pitch \\ 12.
	^yForPitchZero - octaveOffset - noteOffset!

timeToX: time

	^StaffLeft + (time - timeOffset)! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

PianoRollView class
	instanceVariableNames: ''!


!PianoRollView class methodsFor: 'class initialization'!

initialize
	"PianoRollView initialize."

	| quarterSpace halfSpace threeQuarterSpace |

	VOffset _ 300.
	StaffSpace _ 50.
	StaffLeft _ 50.
	StaffRight _ 1050.
	WindowLeft _ 0.
	WindowRight _ 1100.

	halfSpace _ (StaffSpace / 2.0 + 1.0) rounded.
	quarterSpace _ (StaffSpace / 4.0 - 2.0) rounded.
	threeQuarterSpace _ (StaffSpace * 3.0 / 4.0 + 1.0) rounded.
	PitchOffsets _ Dictionary new.
	PitchOffsets at: 0 put: (StaffSpace * 0).
	PitchOffsets at: 1 put: (StaffSpace * 0) + quarterSpace.
	PitchOffsets at: 2 put: (StaffSpace * 0) + halfSpace.
	PitchOffsets at: 3 put: (StaffSpace * 0) + threeQuarterSpace.
	PitchOffsets at: 4 put: (StaffSpace * 1).
	PitchOffsets at: 5 put: (StaffSpace * 1) + halfSpace.
	PitchOffsets at: 6 put: (StaffSpace * 1) + threeQuarterSpace.
	PitchOffsets at: 7 put: (StaffSpace * 2).
	PitchOffsets at: 8 put: (StaffSpace * 2) + quarterSpace.
	PitchOffsets at: 9 put: (StaffSpace * 2) + halfSpace.
	PitchOffsets at: 10 put: (StaffSpace * 2) + threeQuarterSpace.
	PitchOffsets at: 11 put: (StaffSpace * 3).

RightMarker _
	(Form
		extent: 13@15
		fromArray: #(
			2r1111111111111000
			2r1000000011001000
			2r1000000111001000
			2r1000000111001000
			2r1000001111001000
			2r1000011111001000
			2r1000111111001000
			2r1001111111001000
			2r1000111111001000
			2r1000011111001000
			2r1000001111001000
			2r1000000111001000
			2r1000000111001000
			2r1000000011001000
			2r1111111111111000)
		offset: 0@0).

LeftMarker _
	(Form
		extent: 13@15
		fromArray: #(
			2r1111111111111000
			2r1001100000001000
			2r1001110000001000
			2r1001110000001000
			2r1001111000001000
			2r1001111100001000
			2r1001111110001000
			2r1001111111001000
			2r1001111110001000
			2r1001111100001000
			2r1001111000001000
			2r1001110000001000
			2r1001110000001000
			2r1001100000001000
			2r1111111111111000)
		offset: 0@0).! !

!PianoRollView class methodsFor: 'instance creation'!

createOn: aScore
	"Create but do not schedule a view of aScore."
	"(PianoRollView createOn: (Score new)) openView"

	| pianoRollView |
	pianoRollView _ self new
		model: aScore;
		borderWidth: 2;
		insideColor: Form white.
	pianoRollView initializeView.
	^pianoRollView!

openOn: aScore
	"Create and schedule a view of aScore."
	"PianoRollView openOn: (Score new)"

	(self createOn: aScore) open.! !

PianoRollView initialize!
