Object subclass: #Voice
	instanceVariableNames: 'name instruments stream '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Voices'!
Voice comment:
'Subclasses of Voice implement the mappings between symbolic event parameters (like timbre descriptions and pitch names) and concrete output event data.
They can be used for generating sound compiler notelists or (real-time output or file dumps of) MIDI data.

The instance variables are:
	instruments <set of symbols, associations or devices> my instruments
	name 		<String or Symbol> my handle or name
	stream		<Stream> my optional output stream

The global object Voices is a dictionary that note events can refer to with integer or symbolic keys.'!


!Voice methodsFor: 'initialize-release'!

initializeNamed: aName
	"set up a named Voice"

	(aName isKindOf: Number) ifTrue: [
		name _ aName printString
	] ifFalse: [
		name _ aName ].
	instruments _ Set new. 
	Voices at: name asSymbol put: self.! !

!Voice methodsFor: 'accessing'!

instruments
	"return my instruments."

	^instruments!

instruments: newValue
	"set my instruments."

	instruments _ newValue!

stream
	"answer the stream of the receiver voice"

	^stream!

stream: someStream
	"plug in a stream to the receiver voice"

	stream _ someStream! !

!Voice methodsFor: 'events'!

playEvent: anEvent
	"make sure my subclasses implement this"

	self subclassResponsibility!

playEvent: anEvent at: aTime
	"make sure my subclasses implement this"

	self subclassResponsibility! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Voice class
	instanceVariableNames: ''!


!Voice class methodsFor: 'class initialization'!

initialize
	"set up the shared dictionary for caching voices"
	"Voice initialize."

	Smalltalk at: #Voices put: Dictionary new!

reset
	"reset all sub-instances of Voice"
	"Voice reset"

	self allSubInstancesDo: [ :aV |
		aV reset ]! !

!Voice class methodsFor: 'constant access'!

defaultVoice
	"return some random voice (the first in the dictionary)"

	Voices isEmpty
		ifTrue: [ Voices at: #0 put: (MidiVoice onDevice: (MidiDevice newOn: 1) channel: 0) ].
	^Voices associations first value!

named: aName
	"return the shared dictionary's voice item at the given name (or the name itself if it's a voice)"

	(aName isKindOf: Voice) ifTrue: [
		^aName ].
	(aName class == SmallInteger) ifTrue: [
		^PrimMidiVoice new ].
"		(Voices includesKey: aName printString asSymbol)
			ifTrue: [^Voices at: aName printString asSymbol ifAbsent: [ nil ] ]
			ifFalse: [^PrimMidiVoice new
					""^MidiVoice onDevice: (MidiDevice newOn: 1) channel: aName"" ] ].
"	^Voices at: aName ifAbsent: [ nil ]! !

Voice initialize!


Voice subclass: #MidiVoice
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Voices'!
MidiVoice comment:
'A MidiVoice is used to play note events onto one or more output devices.
I store MidiDevices in my instruments and play events onto them when asked to.

Instance variables:
	instruments	<set of Associations>	my instruments, stored as channel -> MidiDevice
'!


!MidiVoice methodsFor: 'initialize-release'!

reInitialize
	"re-set a named MidiVoice by restarting its value (device)"
	
	instruments do: [ :anIn |
		anIn value reInitialize ]!

reset
	"reset my devices to reelease their ports"
	
	instruments do: [ :anIn |
		anIn key reset ]! !

!MidiVoice methodsFor: 'accessing'!

addDevice: aMDevice channel: aChannel
	"add the given pair to the set of devices"
	
	instruments add: (aChannel -> aMDevice)!

number
	"return my channel or icon number"

	| n |
	n _ Number readFrom: (ReadStream on: (name)).
	n = 0 ifTrue: [
		^self channel ].
	^ n! !

!MidiVoice methodsFor: 'events'!

playEvent: anEvent
	"send the given Event to all device/channel pairs (right now)"

	instruments do: [ :anIn |
		anIn value play: anEvent on: anIn key ]!

playEvent: anEvent at: aTime
	"send the given Event to all device/channel pairs"

	instruments do: [ :anIn |
		anIn value play: anEvent on: anIn key ]! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

MidiVoice class
	instanceVariableNames: ''!


!MidiVoice class methodsFor: 'instance creation'!

newNamed: aName onDevice: aMidiDevice channel: aChannel
	"set up a MidiVoice on the given device and channel"
	"MidiVoice newNamed: 'xxx' onDevice: (MidiFB01 newOn: 1) channel: 1"

	| me |
	me _ self new initializeNamed: aName.
	me instruments add: (aChannel -> aMidiDevice).
	^me!

onDevice: aMidiDevice channel: aChannel
	"set up a MidiVoice on the given device and channel"
	"MidiVoice onDevice: (MidiFB01 newOn: 1) channel: 1"

	| me |
	me _ self new initializeNamed: aChannel.
	me instruments add: (aChannel -> aMidiDevice).
	^me! !

RS232Port subclass: #MidiPort
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Voices'!
MidiPort comment:
'An Instance of MidiPort exists for each physical Midi cable and is the output driver for the Midi commands sent by MidiDevices.
A MidiDevice points to a MidiPort and sends it time-stamped byteArrays.
The MidiPort uses the ExternalPort I/O primitives for byteArray input/output'!


!MidiPort methodsFor: 'port initialization'!

resetMidiOut
	"reset the receiver as a MIDI Output channel (***Special SerialPort Routine Code***)"

	self writeStatus: -33 with: 0!

setMidiOut
	"set the receiver to be a MIDI Output channel (***Special SerialPort Routine Code***)"

	self writeStatus: -32 with: 0! !

!MidiPort methodsFor: 'sending'!

send: aByteArray length: length
	"now finally send the byte array out the port using the famous ExternalPort pimitive"

"	Transcript show: 'M:  ',
		aByteArray printString; cr.
"	self send: aByteArray 
		startingAt: 1 
		size: length 
		signal: sendSemaphore.
	sendSemaphore wait! !

!MidiPort methodsFor: 'receiving'!

receive: aByteArray length: length
	"try to read some bytes and return at most length"

	self receiveBuffer.
	^receiveStream next: length! !

!MidiPort methodsFor: 'access primitives'!

close
	"close and reset the port"

	self clearMidiMode.
	super close! !

!MidiPort methodsFor: 'private'!

reopen
	"no-op for now"
"
	self setMidiMode.
	super reopen
"! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

MidiPort class
	instanceVariableNames: ''!


!MidiPort class methodsFor: 'maintenance'!

flush
	"release and close all of my instances"
	"MidiPort flush"

	MidiPort allInstances do: [ :aMP |
		aMP close.
		Smalltalk removeDependent: aMP.
		aMP release
	]! !

Voice subclass: #OpcodeVoice2
	instanceVariableNames: 'fileType tracks ppq tempo '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Voices'!
OpcodeVoice2 comment:
'This voice class is a file I/O convertor that can transform Standard 1.0 format MIDI file dumps into 
EventLists and/or play them in real-time. 
The messages here will look like:
(OpcodeVoice2 newOn: ''Scores:n3.a.sc'') loadOnto: (EventList newNamed: #n3.a)

A type 0 MIDI file contains a single, multi-channel track. 
A type 1 contains one or more simultaneous tracks. 
A type 2 contains a number on independent single-track patterns.

Instance variables:
	fileType <Integer> MIDI file type: 0, 1 or 2.
	tracks <Integer> number of tracks in the file.
	ppq <Integer> time division in pulses per quarter-beat.
	tempo <Integer> the tempo read in from a Midi dump file in microseconds per beat.'!


!OpcodeVoice2 methodsFor: 'accessing'!

defaultEventClass
	"What kind of Events do i normally use?"

	^NoteEvent!

tempo
	"return my default tempo"

	^tempo!

tempo: aValue
	"set my tempo"

	tempo _ aValue! !

!OpcodeVoice2 methodsFor: 'read/write'!

loadOnto: anEventList
	"load a MIDI file into anEventList."
	"For file types 1 and 2, extra EventLists are created within anEventList, one per track."
	"(OpcodeVoice2 newOn: 'Scores:n3.a.sc') loadOnto: (EventList newNamed: #n3.a).
	(EventList named: #n3.a) inspect"
	"(OpcodeVoice2 newOn: '../release/scores/n3.a.sc') loadOnto: (EventList newNamed: #n3.a).
	(EventList named: #n3.a) inspect"

	| anEL |
	self readMThd.
	[stream atEnd] whileFalse: [
		anEL _ EventList newAnonymous.
		self readTrackOnto: anEL.
		anEventList add: anEL at: 0].

	anEventList at: #tempo put: ((60000000/tempo) asInteger).		"tempo in bpm"
	Transcript show: ((anEventList events) size) printString, ' track EventLists'; cr!

readMThd
	"read the header chunk from MIDI binary file."

	| size word |
	size _ stream size.
	Transcript show: 'size: ', size printString; cr.
	word _ stream nextWord.	" 'MT' "
	(word == 16r4D54) ifFalse: [
		self error: 'MT expect in header chunk', word printString ].
	word _ stream nextWord.	" 'hd' "
	word _ stream nextWord.	"null padding"
	word _ stream nextWord.	"length -- always 6"
	fileType _ stream nextWord.	"MIDI file type (0, 1, 2)"
	Transcript show: 'MIDI file format is ', fileType printString; cr.
	tracks _ stream nextWord.	"no. tracks"
	Transcript show: 'number of tracks = ', tracks printString; cr.
	ppq _ stream nextWord.	"division (ticks per quarter beat)"
	Transcript show: 'ticks per quater beat = ', ppq printString; cr.!

readMTrk
	"read a track chunk from MIDI binary file. Answer its length."

	|length word  |
	word _ stream nextWord.	" 'MT' "
	(word == 16r4D54) ifFalse: [self error: 'MT expected in track chunk'].
	word _ stream nextWord.	" 'rk' "
	word _ stream nextWord.	"null padding"
	length _ stream nextWord.	"length"
	^length!

readTrackOnto: anEventList
	"read a track of MIDI events from my stream into anEventList"

	| fileTime trackTime deltaTime byte type vel event length startPos | 
	length _ self readMTrk.		"read track chunk header"
	startPos _ stream position.
	fileTime _ 0.
	trackTime _ 0.
	[stream position < (startPos + length)] whileTrue: [
		"Get the delta time"
		fileTime _ fileTime + self getVarLen.
		deltaTime _ (fileTime/ppq*tempo/1000 asFloat - trackTime) asInteger."convert pulses 
to ms"
		trackTime _ trackTime + deltaTime.

		"Get the event itself"
		byte _ stream next.
		((byte bitAnd: 16r80) = 0) ifFalse:
			[(byte == 16rFF) 	"a meta-event?"
				ifTrue: [type _ stream next]
				ifFalse: [type _ byte.
					(type == 16rF0) ifFalse: [byte _ stream next]] "if not sysex"
			].

		((type < 16r80) or: [type == 16rF0])
			ifTrue: ["Process a meta- or sysex event"
				type == 16r51 
					ifTrue: [
						tempo _0.
						self getVarLen timesRepeat: [tempo _ tempo * 256 + stream 
next]]
					ifFalse: [
						Transcript 
							show: 'meta or sysex, ', (type printStringRadix: 16), ' -- '.
						self getVarLen timesRepeat: [
							Transcript show: (stream next printStringRadix:16), ' '].
						Transcript cr]
			]
			ifFalse: ["Process normal MIDI event"
				(type bitAnd: 16rE0) = 16rC0 
					ifTrue: [vel _ nil]		"program change or channel pressure"
					ifFalse: [vel _ stream next].

				"use MIDI channel number as event's voice"
				event _ self defaultEventClass 
					pitch: byte ampl: vel voice: (type bitAnd: 16r0F).

				(((type bitAnd: 16rF0) = 16r90) and: [vel ~~ 0])
					ifTrue: [anEventList add: event at: trackTime].	"note-on"

				((type bitAnd: 16rF0) = 16r80 or: [vel == 0])
					ifTrue:["note-off"
						"find the corresponding note-on and set its duration"
						anEventList fixNoteOn: (trackTime -> event)]
			]
	].

	anEventList at: #bytes put: self size; 
			at: #tempo put: tempo; 
			code: #noteOnOff;
			at: #size put: (anEventList events) size.
	Transcript show: ((anEventList events) size) printString, ' events'; cr! !

!OpcodeVoice2 methodsFor: 'private'!

getVarLen
	"Combine 1 or more input bytes from my stream into a variable length item"

	| answer byte |
	answer _ 0.
	byte _ 16r80.
	[(byte bitAnd: 16r80) == 0] whileFalse: [
		byte _ stream next. 
		answer _ (answer bitShift: 7) + (byte bitAnd: 16r7F)].
	^answer! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

OpcodeVoice2 class
	instanceVariableNames: ''!


!OpcodeVoice2 class methodsFor: 'instance creation'!

newOn: fileName
	"open an new binary MIDI file for R/W."

	| me |
	me _ self new.
	me stream: ((FileStream fileNamed: fileName) binary; readOnly).
	^me! !

Model subclass: #MidiDevice
	instanceVariableNames: 'port num '
	classVariableNames: 'MStream '
	poolDictionaries: ''
	category: 'Music-Voices'!
MidiDevice comment:
'A MidiDevice is a subclass of Model used to model real Midi input/output hardware channels.
The abstract class MidiDevice implements the generic MIDI note on/off type events.
Subclasses of MidiDevice exist for specific models of devices and  implement the device-specific (system exclusive) commands.
MidiDevices are passed Events and channel numbers by their voices and generate commanddds as ByteArrays that they pass to MidiPorts.
Several devices may share one port (if there are several MIDI-capable devices on one cable), and one voice may point to several device/channel pairs on one or more devices.

Instance variables:
	port			<MidiPort> the MidiPort i use
	num			<SmallInteger> my port number

Class variable:
	MStream	<WriteStream> the stream I dump by bytes on if my port is nil (good for debugging new subclasses)

Standard MIDI Commands:
		Note On = 9n kk vv
		Note Off = 8n kk vv
		Key Pressure = An kk vv
		Pitch Wheel = En ll ll
		After Touch Channel Pressure = Dn vv
		Control Change = Bn cc vv
		Program change = Cn pp

Relevant Constants for description of MIDI commands:
		n = 4-bit channel number
		kk = 7-bit key number
		vv = 7-bit key velocity
		ll = 7-bit low-order value
		hh = 7-bit high-order value
		cc = 7-bit control number
		pp = 7-bit program number'!


!MidiDevice methodsFor: 'initialization'!

initialize: portNum
	"set me up on the given output port."

"	Transcript show: 'MidiDevice init...'; cr."
	port == nil ifTrue: [
		port _ MidiPort open: portNum ].
	port == nil ifTrue: [
		port _ (MidiPort allInstances at: 1) initializePort ].
	port setMidiMode.
"	num _ portNum.
	MStream nextPutAll: 'MidiInit: ', portNum printString; cr.
	^self
"!

reInitialize
	"restore my port when restarting me"

	port == nil ifTrue: [
		port _ MidiPort open: num ].
	port == nil ifTrue: [
		port _ (MidiPort allInstances at: 1) initializePort ].
	port setMidiOut.
	Transcript show: 'MidiPort reInit...'; cr.!

reset
	"release my hold on the port"

	Transcript show: 'MidiPort reset: ', num printString; cr.
	port _ nil! !

!MidiDevice methodsFor: 'note events'!

play: anEvent on: channelNumber
	"play the given event on my port as a default MIDI noteOn/noteOff pair"

	| dur |
	dur _ self mapDuration: anEvent dur.
"	Sensor leftShiftDown ifTrue: [ self halt ]."
	self playNoteOn: anEvent on: channelNumber.
	dur > 0 
		ifTrue: [[ (Delay forMilliseconds: dur) wait.
			self playNoteOff: anEvent on: channelNumber ] fork ].! !

!MidiDevice methodsFor: 'parameter mapping'!

mapAmplitude: aVal
	"map a numerical or symbolic amplitude to a MIDI-compatible volume number (key velocity)"


	aVal class == Amplitude ifTrue: [ ^aVal asMidi ].
	aVal class == SmallInteger ifTrue: [
		^aVal ].								"assume a volume number"
	^Amplitude value: aVal asMidi		"assume a dynamic name"!

mapDuration: aVal
	"map a numerical or symbolic duration value to a MIDI-compatible duration in mS"

	aVal class == Duration ifTrue: [ ^aVal asMS ].
	aVal class == SmallInteger ifTrue: [
		^aVal ].							"assume val is already mS"
	^Duration value: aVal asMidi	"assume a symbolic duration"!

mapPitch: aVal
	"map a numerical or symbolic pitch to a MIDI-compatible note number"

	aVal class == Pitch ifTrue: [ ^aVal asMidi ].
	aVal class == SmallInteger ifTrue: [
		^aVal ].							"assume a note number"
	^Pitch value: aVal asMidi		"assume a note name"! !

!MidiDevice methodsFor: 'backward compatibility'!

play: aNote amp: theAmp
	"play the given event--old MidiChannel-style format"

	| bytes |
	bytes _ (ByteArray new: 3).
	bytes at: 1 put: (16r90 bitOr: num);
		at: 2 put: aNote;
		at: 3 put: theAmp.
	port == nil 
		ifTrue: [ MStream nextPutAll: Time millisecondClockValue printString, 
				'  NOn: ', bytes printString; cr ]
		ifFalse: [ port send: bytes length: 3 ]!

play: aNote at: aTime for: aDur
	"play the given event--old MidiChannel-style format"

	[(Delay untilMilliseconds: aTime) wait.
	self play: aNote amp: 100] fork.

	[(Delay untilMilliseconds: aTime + aDur) wait.
	self play: aNote amp: 0] fork!

play: aNote at: aTime for: aDur amp: theAmp
	"play the given event--old MidiChannel-style format"

	[(Delay untilMilliseconds: aTime) wait.
	self play: aNote amp: theAmp] fork.

	[(Delay untilMilliseconds: aTime + aDur) wait.
	self play: aNote amp: 0] fork! !

!MidiDevice methodsFor: 'private'!

playNoteOff: anEvent on: channelNumber
	"play the note-off part of the given event on your port"

	| bytes |
	bytes _ (ByteArray new: 3).
	bytes at: 1 put: (16r90  bitOr: channelNumber);
		at: 2 put: (self mapPitch: anEvent pitch);
		at: 3 put: 16r00.
	port == nil 
		ifTrue: [ MStream nextPutAll: '	', Time millisecondClockValue printString, 
				' NOff: ', bytes printString; cr ]
		ifFalse: [ port send: bytes length: 3 ]!

playNoteOn: anEvent on: channelNumber
	"play the note-on part of the given event"

	| bytes |
	bytes _ (ByteArray new: 3).
	bytes at: 1 put: (16r90  bitOr: channelNumber);
		at: 2 put: (self mapPitch: anEvent pitch);
		at: 3 put: (self mapAmplitude: anEvent ampl).
	port == nil 
		ifTrue: [ MStream nextPutAll: Time millisecondClockValue printString, 
				'  NOn: ', bytes printString; cr ]
		ifFalse: [ port send: bytes length: 3 ]! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

MidiDevice class
	instanceVariableNames: ''!


!MidiDevice class methodsFor: 'class initialization'!

initialize
	"Set up the one class variable"
	"MidiDevice initialize"

	MStream _ "Transcript" WriteStream on: (String new: 1024).!

stream
	"return the stream of bytes i might have"
	"MidiDevice stream inspect"

	^MStream! !

!MidiDevice class methodsFor: 'instance creation'!

newOn: portNumber
	"Open a generic MidiDevice on the given Port"

	| me |
	me _ self new initialize: portNumber.
	^me! !

MidiDevice initialize!


Voice subclass: #FastMidiVoice
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Voices'!
FastMidiVoice comment:
'Instances of class FastMidiVoice use John Maloney''s MidiInterface class for fast real-time output.'!


!FastMidiVoice methodsFor: 'note events'!

play: anEvent on: channelNumber
	"play the given event on my port as a default MIDI noteOn/noteOff pair"

	| noteOff |
	MidiInterface 
		noteOn: anEvent pitch asMidi 
		vel: anEvent ampl asMidi 
		chan: anEvent voice.
	noteOff _ anEvent copy.
	noteOff ampl: 0.
	EventScheduler addAppointment: 
		((Time millisecondClockValue + anEvent duration asMS) => noteOff).! !

Voice subclass: #CmusicVoice
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Voices'!
CmusicVoice comment:
'Cmusic voices know how to print events onto streams in standard cmusic format.
The ignore their instrument associations, assuming they are being used as holders for note parameter transformers.
CmusicVoices use the property dictionaries of events for an arbitratry number of note parameters.
If events are played on them that have properties names like ''p5'' or ''p21'', they will be printed out in order.
For complex instruments, one can create new subclasses of CmusicVoice that perform special parameter translations or printout formatting.

Instance variables:
	stream			<Stream> my output stream (usually a file or the Transcript)'!


!CmusicVoice methodsFor: 'initialize-release'!

close
	"close my output stream (if it's a file)"

	(stream isKindOf: FileStream) ifTrue:
		[ stream flush. 
		stream close ]!

initializeNamed: aName onStream: theStream
	"create a new named CMusicVoice on the given stream"

	self initializeNamed: aName.
	stream _ theStream! !

!CmusicVoice methodsFor: 'events'!

playEvent: anEvent at: aTime 
	"send the given Event to all device/channel pairs in cmusic format (note start instrument dur p-fields)"

	| pFields aString |
"	Transcript show: aTime printString; cr. "
	stream == nil ifTrue: [ self halt ].
							"print out the standard fields"
	stream nextPutAll: 'note	' , (aTime / 1000) asFloat printString , '	' ,
			name , '	', anEvent duration printString.
							"select the relevant p-fields (i.e., properties with names like p5 or p21)"
	anEvent properties == nil ifFalse:
		[ pFields _ anEvent properties keys select: [:aSymbol | 
			aString _ aSymbol asString.
			(aString at: 1 = $p)
				and: [(Number readFrom: (aString copyFrom: 2 to: aString size)) !!= 0]].
							"and now the extra p-fields"
		pFields asSortedCollection do: [ :aProperty | 
			stream nextPutAll: '	', (anEvent perform: aProperty printString)]].
	stream nextPut: $;; cr! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

CmusicVoice class
	instanceVariableNames: ''!


!CmusicVoice class methodsFor: 'instance creation'!

newNamed: aName onStream: aStream
	"set up a CmusicVoice on the given output stream"
	"CmusicVoice newNamed: 'violin' onStream: Transcript"

	^self new initializeNamed: aName onStream: aStream! !

Voice subclass: #OpcodeVoice1
	instanceVariableNames: 'tempo time command byte2 byte3 eventList '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Voices'!
OpcodeVoice1 comment:
'This voice class is a file I/O convertor that can transform **OLD** Opcode format MIDI file dumps into EventLists and/or play them in real-time. 
The messages here will look like:
(OpCodeVoice newOn: ''Scores:n3.a.sc'') loadOnto: (EventList newNamed: #n3.a)

Instance variables:
	tempo <Integer> the tempo read in from a Midi dump file
	time <Integer> the current time
	command <Byte> first byte of the command
	byte2 <Byte> second byte of the command
	byte3 <Byte> third byte of the command
	eventList <EventList> the resulting event list'!


!OpcodeVoice1 methodsFor: 'accessing'!

command
	"return the command byte of the current event"

	^command!

command: aValue
	"set the command byte"

	command _ aValue!

defaultEventClass
	"What kind of Events do i normally use?"

	^NoteEvent!

eventList: aValue
	"plug in an event list"

	eventList _ aValue!

tempo
	"return my default tempo"

	^tempo!

tempo: aValue
	"set my tempo"

	tempo _ aValue! !

!OpcodeVoice1 methodsFor: 'read/write'!

loadOnto: anEventList
	"load a MIDI file into an EventList."
	"(MidiStream newOn: 'Scores:n3.a.sc') loadOnto: (EventList newNamed: #n3.a)
	(EventList named: #n3.a) setDurations
	(EventList named: #n3.a) inspect"

	| dict t temp event |
	self readHeader.
	anEventList at: #bytes put: self size; 
			at: #tempo put: tempo; 
			code: #noteOnOff.
	t _ 0.
	temp _ ((24 * 60) / (tempo / 128)) asInteger. 
	[stream atEnd] whileFalse:
		[self readEvent.
		time _ time * temp.
		t _ t + time.			"sum MIDI delta times"
		(command bitAnd: 16rF0) = 16rC0 ifFalse:
			[ event _ self defaultEventClass 
					pitch: byte2 ampl: byte3 voice: command.
			 anEventList add: event at: t]].
	anEventList  at: #size put: (anEventList events) size.
	Transcript show: ((anEventList events) size) printString, ' events'; cr!

printEventOn: aFile
	"print out 1 MIDI event by its instance variables"

	| str |
	str _ WriteStream on: (String new: 12).
	time printOn: str base: 16.  str nextPut: $	.
	command printOn: str base: 16.  str nextPut: $	.
	byte2 printOn: str base: 16.  str nextPut: $	.
	command > 127 ifTrue:
		[byte3 printOn: str base: 16.  str nextPut: $	.
	 ].
	aFile nextPutAll: (str contents); nextPut: Character cr!

readEvent
	"read 1 MIDI event into instance variables, parse out running events"
	"unbuffered version."

	| byte |
	time _ stream next.
	time > 127 ifTrue:   "overlength encoded times"
		[ time _ (time - 127) bitShift: 8.
		time _ time + (self next) ].
	byte _ stream next.
	byte > 127 ifTrue:  	"real event"
		[ command _ byte.
		byte2 _ stream next.
		(command bitAnd: 16rF0) = 16rC0 
			ifTrue: [ byte3 _ nil ]
			ifFalse: [  "not a command code"
				byte3 _ stream next ].
		(command bitAnd: 16rFF) = 16rFF ifTrue: 
			[ self readMetaEvent ] ]
	ifFalse: 			"not a real event, running status"
		[ byte2 _ byte.
		byte3 _ stream next ]!

readHeader
	"strip MIDI header from binary file."

	| size word |
	size _ stream size.
	Transcript show: 'size: ', size printString. 
	word _ stream nextWord.
	(word = 16rFF) ifTrue: [
		byte2 _ stream next.
		byte3 _ stream next.
		self readMetaEvent ]
	ifFalse: [
		self error: 'unknown header: ', word printString ]!

readMetaEvent
	"parse a MIDI meta event - tempo or end-of-track"

	| word |
	word _ (byte2 bitShift: 8) bitOr: byte3.
	(word = 16r5002) ifTrue: [  "tempo meta event"
		tempo _ stream nextWord.
		Transcript show: '  	tempo: ', tempo printString; cr ]
	ifFalse: [ (word = 16r2F00) ifTrue: [  "end of track"
		Transcript show: 'end of track  	' ]
		ifFalse: [
			self error: 'unknown meta event', word printString ] ]! !

!OpcodeVoice1 methodsFor: 'play/record'!

play
	"play MIDI data from a file starting now"
	"Use like (OpCodeVoice fileNamed: 'Scores:piece1') play."

	self checkMidiHeader.
	self openData.
	eventList _ self loadOnto: (EventList new)
	eventList schedule.
	eventList play! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

OpcodeVoice1 class
	instanceVariableNames: ''!


!OpcodeVoice1 class methodsFor: 'instance creation'!

newOn: fileName
	"open an new binary MIDI file for R/W."

	| me |
	me _ self new.
	me stream: ((FileStream fileNamed: fileName) binary; readOnly).
	^me! !

Voice subclass: #PrimMidiVoice
	instanceVariableNames: 'clockOffset '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Voices'!
PrimMidiVoice comment:
'Instances of this class are voices that use Guy Garnett and Lee Boynton''s primitives for Midi output on Macintosh computers.
They send events to the class Midi for real-time output.
See the example workspace for their usage.

Instance variables:
	clockOffset <SmallInteger>	the midi driver''s clock offset, used in calculatting real-time'!


!PrimMidiVoice methodsFor: 'initialize release'!

initialize
	"set up the primitive driver's voice"

	Midi setTickSize: 1.					"1 ms resolution"
	clockOffset _ Midi getClockTime		"get the driver's offset"! !

!PrimMidiVoice methodsFor: 'events'!

playEvent: anEvent at: aTime
	"send the given Event to the global output - midi assumes"

	| realTime | 
	realTime _ ((aTime - Time millisecondClockValue) max: 0) + clockOffset.
"	Transcript show: realTime printString, ' '."
	^Midi
		play: anEvent pitch asMidi
		at: realTime
		dur: anEvent duration asMS
		amp: anEvent ampl asMidi
		voice: anEvent voice! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

PrimMidiVoice class
	instanceVariableNames: ''!


!PrimMidiVoice class methodsFor: 'instance creation'!

new
	^super new initialize! !

MidiDevice subclass: #MidiFB01
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Voices'!
MidiFB01 comment:
'MidiFB01 is a type of MidiDevice for the system exclusive commands of the Yamaha FB-01 synthesiser.
It implements the special commands available here (like note commands with fractional pitch and given duration).

The standard FB01 note command is like:
		2n k f v d1 d2 = cmd:channel, noteNum, frac, vel, durLow, durHigh'!


!MidiFB01 methodsFor: 'note events'!

play: anEvent on: aChannel
	"send FB01-specific code for note with fraction and duration
		2n k f v d1 d2 = cmd:channel, noteNum, frac, vel, durLow, durHigh"

	| bytes realDur |
	bytes _ ByteArray new: 6.
	bytes at: 1 put: (16r20 + aChannel).
	bytes at: 2 put: (self mapPitch: anEvent pitch).
	bytes at: 3 put: 0.
	bytes at: 4 put: (self mapAmplitude: anEvent ampl).
	realDur _ self mapDuration: anEvent duration.
	bytes at: 5 put: (realDur bitAnd: 127).
	bytes at: 6 put: (realDur bitShift: -7 ).

	port == nil
		ifTrue: [ MStream nextPutAll: 'FB01: ', bytes printString; cr ]
		ifFalse: [ port send: bytes length: 6 ]! !