/*
MidiPrimDriver.c
Last change: January 14, 1988.

Copyright 1989 John H. Maloney. All rights reserved.

This module implements a low-overhead interface to the Macintosh
serial ports for MIDI input and output. The entry points are:

	MidiOpen	- initializes a serial port for MIDI including
				  configuring the SCC chip, allocating input and output
				  buffers, and initializing the interrupt vectors
	MidiClose	- disables interrupts and frees buffer
	MidiPutByte	- queues one byte of data for transmission
	MidiPut		- queues many bytes for transmission
	MidiGet		- copies data from the input queue into a client buffer

Ports:
The modem port is port 0 and the printer port is port 1.

Sending:
A call to MidiPutByte places a byte in the output queue. If the output
driver is not busy, transmission of the byte is initiated immediately.
If the output driver *is* busy, then an SCC "transmit buffer empty"
interrupt (or "transmit interrupt") will occur when the hardware has
finished sending the current byte. This interrupt is used to start
transmission of the next byte in the queue; eventually this will
cause every byte in the queue to be transmitted.

When the queue is empty, no further transmit interrupts will occur until
another call to MidiPutByte initiates another round of transmissions. If
the queue is full, MidiPutByte will busy-wait until the queue has room
before inserting the new byte. By definition, if the output queue is full,
there will eventually be a transmit interrupt which will grab the
next byte for tranmission, making space to queue the new byte.

Receiving:
An interrupt occurs whenever a byte is received. The received byte is
placed in the input queue if there is room or thrown away if there isn't.
In the latter case, a flag is set in the error word so that higher levels
of software may detect the "input overrun" and report it to the user.
It is the responsibility of the client program to periodically collect
the bytes that have collected in the input queue by calling MidiGet.

Interrupts:
Care should be taken to avoid overly long computations at interrupt time.
Furthermore, one must not call any Toolbox routines that cause memory
to be allocated or freed at interrupt time. See "Inside Macintosh" for
further details on what is and is not allowed during interrupt processing.
The relevant chapters are "The Device Manager" and "The Serial Drivers",
both in Volume II.
*/

/* define "MPW" to compile with MPW C */
#ifdef MPW
#include <OSUtils.h>
#include <Strings.h>
extern void SetUpA5();
extern void RestoreA5();
#else
#include <OSUtil.h>
#include <Proto.h>
#endif

#include "cext.h"
#include "MidiPrim.h"

/* Base addresses for the SCC read and write registers. */
#define SCCRBase		0x1D8L
#define SCCWBase		0x1DCL

/* Offsets for the SCC control and data registers (in bytes). */
#define port1Ctl		0
#define port0Ctl		2
#define port1Data		4
#define port0Data		6

/* Base address for the dispatch table for Level 2 interrupts. */
#define Vectors			0x1B2L

/* Offsets into the Level 2 interrupt dispatch table (in bytes). */
#define Xmit1Vector		 0
#define Rcv1Vector		 8
#define RcvErr1Vector	12
#define Xmit0Vector		16
#define Rcv0Vector		24
#define RcvErr0Vector	28

/* The MIDIPorts (global, in case anyone wants to look at them) */
public MIDIPort port0 = {0, 0, false, NULL, NULL};
public MIDIPort port1 = {0, 0, false, NULL, NULL};

/* Pre-computed pointers to the control and data registers. These
 * be initialized the first time MidiOpen is called. */
private byte *port0ControlReg = NULL;
private byte *port0ReadReg = NULL;
private byte *port0WriteReg = NULL;
private byte *port1ControlReg = NULL;
private byte *port1ReadReg = NULL;
private byte *port1WriteReg = NULL;

/* Private Procs */
void	WriteSCCReg(short, byte, byte);
void	Pause(void);
void	Xmit0ISR(void);
void	Rcv0ISR(void);
void	RcvErr0ISR(void);
void	Xmit1ISR(void);
void	Rcv1ISR(void);
void	RcvErr1ISR(void);

/* MidiOpen - Open the given midi port
 * NOTE: Most MIDI interface suppliers now use 1 MHz as the standard rate
 * for the external clock to be supplied to the SCC chip. However, some
 * older MIDI interfaces use different rates. When you set up the SCC,
 * you must supply a divide constant to set the clock rate, as follows:
 * 	0x40 for 500 kHz,
 * 	0x80 for 1 MHz, or
 * 	0xC0 for 2 MHz.
 */
public void MidiOpen(port)
	int port;
{
	register MIDIPort *portData;
	int clockDivide = 0x80;
		/*	use 0x40 for 500 kHz, 0x80 for 1 MHz, or 0xC0 for 2 MHz */

	port0ControlReg = (*(byte **)SCCWBase) + port0Ctl;
	port0ReadReg = (*(byte **)SCCRBase) + port0Data;
	port0WriteReg = (*(byte **)SCCWBase) + port0Data;
	port1ControlReg = (*(byte **)SCCWBase) + port1Ctl;
	port1ReadReg = (*(byte **)SCCRBase) + port1Data;
	port1WriteReg = (*(byte **)SCCWBase) + port1Data;

	portData = (port == 0) ? &port0 : &port1;
	if (portData->in != NULL) MidiClose(port);	/* port was open! */
	portData->errors = 0;
	portData->inputOverruns = 0;
	portData->sending = false;
	portData->in = MakeQueue(IN_BUF_SIZE);
	portData->out = MakeQueue(OUT_BUF_SIZE);

	if (port == 0) {
		WriteSCCReg(0, 9, 0x82);	/* reset port 0 */
		*((ulong *)(Vectors + Xmit0Vector)) = (ulong) Xmit0ISR;
		*((ulong *)(Vectors + Rcv0Vector)) = (ulong) Rcv0ISR;
		*((ulong *)(Vectors + RcvErr0Vector)) = (ulong) RcvErr0ISR;
	} else {
		WriteSCCReg(0, 9, 0x42);	/* reset port 1 */
		*((ulong *)(Vectors + Xmit1Vector)) = (ulong) Xmit1ISR;
		*((ulong *)(Vectors + Rcv1Vector)) = (ulong) Rcv1ISR;
		*((ulong *)(Vectors + RcvErr1Vector)) = (ulong) RcvErr1ISR;
	}
	WriteSCCReg(port, 11, 0x28);	/* use external clock */
	WriteSCCReg(port,  4, clockDivide + 4);	/* clock divider, 1 stop bit */
	WriteSCCReg(port,  3, 0xC1);	/* receive enabled (8 data bits) */
	WriteSCCReg(port,  5, 0x6A);	/* transmit enabled (8 data bits), rts */
	WriteSCCReg(port,  1, 0x13);	/* rcv, xmit, and external ints enabled */
	WriteSCCReg(port, 15, 0x08);	/* all ext/stat but DCD disabled */
	WriteSCCReg(port,  0, 0x10);	/* reset ext/stat interrupts */
	WriteSCCReg(0,     9, 0x0A);	/* master interrupt enable, no vector */
}

/* MidiClose - Close the given midi port */
public void MidiClose(port)
	int port;
{
	/* disable rcv and xmit interrupts but allow mouse interrupts */
	WriteSCCReg(port, 1, 1);
	/* deallocate buffers */
	if (port == 0) {
		if (port0.in == NULL) return;	/* port is closed! */
		DisposPtr(port0.in);
		DisposPtr(port0.out);
		port0.in = NULL;
		port0.out = NULL;
	} else {
		if (port1.in == NULL) return;	/* port is closed! */
		DisposPtr(port1.in);
		DisposPtr(port1.out);
		port1.in = NULL;
		port1.out = NULL;
	}
}

/* MidiPutByte - Send the given data byte to the given midi port */
public void MidiPutByte(port, dataByte)
	int port;
	byte dataByte;
{
	register MIDIPort *portData;
	register ByteQueue *outQueue;

	portData = (port == 0) ? &port0 : &port1;
	outQueue = portData->out;
	if (outQueue == NULL) return;	/* port is closed! */

	while (QFull(outQueue)) {
		/* busy wait until there is room in the buffer */
	}
	QInsert(outQueue, dataByte);
	if (!portData->sending) {
		portData->sending = true;
		if (port == 0) {
			QRemove(outQueue, *port0WriteReg);
		} else {
			QRemove(outQueue, *port1WriteReg);
		}
	}
}

/* MidiPut - Send the "count" data bytes from the client supplied buffer
 * to the given midi port.
 */
public void MidiPut(port, buffer, count)
	int port;
	byte *buffer;
	int count;
{
	register MIDIPort *portData;
	register ByteQueue *outQueue;
	register byte *portWriteReg;
	register byte *next, *end;

	if (port == 0) {
		portData = &port0;
		portWriteReg = port0WriteReg;
	} else {
		portData = &port1;
		portWriteReg = port1WriteReg;
	}
	outQueue = portData->out;
	if (outQueue == NULL) return;	/* port is closed! */

	for (next = &buffer[0], end = &buffer[count]; next < end; ) {
		while (QFull(outQueue)) {
			/* busy wait until there is room in the buffer */
		}
		QInsert(outQueue, *next++);
		if (!portData->sending) {
			portData->sending = true;
			QRemove(outQueue, *portWriteReg);
		}
	}
}

/* MidiGet - Copy data bytes from the given midi port's input queue
 * into the client supplied buffer of size "count" bytes.
 */
public int MidiGet(port, buffer, count)
	int port;
	byte *buffer;
	int count;
{
	ByteQueue *inQueue = (port == 0) ? port0.in : port1.in;
	int copied = 0;

	if (inQueue == NULL) {
		return copied;	/* port is closed! */
	} else {
		while (true) {
			if (QEmpty(inQueue) || (copied >= count)) {
				return copied;
			} else {
				QRemove(inQueue, *(buffer++));
				copied++;
			}
		}
	}
}

/* MakeQueue - Create a queue of the given size.
 * Note: a queue of size N may hold up to N-1 bytes of data.
 * ESSENTIAL: size *must* be an even power of 2.
 */
public ByteQueue * MakeQueue(size)
	int size;
{
	register ByteQueue *newQueue;

	newQueue = (ByteQueue *) NewPtr((ulong) (sizeof(ByteQueue) + size));
	if (newQueue == NULL) {
		DebugStr("\pCould not allocate a MIDI buffer");
	}
	newQueue->mask = size - 1;
	newQueue->head = 0;
	newQueue->tail = 0;
	return newQueue;
}

/* Insert - Add the given data byte to the end of the given queue
 * ASSUMPTION: The caller has already ensured there is room in the queue.
 */
public void Insert(queue, dataByte)
	register ByteQueue *queue;
	byte dataByte;
{
	queue->data[queue->tail++] = dataByte;
	queue->tail &= queue->mask;		/* mod the queue size */
}

/* Remove - Remove a data byte from the given queue and return it
 * ASSUMPTION: The caller has already ensured there is data in the queue.
 */
public byte Remove(queue)
	register ByteQueue *queue;
{
	byte dataByte;

	dataByte = queue->data[queue->head++];
	queue->head &= queue->mask;		/* mod the queue size */
	return dataByte;
}

/* WriteSCCReg - write a byte to the given register of the SCC for
 * the given port. One must be careful not the access the SCC more
 * often than once every 2.2 microseconds to give its internal lines
 * time to settle. Noops have been inserted to guarantee that timing
 * for 68020 processors with clock speeds up to 25 MHz.
 */
private void WriteSCCReg(port, sccRegister, dataByte)
	short port;
	register byte sccRegister;
	register byte dataByte;
{
	if (port == 0) {
		*port0ControlReg = sccRegister;
		Pause();
		*port0ControlReg = dataByte;
	} else {
		*port1ControlReg = sccRegister;
		Pause();
		*port1ControlReg = dataByte;
	}
	Pause();
}

/* Pause - Must onsume at least 2.2 microseconds on a 68000 or 68020
 * running at up to 25 MHz, even with zero wait states and caching.
 */
private void Pause()
{
	long junk, *junkPtr1, *junkPtr2;

	junkPtr1 = junkPtr2 = &junk;
	*junkPtr1 = *junkPtr2;
	*junkPtr2 = *junkPtr1;
	/* add more if necessary... */
}

/* ================ Interrupt Service Routines ================ */
void Xmit0ISR()
{
	SetUpA5();
	if (QEmpty(port0.out)) {
		/* clear the "transmit buffer empty" interrupt pending flag */
		WriteSCCReg(0, 0, 0x28);
		port0.sending = false;
	} else {
		/* send the next character */
		QRemove(port0.out, *port0WriteReg);
	}
	RestoreA5();
}

private void Xmit1ISR()
{
	SetUpA5();
	if (QEmpty(port1.out)) {
		/* clear the "transmit buffer empty" interrupt pending flag */
		WriteSCCReg(1, 0, 0x28);
		port1.sending = false;
	} else {
		/* send the next character */
		QRemove(port1.out, *port1WriteReg);
	}
	RestoreA5();
}

private void Rcv0ISR()
{
	register byte dataByte;

	SetUpA5();
	dataByte = *port0ReadReg;	/* read data byte */
	if (QFull(port0.in)) {
		port0.inputOverruns++;
	} else {
		/* put byte into input queue */
		QInsert(port0.in, dataByte);
	}
	RestoreA5();
}

private void Rcv1ISR()
{
	register byte dataByte;

	SetUpA5();
	dataByte = *port1ReadReg;	/* read data byte */
	if (QFull(port1.in)) {
		port1.inputOverruns++;
	} else {
		/* put byte into input queue */
		QInsert(port1.in, dataByte);
	}
	RestoreA5();
}

private void RcvErr0ISR()
{
	register byte junk;

	SetUpA5();
	junk = *port0ReadReg;		/* read error byte */
	*port0ControlReg = 0x30;	/* clear the error */
	port0.errors++;				/* record the error */
	RestoreA5();
}

private void RcvErr1ISR()
{
	register byte junk;

	SetUpA5();
	junk = *port1ReadReg;		/* read error byte */
	*port1ControlReg = 0x30;	/* clear the error */
	port1.errors++;				/* record the error */
	RestoreA5();
}
