/*
MidiPrimTests.c
Last change: January 14, 1988.

Copyright 1989 John H. Maloney. All rights reserved.

This module tests the module MidiPrimDriver.c.
Tests that may be run include:
	
	QueueTest - Tests the ByteQueue functions
	SendTest - Sends to the Midi port
	ReceiveTest - Collects incoming Midi data for 10 seconds
	PlayTest - Plays an octave of the chromatic scale
	LoopBackTest - Tests both sending and receiving. To run this
		test you should connect the Midi Out to the Midi In on your
		Midi interface; the bytes transferred should be 10000.

The modem port is port 0 and the printer port is port 1.

Compile with the "-dMPW" switch to compile using MPW C.
Compile normally using Lightspeed C 2.15. This test program will
run under Lightspeed C even though the primitive itself must be
installed using MPW C 2.0.

******************** MPW 2.0 Build Script *******************
Asm SetUpA5.a
C -dMPW MidiPrimDriver.c
C -dMPW MidiPrimTests.c
Link -d -t MPST -c 'MPS ' 
	SetUpA5.a.o 
	MidiPrimDriver.c.o 
	MidiPrimTests.c.o 
	"{Libraries}"Runtime.o 
	"{CLibraries}"CInterface.o 
	"{CLibraries}"CRuntime.o 
	"{CLibraries}"StdCLib.o 
	"{CLibraries}"CSANELib.o 
		-o MidiTests
************************************************************/

/* define "MPW" to compile with MPW C */
#ifdef MPW
#include <StdIO.h>
#include <Events.h>
#include <OSUtils.h>
#else
#include <Proto.h>
#endif

#include "cext.h"
#include "MidiPrim.h"

/* Import the port data structures */
extern MIDIPort port0;
extern MIDIPort port1;

#define TEST_PORT 0	/* 0 for the modem port, 1 for the printer port */

void ShowQueue(ByteQueue *);
void ShowQueue(q)
	ByteQueue *q;
{
	int i;
	printf("size = %d, tail = %d, head = %d [",
		q->mask + 1, q->tail, q->head);
	for (i = 0; i < 8; i++) {
		printf(" %d", q->data[i]);
	}
	printf(" ] count = %d %s%s\n",
		QCount(q),
		QFull(q) ? "FULL" : "",
		QEmpty(q) ? "EMPTY" : "");
}

void QueueTest(void);
void QueueTest()
{
	ByteQueue *q;
	int i;
	byte junk;

	q = MakeQueue(8);
	QInsert(q, 0);
	QInsert(q, 0);
	QInsert(q, 0);
	QRemove(q, junk);
	QRemove(q, junk);
	QRemove(q, junk);
	for (i = 1; i <= 12; i = i + 2) {
		ShowQueue(q);
		QInsert(q, i);
		ShowQueue(q);
		QInsert(q, i + 1);
		ShowQueue(q);
		QRemove(q, junk);
	}
	printf("\n");
	ShowQueue(q);
	QRemove(q, junk);
	ShowQueue(q);
	QRemove(q, junk);
	ShowQueue(q);
	QRemove(q, junk);
	ShowQueue(q);
}

void SendTest(void);
void SendTest()
{
	MIDIPort *port = (TEST_PORT == 0) ? &port0 : & port1;
	ushort errors, overruns;
	long i;

	printf("Sending 10000 bytes...");
	fflush(stdout);
	MidiOpen(TEST_PORT);
	for (i = 10000; i > 0; i--) {
		MidiPutByte(TEST_PORT, 0xF8);	/* MIDI clock */
	}
	while (!QEmpty(port->out)) {
		/* busy wait until all bytes have been sent */
	}
	errors = port->errors;
	overruns = port->inputOverruns;
	MidiClose(TEST_PORT);
	printf("Done\n");
	printf("errors = %d, input overruns = %d\n", errors, overruns);
}

void ReceiveTest(void);
void ReceiveTest()
{
	MIDIPort *port = (TEST_PORT == 0) ? &port0 : & port1;
	ushort errors, overruns;
	long ticksToListen = 600;
	long startTicks, ticks;
	byte buf[100];
	int receiveCount = 0;

	printf("Receiving Midi data for 10 seconds...");
	fflush(stdout);
	MidiOpen(TEST_PORT);
	startTicks = TickCount();
	ticks = 0;
	for (ticks = 0;
		 ticks < ticksToListen;
		 ticks = TickCount() - startTicks) {
			receiveCount += MidiGet(TEST_PORT, &buf[0], 100);
	}
	errors = port->errors;
	overruns = port->inputOverruns;
	MidiClose(TEST_PORT);
	printf("Done\n");
	printf("Received %d bytes\n", receiveCount);
	printf("errors = %d, input overruns = %d\n", errors, overruns);
}

void PlayTest(void);
void PlayTest()
{
	MIDIPort *port = (TEST_PORT == 0) ? &port0 : & port1;
	ushort errors, overruns;
	static byte noteOn[4] = {144, 0, 75, 0};
	static byte noteOff[4] = {144, 0, 0, 0};
	int i;
	long junk;

	printf("Playing...");
	fflush(stdout);
	MidiOpen(TEST_PORT);
	for (i = 48; i <= 72; i++) {
		noteOn[1] = noteOff[1] = i;
		MidiPut(TEST_PORT, &noteOn[0], 3);
		Delay(4L, &junk);
		MidiPut(TEST_PORT, &noteOff[0], 3);
		Delay(1L, &junk);
	}
	errors = port->errors;
	overruns = port->inputOverruns;
	MidiClose(TEST_PORT);
	printf("Done\n");
	printf("errors = %d, input overruns = %d\n", errors, overruns);
}

void LoopBackTest(void);
void LoopBackTest()
{
	MIDIPort *port = (TEST_PORT == 0) ? &port0 : & port1;
	ushort errors, overruns;
	long count = 100;
	long i, startTicks, time, baseTime;
	byte buf[100];
	int transferCount = 0;
	long junk;

	printf("LoopBack Test: You should have MIDI IN connected to MIDI OUT.\n");
	printf("LoopBack of 10000 bytes...");
	fflush(stdout);
	MidiOpen(TEST_PORT);
	startTicks = TickCount();
	for (i = count; i > 0; i--) { }
	baseTime = TickCount() - startTicks;
	startTicks = TickCount();
	for (i = count; i > 0; i--) {
		MidiPut(TEST_PORT, &buf[0], 100);
		transferCount += MidiGet(TEST_PORT, &buf[0], 100);
	}
	while (!QEmpty(port->in) || !QEmpty(port->out)) {
		transferCount += MidiGet(TEST_PORT, &buf[0], 100);
	}
	Delay(1L, &junk);	/* let port settle down, then get stragglers */
	transferCount += MidiGet(TEST_PORT, &buf[0], 100);
	time = ((TickCount() - startTicks) - baseTime) - 1;
	errors = port->errors;
	overruns = port->inputOverruns;
	MidiClose(TEST_PORT);
	printf("Done\n");
	printf("Transfered %d bytes", transferCount);
	printf(" in %ld ticks = %f seconds\n", time, (float) time / 60.0);
	printf("errors = %d, input overruns = %d\n", errors, overruns);
}

void main(void);
void main()
{
	QueueTest();
	PlayTest();
	SendTest();
	ReceiveTest();
	LoopBackTest();
}
