Model subclass: #Value
	instanceVariableNames: 'value '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Numeric-Values'!
Value comment:
'Value is the abstract superclass of most value models.

Instance Variable:
	value <String or Number> - my value'!


!Value methodsFor: 'initialize-release'!

initialize
	"Initialize instance variables"

	^self! !

!Value methodsFor: 'access'!

value
	"Answer the value the receiver holds on to."

	^value!

value: aNewValue
	"Change the value the receiver holds on to."

	value _ aNewValue.
	self changed!

valueAsString
	"Answer a string representation of the value the receiver holds on to."

	^value printString!

valueAsText
	"Answer a text representation of the value the receiver holds on to."

	^value printString asText!

valueFromString: aString
	"Change the value the receiver holds on to by evaluating aString"

	^self value: (Object readFrom: aString readStream)!

valueFromText: aText
	"Change the value the receiver holds on to by evaluating
	the string form aText "

	^self valueFromString: aText string! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Value class
	instanceVariableNames: ''!


!Value class methodsFor: 'instance creation'!

new
	"Answer an initialized instance of the receiver."

	^super new initialize! !

Value subclass: #NumberValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Numeric-Values'!
NumberValue comment:
'I am the superclass of BoundedValues and IncrementingValues and provide them with the message to read a number from a string.'!


!NumberValue methodsFor: 'initialize-release'!

initialize
	"Initialize instance variables"

	super initialize.
	value _ 0! !

!NumberValue methodsFor: 'access'!

decrement
	"Subtract one from the receivers value."
	
	^self value: value - 1!

increment
	"Add one to the receivers value."

	^self value: value + 1!

valueFromString: aString
	"Change the value the receiver holds on to by evaluating aString"

	^self value: (Number readFrom: aString readStream)! !

NumberValue subclass: #BoundedValue
	instanceVariableNames: 'min max '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Numeric-Values'!
BoundedValue comment:
'This is a subclass of NumberValue that can be used as the model of an interval or range.
We are useful as the model of sliders and gague views.

Instance Variables:
	min <Float> - my minimum value
	max <Float> - my maximum value
'!


!BoundedValue methodsFor: 'initialize-release'!

initialize
	"Initialize instance variables"

	super initialize.
	min _ 0.0.
	max _ 100.0! !

!BoundedValue methodsFor: 'access'!

max
	"Answer the maximun value the receiver may hold."

	^max!

max: aNumber
	"Set the maximun value the receiver may hold."

	max _ aNumber asFloat.
	self changed: #max.
	self changed: #percent!

min
	"Answer the minimum value the receiver may hold."

	^min!

min: aNumber
	"Set the minimum value the receiver may hold."

	min _ aNumber asFloat.
	self changed: #min.
	self changed: #percent!

percent
	"Answer the percent the receiver holds relative to the 
	min and max values."

	^(value asFloat - min) / (max - min)!

percent: x
	"Answer the value the receiver holds relative to the percent x and
	the min and max values."

	self value: (((max - min)* x) + min)!

value: aNumber
	"Change the value the receiver holds on to relative to the
	receivers min and max values."

	aNumber asFloat < min ifTrue: [value _ min]
	ifFalse: [aNumber asFloat > max ifTrue: [value _ max]
	ifFalse: [value _ aNumber]].
	self changed! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

BoundedValue class
	instanceVariableNames: ''!


!BoundedValue class methodsFor: 'instance creation'!

min: aMin max: aMax value: aValue
	"Answer an instance of the receiver with the given min and max values."

	|av|
	av _ super new.
	av min: aMin.
	av max: aMax.
	av value: aValue.
	^av! !