Magnitude subclass: #MusicMagnitude
	instanceVariableNames: 'value '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Magnitudes'!
MusicMagnitude comment:
'MusicValue is a place for the music magnitudes to share their state and basic accessing behavior.
Instance variables:
	value		<Number or Symbol> the magnitude''s value
	type 		<Symbol> the type of the value, normally SmallInteger, Float, 
					Fraction or Symbol, chached for efficiency
'!


!MusicMagnitude methodsFor: 'accessing'!

realValue: theValue
	"plug in the given value"

	value _ theValue!

value
	"Return the instance variable 'value'."

	^value! !

!MusicMagnitude methodsFor: 'arithmetic'!

* aValue
	"multiply me by the given value"

	aValue class == self class 
		ifTrue: [ ^self mostGeneral * aValue mostGeneral ]
		ifFalse: [ ^self mostGeneral * aValue ]!

+ aValue
	"add the given value to me"

	aValue class == self class 
		ifTrue: [ ^self mostGeneral + aValue mostGeneral ]
		ifFalse: [ ^self mostGeneral + aValue ]!

- aValue
	"subtract the given value from me"

	aValue class == self class 
		ifTrue: [ ^self mostGeneral - aValue mostGeneral ]
		ifFalse: [ ^self mostGeneral - aValue ]!

/ aValue
	"divide me by the given value"

	aValue class == self class 
		ifTrue: [ ^self mostGeneral / aValue mostGeneral ]
		ifFalse: [ ^self mostGeneral / aValue ]!

= aValue
	"add the given value to me"

	aValue class == self class 
		ifTrue: [ ^self mostGeneral = aValue mostGeneral ]
		ifFalse: [ ^self mostGeneral = aValue ]!

generality
	"Answer the number representing the ordering of the receiver in the
	generality hierarchy."

	^10! !

!MusicMagnitude methodsFor: 'comparing'!

< aMM

	^value < aMM! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

MusicMagnitude class
	instanceVariableNames: ''!


!MusicMagnitude class methodsFor: 'instance creation'!

value: aValue
	"create and set up a new instance"

	| me | 
	me _ self new value: aValue.
	(aValue isMemberOf: me class)
			ifTrue: [ ^aValue ].
	^me! !

MusicMagnitude subclass: #Pitch
	instanceVariableNames: ''
	classVariableNames: 'NameMap '
	poolDictionaries: ''
	category: 'Music-Magnitudes'!
Pitch comment:
'Pitches are used for the pitch or frequency parameters of musical events.
Pitches generally come in several flavors, as signified by the instance variable "type".
This is a pitch notation, NOT a real deep representation.  A more interesting model, such as that used by Lounette Dyer in her 1984 ICMC paper, would use these magnitudes as elements in pitch sets that also include information about key signature or tonal center.

Class variable:
	NameMap	<Dictionary of string -> integer> key names mapped onto a-based integers

Examples:
	Pitch value: 440.0			"create an instance with units of Hertz"
	Pitch value: 77				"create an instance with units of Midi key numbers"
	Pitch value: #e4			"create a symbolic instance"
	Pitch value: ''e4''			"same as using a symbol"

Note that new pitch representations such as music11-like pch (4.11 = 11th note in oct 4) or oct (4.1100 = oct4 + 1100 cts) notations can be added by overriding the float to Hz or float-to-midi conversions.'!


!Pitch methodsFor: 'accessing'!

realValue: theValue
	"plug in the given value"

	value _ theValue!

value: theValue
	"Accept the argument, 'theValue', as the new instance variable 'value';."

	theValue class == Float					"assume frequency"
			ifTrue: [^HertzPitch new realValue: theValue].
	theValue class == SmallInteger			"assume key number"
			ifTrue: [^MidiPitch new realValue: theValue].
	(theValue isKindOf: String) 		"assume note name string"
			ifTrue: [^SymbolicPitch new realValue: theValue].
	self error: 'unknown Pitch type.'! !

!Pitch methodsFor: 'converting'!

mostGeneral
	"what's the most general type of me"

	^self asMidi!

name
	"return my symbolic name"

	^self asSymbol! !

!Pitch methodsFor: 'printing'!

storeOn: aStream
	"print me as a Pitch declaration"

	aStream nextPutAll: '(Pitch value: ', self value printString, ')'.! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Pitch class
	instanceVariableNames: ''!


!Pitch class methodsFor: 'class initialization'!

initialize
	"set up the class name mapping dictionary"
	"Pitch initialize"

	NameMap _ Dictionary new.
	NameMap 
			at: 'a#' put: 1;
			at: 'bb' put: 1;
			at: 'b' put: 2;
			at: 'cb' put: 2;
			at: 'c' put: 3;
			at: 'c#' put: 4;
			at: 'db' put: 4;
			at: 'd' put: 5;
			at: 'd#' put: 6;
			at: 'eb' put: 6;
			at: 'e' put: 7;
			at: 'fb' put: 7;
			at: 'f' put: 8;
			at: 'f#' put: 9;
			at: 'gb' put: 9;
			at: 'g' put: 10;
			at: 'g#' put: 11;
			at: 'ab' put: 11;
			at: 'a' put: 12.! !

!Pitch class methodsFor: 'map access'!

pitchMap
	"return the name dictionary"

	^NameMap! !

!Pitch class methodsFor: 'examples'!

pitchExample
	"print a simple message to the transcript"
	"Pitch pitchMap inspect"
	"Pitch pitchExample"

	| me |
	me _ Pitch value: #c3.
	Transcript show: me printString; cr;
		show: me asHertz printString; cr;
		show: me asMidi printString; cr.! !

Pitch initialize!


MusicMagnitude subclass: #PField
	instanceVariableNames: 'name '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Magnitudes'!
PField comment:
'PFields are used for arbitrary parameters of musical events.
Instance variable:
	field		<Integer> field in the score

Note also that i use my type as a symbol for the optional declaration of my parameter or units.

Examples:
	PField field: 5 value: 0.7 type: #position
	PField field: 9 value: ''legato'' type: #expression
'!


!PField methodsFor: 'accessing'!

name
	"Return the instance variable 'name'."

	^name!

name: theName
	"Accept the argument, 'theName', as the new instance nariable 'name'."

	name _ theName! !

Pitch subclass: #HertzPitch
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Magnitudes'!


!HertzPitch methodsFor: 'converting'!

asHertz
	^value!

asMidi
	"return the MIDI-fied value of the receiver pitch"

	^self convertHertzToMidi!

asSymbol
	"return the symbolic value of the receiver pitch"

	^self convertHertzToSymbol! !

!HertzPitch methodsFor: 'printing'!

printOn: aStream
	"print me as a pitch string"

	^aStream nextPutAll: 'Pitch: ', value printString, 'Hz.'! !

!HertzPitch methodsFor: 'private'!

convertHertzToMidi
	"assuming value is a frequency, return a key number"
	"(Pitch value: 440.0) asMidi"		"440.0 is Midi key 45"
	"(Pitch value: 262.0) asMidi"		"261.3 is Midi key 36"

	| theOctave theA theNote theHalfStep |
	theOctave _ (value / 27.5) truncated floorLog: 2.
	theA _ 27.5 timesTwoPower: theOctave.
	theHalfStep _ 2 raisedTo: (1 / 12).
	theNote _ value asFloat / theA.
	theNote _ theNote floorLog: theHalfStep.
	^(theOctave * 12 + theNote - 3) truncated!

convertHertzToSymbol
	"assuming value is a frequency, return a symbolic note name"
	"(Pitch value: 440.0) asSymbol"			"440.0 is Midi key 45"
	"(Pitch value: 262.0) asSymbol"			"261.3 is Midi key 36"

	^(Pitch value: self convertHertzToMidi) convertMidiToSymbol! !

MusicMagnitude subclass: #Amplitude
	instanceVariableNames: ''
	classVariableNames: 'DynamicMap '
	poolDictionaries: ''
	category: 'Music-Magnitudes'!
Amplitude comment:
'Amplitudes are used for the loudness parameters of musical events.
Amplitudes generally come in several flavors, as signified by the instance variable "type".
Class variable:
	DynamicMap		<Dictionary of interval -> symbol> map of symbolic dynamics to the range of 0 to 1000

Examples:
	Amplitude value: 0.77		"create a float instance - range 0.0 to 1.0 (cmusic)"
	Amplitude value: 77			"create an integer instance - range 0 to 127 (Midi)"
	Amplitude value: #mp		"create a symbolic instance - range #ppp to #fff"
'!


!Amplitude methodsFor: 'accessing'!

value: theValue
	"Accept the argument, 'theValue', as the new instance variable 'value';."

	theValue class == Float					"assume ratio"
			ifTrue: [^RatioAmplitude new realValue: theValue].
	theValue class == SmallInteger			"assume key velocity"
			ifTrue: [^MidiAmplitude new realValue: theValue].
	(theValue isKindOf: String) 		"assume dynamic name string"
			ifTrue: [^SymbolicAmplitude new realValue: theValue].
	self error: 'unknown Amplitude type.'! !

!Amplitude methodsFor: 'converting'!

mostGeneral
	"what's the most general type of me"

	^self asMidi! !

!Amplitude methodsFor: 'printing'!

printOn: aStream
	"print me as an amplitude string"

	^aStream nextPutAll: 'Amp: ', value printString!

storeOn: aStream
	"print me as an amplitude declaration"

	aStream nextPutAll: '(Amplitude value: ', self value printString, ')'.! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Amplitude class
	instanceVariableNames: ''!


!Amplitude class methodsFor: 'class initialization'!

initialize
	"set up the class dynamic mapping dictionary"
	"Amplitude initialize"
	
	DynamicMap _ Dictionary new.
	DynamicMap at: (1 to: 50) put: #ppp.
	DynamicMap at: (50 to: 150) put: #pp.
	DynamicMap at: (150 to: 250) put: #p.
	DynamicMap at: (250 to: 500) put: #mp.
	DynamicMap at: (500 to: 750) put: #mf.
	DynamicMap at: (750 to: 850) put: #f.
	DynamicMap at: (850 to: 950) put: #ff.
	DynamicMap at: (950 to: 1000) put: #fff.! !

!Amplitude class methodsFor: 'examples'!

ampExample
	"print a simple message to the transcript"
	"Amplitude ampExample"

	| me |
	me _ Amplitude value: #mf.
	Transcript show: me printString; cr;
		show: me asFloat printString; cr;
		show: me asMidi printString; cr.! !

Amplitude initialize!


Amplitude subclass: #RatioAmplitude
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Magnitudes'!


!RatioAmplitude methodsFor: 'converting'!

asFloat
	"return the value of the receiver as a loudness between 0 and 1"

	^value!

asMidi
	"return the MIDI-fied value of the receiver amplitude"

	^self convertFloatToMidi!

asSymbol
	"return the symbolic value of the receiver amplitude"

	^self convertFloatToName! !

!RatioAmplitude methodsFor: 'private'!

convertFloatToMidi
	"assuming value is a float (0.0 to 1.0), return a key velocity (0 to 127)"
	
	^(value * 127.0) asInteger!

convertFloatToName
	"assuming value is a float, return a symbolic dynamic"

	| key |
	key _ DynamicMap keys 
		select: [ :aK | aK includes: (value * 1000) ].	"map's range = 0 to 1000"
	key == nil 
		ifFalse: [^DynamicMap at: key ]! !

MusicMagnitude subclass: #Duration
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Magnitudes'!
Duration comment:
'Durations are used for the duration parameters of musical events.
Durations generally come in several flavors, as signified by the instance variable "type".

Examples:
	Duration value: 0.77		"create an instance for with seconds as the unit"
	Duration value: 770		"create an instance for with milliseconds as the unit"
	Duration value: 1/4			"create an instance for with beats as the unit"
'!


!Duration methodsFor: 'accessing'!

value: theValue
	"Accept the argument, 'theValue', as the new instance variable 'value';."

	theValue class == Float					"assume ratio"
			ifTrue: [^MSDuration new realValue: (theValue * 1000) truncated].
	theValue class == SmallInteger			"assume key velocity"
			ifTrue: [^MSDuration new realValue: theValue].
	theValue class == Fraction			"assume key velocity"
			ifTrue: [^BeatDuration new realValue: theValue].
	self error: 'unknown Duration type.'! !

!Duration methodsFor: 'converting'!

mostGeneral
	"what's the most general type of me"

	^self asMS!

truncated
	^self asMS! !

!Duration methodsFor: 'printing'!

printOn: aStream
	"print me as a duration string"

	^aStream nextPutAll: 'Dur: ', value printString!

storeOn: aStream
	"print me as a Duration declaration"

	aStream nextPutAll: '(Duration value: ', self value printString, ')'.! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Duration class
	instanceVariableNames: ''!


!Duration class methodsFor: 'examples'!

durExample
	"print a simple message to the transcript"
	"Duration durExample"

	| me |
	me _ Duration value: 1/4.
	Transcript show: me printString; cr;
		show: me asBeats printString; cr;
		show: me asMS printString; cr.! !

Pitch subclass: #SymbolicPitch
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Magnitudes'!


!SymbolicPitch methodsFor: 'converting'!

asHertz
	"return the value of the receiver pitch as a frequency"

	^self convertSymbolToHertz!

asMidi
	"return the MIDI-fied value of the receiver pitch"

	^self convertSymbolToMidi!

asSymbol
	^value! !

!SymbolicPitch methodsFor: 'printing'!

printOn: aStream
	"print me as a pitch string"

	^aStream nextPutAll: 'Pitch: ', value printString! !

!SymbolicPitch methodsFor: 'private'!

convertSymbolToHertz
	"assuming value is a symbolic note name, return a frequency "
	"(Pitch value: #a4) asHertz"		"440.0 is Midi key 45"
	"(Pitch value: #c3) asHertz"		"261.0 is Midi key 36"

	^27.50 * (2.0 raisedTo: 
		((self convertSymbolToMidi + 3) asFloat / 12.0))!

convertSymbolToMidi
	"assuming value is a symbolic note name, return a key number"
	"Midi is 0=c-based, NameMap is 0=a-based"
	"(Pitch value: #c3) asMidi"
	"(Pitch value: 'c#3' asSymbol) asMidi"

	| not oct |
	not _ String with: (value at: 1).
	oct _ 0.
	value size > 2 
		ifTrue: [ not _ not, (String with: (value at: 2)).
				oct _ value at: 3 ]
		ifFalse: [ oct _ value at: 2 ].
	^(NameMap at: not) - 3 + ((oct digitValue) * 12)! !

Amplitude subclass: #SymbolicAmplitude
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Magnitudes'!


!SymbolicAmplitude methodsFor: 'converting'!

asFloat
	"return the value of the receiver as a loudness between 0 and 1"

	^self convertNameToFloat!

asMidi
	"return the MIDI-fied value of the receiver amplitude"

	^self convertNameToMidi!

asSymbol
	"return the symbolic value of the receiver amplitude"

	^value asSymbol! !

!SymbolicAmplitude methodsFor: 'private'!

convertNameToFloat
	"assuming value is a symbolic dynamic, return a floating amplitude"

	| interval |
	interval _ DynamicMap keyAtValue: value.
	^(interval first / 1000) asFloat!

convertNameToMidi
	"assuming value is a symbolic dynamic, return a key velocity"

	| interval |
	interval _ DynamicMap keyAtValue: value asSymbol.
	^(interval first / 8) asInteger! !

Pitch subclass: #MidiPitch
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Magnitudes'!


!MidiPitch methodsFor: 'converting'!

asHertz
	"return the value of the receiver pitch as a frequency"

	^self convertMidiToHertz!

asMidi	
	^value!

asSymbol
	"return the symbolic value of the receiver pitch"

	^self convertMidiToSymbol! !

!MidiPitch methodsFor: 'printing'!

printOn: aStream
	"print me as a pitch string"

	^aStream nextPutAll: 'Pitch: ', value printString! !

!MidiPitch methodsFor: 'private'!

convertMidiToHertz
	"assuming value is a key number, return a frequency"
	"(Pitch value: 45) asHertz"		"440.0 is Midi key 45"
	"(Pitch value: 36) asHertz"		"261.0 is Midi key 36"

	^27.50 * (2.0 raisedTo: ((value + 3) asFloat / 12.0))!

convertMidiToSymbol
	"assuming value is a key number, return a symbolic pitch name"
	"(Pitch value: 37) name"

	^((NameMap keyAtValue: (value \\ 12) + 3), (value // 12) printString)! !

Amplitude subclass: #MidiAmplitude
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Magnitudes'!


!MidiAmplitude methodsFor: 'converting'!

asFloat
	"return the value of the receiver as a loudness between 0 and 1"

	^self convertMidiToFloat!

asMidi
	"return the MIDI-fied value of the receiver amplitude"

	^value!

asSymbol
	"return the symbolic value of the receiver amplitude"

	^self convertMidiToName! !

!MidiAmplitude methodsFor: 'private'!

convertMidiToFloat
	"assuming value is a key velocity, return a floating amplitude"
	
	^value / 127!

convertMidiToName
	"assuming value is a key velocity, return a symbolic dynamic name"

	| key |
	key _ DynamicMap keys 
		select: [ :aK | aK includes: (value * 8) ].	"map's range = 0 to 1000"
	key == nil 
		ifFalse: [^DynamicMap at: key ]! !

Duration subclass: #BeatDuration
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Magnitudes'!


!BeatDuration methodsFor: 'converting'!

asBeats

	^value!

asInteger

	^self asMS!

asMS
	"return the value of the receiver duration"

	^(value * 1000) truncated! !

Duration subclass: #MSDuration
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Music-Magnitudes'!


!MSDuration methodsFor: 'converting'!

as10MS
	"return the value of the receiver duration"

	^value // 10!

asBeats

	^(value / 1000) reduced!

asMS
	"return the value of the receiver duration"

	^value! !